#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include "LRSolver.h"
#include "simplex_opt.h"

LRSolver::SimplexOptimizer::SimplexOptimizer(LRSolver* _LR)
	: lambda(NULL), point_num(0), point_num_max(0), buf_num(0), buf_num_max(0), atom_num_max(0),
	c_buf(NULL), atom_id_buf(NULL), LR(_LR), total_time(0)
{
}

LRSolver::SimplexOptimizer::~SimplexOptimizer()
{
	Dealloc();
}

void LRSolver::SimplexOptimizer::Dealloc()
{
	if (lambda)
	{
		delete [] lambda;
		lambda = NULL;
	}
	if (point_num_max > 0)
	{
		delete [] buf;
		delete [] A;
		free(buf_atom_list);
		free(points);
		point_num_max = 0;
	}
	if (buf_num_max > 0)
	{
		free(c_buf);
		free(atom_id_buf);
		c_buf = NULL;
		atom_id_buf = NULL;
		buf_num_max = 0;
	}
	if (atom_num_max > 0)
	{
		delete [] _lambda_atom;
		atom_num_max = 0;
	}
}

void LRSolver::SimplexOptimizer::Alloc(int n_global_max)
{
	if (lambda) Dealloc();
	lambda = new double[3*n_global_max];
	weighted_lambda = lambda + n_global_max;
	n_max_buf = lambda + 2*n_global_max;
}

void LRSolver::SimplexOptimizer::Init(Subproblem* _S, int num_max)
{
	int p;
	S = _S;
	if (num_max > atom_num_max)
	{
		if (atom_num_max > 0) delete [] _lambda_atom;
		atom_num_max = 2*num_max;
		_lambda_atom = new MyDouble[2*atom_num_max];
		_weighted_lambda_atom = _lambda_atom + atom_num_max;
	}
	if (num_max >= point_num_max)
	{
		if (point_num_max == 0) 
		{
			point_num_max = 16;
			points = (Point*) malloc(point_num_max*sizeof(Point));
		}
		else
		{
			if (point_num_max < num_max) point_num_max = num_max;
			point_num_max *= 2;
			points = (Point*) realloc(points, point_num_max*sizeof(Point));
			delete [] buf;
			delete [] A;
			free(buf_atom_list);
		}
		A = new double*[point_num_max];
		buf = new double[point_num_max*point_num_max + 3*point_num_max];
		b     = buf +   point_num_max;
		alpha = buf + 2*point_num_max;
		A[0]  = buf + 3*point_num_max;
		for (p=1; p<point_num_max; p++) A[p] = A[p-1] + point_num_max;
		buf_atom_list = malloc(point_num_max*( (sizeof(Problem::Atom)>sizeof(int)) ? sizeof(Problem::Atom) : sizeof(int) ));
	}
	for (p=0; p<num_max; p++)
	{
		_lambda_atom[p].reset(); // = S->P->dot_product(S->atoms[i].atom, lambda);
		_weighted_lambda_atom[p].reset(); // = S->P->dot_product(S->atoms[i].atom, weighted_lambda);
	}
	point_num = 0;
	for (p=0; p<num_max; p++) alpha[p] = (p == 0) ? 1 : 0;
}



double* LRSolver::SimplexOptimizer::GetLambda(Subproblem* _S, double* xlambda_ptr)
{
	S = _S;
	double GAMMA = LR->GAMMA;

	double* xi = S->xi;
	double* mui = S->mui;

	if (xlambda_ptr == NULL)
	{
		MAPPING_FOR(S, k, i,\
		{\
			lambda[k] = 0;\
		},\
		{\
			lambda[k] += GAMMA*xi[k] + mui[k] - LR->nu[i];\
		})
	}
	else
	{
		_xlambda = 0;
		MAPPING_FOR(S, k, i,\
		{\
			lambda[k] = 0;\
		},\
		{\
			double L  = GAMMA*xi[k] + mui[k] - LR->nu[i];\
			////////////////////////////////////////////////
			lambda[k]          += L;\
			_xlambda          += xi[k] * L;\
		})
		*xlambda_ptr = _xlambda;
	}

	return lambda;
}

double* LRSolver::SimplexOptimizer::InitAll()
{
	double GAMMA = LR->GAMMA;

	_xlambda = 0;
	_weighted_xx = 0;
	_weighted_xlambda = 0;
	double* xi = S->xi;
	double* mui = S->mui;

	MAPPING_FOR(S, k, i,\
	{\
		lambda[k] = weighted_lambda[k] = 0;\
	},\
	{\
		double L  = GAMMA*xi[k]                     + mui[k] - LR->nu[i];\
		double WL = GAMMA*xi[k] / LR->var_counts[i] + mui[k] - LR->nu[i];\
		double weight = 1.0 - 1.0 / LR->var_counts[i];\
		////////////////////////////////////////////////
		lambda[k]          += L;\
		weighted_lambda[k] += WL;\
		_xlambda          += xi[k] * L;\
		_weighted_xx      += weight * xi[k] * xi[k];\
		_weighted_xlambda += xi[k] * WL;\
	})

	point_num = buf_num = 0;

	return lambda;
}

void LRSolver::SimplexOptimizer::InitWeightedLambda()
{
	double GAMMA = LR->GAMMA;

	_xlambda = 0;
	_weighted_xx = 0;
	_weighted_xlambda = 0;
	double* xi = S->xi;
	double* mui = S->mui;

	MAPPING_FOR(S, k, i,\
	{\
		weighted_lambda[k] = 0;\
	},\
	{\
		double WL = GAMMA*xi[k] / LR->var_counts[i] + mui[k] - LR->nu[i];\
		////////////////////////////////////////////////
		weighted_lambda[k] += WL;\
	})

	point_num = buf_num = 0;

}

void LRSolver::SimplexOptimizer::PointStart(double c0)
{
	if (point_num >= point_num_max)
	{
		printf("Fatal error in PointStart()\n"); exit(1);
	}
	points[point_num].c0 = c0;
	points[point_num].num = 0;
	points[point_num].c = c_buf + buf_num;
	points[point_num].atom_id = atom_id_buf + buf_num;
}

void LRSolver::SimplexOptimizer::PointAddAtom(double c, int atom_id)
{
	Point* P;
	if (buf_num >= buf_num_max)
	{
		double* c_buf_old = c_buf;
		int* atom_id_buf_old = atom_id_buf;
		if (buf_num_max == 0)
		{
			buf_num_max = 16;
			c_buf = (double*) malloc(buf_num_max*sizeof(double));
			atom_id_buf = (int*) malloc(buf_num_max*sizeof(int));
		}
		else
		{
			buf_num_max *= 2;
			c_buf = (double*) realloc(c_buf, buf_num_max*sizeof(double));
			atom_id_buf = (int*) realloc(atom_id_buf, buf_num_max*sizeof(int));
		}
		for (P=points; P<=points+point_num; P++)
		{
			P->c = c_buf + (int)(P->c - c_buf_old);
			P->atom_id = atom_id_buf + (int)(P->atom_id - atom_id_buf_old);
		}
	}
	P = &points[point_num];
	P->c[P->num] = c;
	P->atom_id[P->num] = atom_id;
	P->num ++;
	buf_num ++;
}

void LRSolver::SimplexOptimizer::PointFinish()
{
	point_num ++;
}

void LRSolver::SimplexOptimizer::ResetPoints()
{
	int p;
	point_num = 0;
	for (p=0; p<point_num_max; p++) alpha[p] = (p == 0) ? 1 : 0;
}


void LRSolver::SimplexOptimizer::MoveToPoint(bool call_update_x)
{
	int p, q, k;
	Point* P;

	double c0 = 0;
	for (p=0; p<S->atom_num; p++) S->atoms[p].c = 0;
	for (p=0, P=points; p<point_num; p++, P++)
	{
		if (alpha[p] <= 0) continue;
		if (P->c0 == 0 && P->num == 1)
		{
			k = P->atom_id[0];
			S->atoms[k].c += alpha[p];
			S->TouchAtom(k);
		}
		else
		{
			c0 += P->c0*alpha[p];
			for (k=0; k<P->num; k++)
			{
				S->atoms[P->atom_id[k]].c += P->c[k]*alpha[p];
			}
		}
	}

	double* y = weighted_lambda;
	memcpy(y, S->xi, S->n*sizeof(double));
	double y_cost = S->xi_cost;
	if (c0 != 1)
	{
		for (k=0; k<S->n; k++) y[k] *= c0;
		y_cost *= c0;
	}
	for (p=0; p<S->atom_num; p++)
	{
		if (S->atoms[p].c != 0)
		{
			S->P->add_atom_to_vector(S->atoms[p].c, S->atoms[p].atom, y);
			y_cost += S->atoms[p].c * S->atoms[p].cost;
		}
	}
	S->MoveToPoint(y, y_cost);
	if (call_update_x)
	{
		Problem::Atom* atom_list = (Problem::Atom*) buf_atom_list;
		for (p=q=0; p<S->atom_num; p++)
		{
			if (S->atoms[p].c == 0) continue;
			buf[q] = S->atoms[p].c;
			atom_list[q] = S->atoms[p].atom;
			q ++;
		}
		S->P->update_x(S->xi, c0, q, buf, atom_list);
	}
}




double* LRSolver::SimplexOptimizer::GetUpdatedLambda()
{
	double* updated_lambda = GetBuf();
	double GAMMA = LR->GAMMA;
	double* xi = S->xi;

	int p, k;
	Point* P;

	double c0 = -1;
	for (p=0; p<S->atom_num; p++) S->atoms[p].c = 0;
	for (p=0, P=points; p<point_num; p++, P++)
	{
		if (alpha[p] <= 0) continue;
		c0 += P->c0*alpha[p];
		for (k=0; k<P->num; k++)
		{
			S->atoms[P->atom_id[k]].c += P->c[k]*alpha[p];
		}
	}

	for (k=0; k<S->n; k++) updated_lambda[k] = lambda[k] + GAMMA*c0*xi[k];
	for (p=0; p<S->atom_num; p++)
	{
		if (S->atoms[p].c != 0)
		{
			S->P->add_atom_to_vector(LR->GAMMA*S->atoms[p].c, S->atoms[p].atom, updated_lambda);
		}
	}

	return updated_lambda;
}



//#define DEBUG_MINIMIZE_SIMPLEX

double* LRSolver::SimplexOptimizer::Minimize(double gap_threshold, double gap_reduction_threshold, double* delta_f)
{
	if (S->atom_num > atom_num_max)
	{
		printf("Fatal error in Minimize()\n");
		exit(1);
	}
	double T_start = get_time();

	double GAMMA = LR->GAMMA;
	Point* P;
	Point* Q;
	int p, q, i, j;
	int N = point_num;
	MyAtom* atoms = S->atoms;

#ifdef DEBUG_MINIMIZE_SIMPLEX
	int k, n=S->n;
	double* xi = S->xi;
	double** POINTS = new double*[point_num];
	POINTS[0] = new double[point_num*n+point_num];
	double* POINT_COSTS = POINTS[0] + point_num*n;
	for (p=1; p<point_num; p++) POINTS[p] = POINTS[p-1] + n;
	for (p=0, P=points; p<N; p++, P++)
	{
		for (k=0; k<n; k++) POINTS[p][k] = P->c0*xi[k];
		POINT_COSTS[p] = P->c0*S->xi_cost;
		for (i=0; i<P->num; i++) 
		{
			MyAtom* a = &atoms[P->atom_id[i]];
			S->P->add_atom_to_vector(P->c[i], a->atom, POINTS[p]);
			POINT_COSTS[p] += P->c[i]*a->cost;
		}
	}
#endif


	for (p=0, P=points; p<N; p++, P++)
	for (q=0, Q=points; q<N; q++, Q++)
	{
		A[p][q] = GAMMA * P->c0 * Q->c0 * _weighted_xx;
		for (i=0; i<P->num; i++)
		for (j=0; j<Q->num; j++)
		{
			int _i = P->atom_id[i], _j = Q->atom_id[j];
			A[p][q] += GAMMA * P->c[i] * Q->c[j] * S->get_dot_product(_i, _j);
		}
		for (j=0; j<Q->num; j++)
		{
			int _j = Q->atom_id[j];
			A[p][q] += P->c0 * Q->c[j] * (lambda_atom(_j) - weighted_lambda_atom(_j));
		}
		for (i=0; i<P->num; i++)
		{
			int _i = P->atom_id[i];
			A[p][q] += P->c[i] * Q->c0 * (lambda_atom(_i) - weighted_lambda_atom(_i));
		}
#ifdef DEBUG_MINIMIZE_SIMPLEX
		double Apq = 0;
		for (k=0; k<n; k++) Apq += GAMMA*S->P->dot_product_weights[k]*POINTS[p][k]*POINTS[q][k];
		if (fabs(Apq - A[p][q]) > 1e-6)
		{
			printf("Error in MinimizeOverSimplex() [%f %f]\n", Apq, A[p][q]);
			exit(1);
		}
#endif
	}
	for (p=0, P=points; p<N; p++, P++)
	{
		b[p] = P->c0 * ( _weighted_xlambda + S->xi_cost );
		for (i=0; i<P->num; i++)
		{
			int _i = P->atom_id[i];
			b[p] += P->c[i] * ( weighted_lambda_atom(_i) + atoms[_i].cost );
		}

#ifdef DEBUG_MINIMIZE_SIMPLEX
		double bp = POINT_COSTS[p];
		for (k=0; k<n; k++) bp += POINTS[p][k]*weighted_lambda[k];
		if (fabs(bp - b[p]) > 1e-6)
		{
			printf("Error in MinimizeOverSimplex() [%f %f]\n", bp, b[p]);
			exit(1);
		}
#endif
	}

	OPT.setup(A, b, point_num);
	*delta_f += OPT.solve(alpha, gap_threshold, gap_reduction_threshold);

#ifdef DEBUG_MINIMIZE_SIMPLEX
	delete [] POINTS[0];
	delete [] POINTS;
#endif

	total_time += get_time() - T_start;

	return alpha;
}



void LRSolver::SimplexOptimizer::MinimizeConv(double gap_threshold, double gap_reduction_threshold, double* delta_f)
{
	if (S->atom_num > atom_num_max || S->atom_num > point_num_max)
	{
		printf("Fatal error in Minimize()\n");
		exit(1);
	}

	double T_start = get_time();

	int p, q, atom_num = S->atom_num;
	double GAMMA = LR->GAMMA;

	for (p=0; p<atom_num; p++)
	for (q=0; q<atom_num; q++)
	{
		A[p][q] = GAMMA*S->get_dot_product(p, q);
	}
	for (p=0; p<atom_num; p++)
	{
		b[p] = S->P->dot_product(S->atoms[p].atom, weighted_lambda) + S->atoms[p].cost;
		alpha[p] = S->atoms[p].weight;
	}

	int* mapping = (int*) buf_atom_list;
	for (p=0; p<atom_num; p++) mapping[p] = p;
	int N = atom_num;

	OPT.setup(A, b, N);
	double cost, gap;
	cost = OPT.evaluate(alpha, &gap);
	gap *= gap_reduction_threshold;
	if (gap_threshold < gap) gap_threshold = gap;
	while ( 1 )
	{
		OPT.setup(A, b, N);
		*delta_f += OPT.solve_inner(alpha, gap_threshold, 0);
		int N_old = N;
		for (p=0; p<N; ) 
		{
			if (alpha[p] > 0) p ++;
			else
			{
				// move point at position N-1 to position p 
				N --;
				mapping[p] = mapping[N];
				alpha[p] = alpha[N];
				b[p] = b[N];

				for (q=0; q<N; q++)
				{
					A[q][p] = A[q][N];
					A[p][q] = A[N][q];
				}
				A[p][p] = A[N][N];
			}
		}
		if (N == N_old || N <= 1) break;
	}

	for (p=0; p<atom_num; p++) S->atoms[p].weight = 0;
	for (p=0; p<N; p++) S->atoms[mapping[p]].weight = alpha[p];

	total_time += get_time() - T_start;
}

void LRSolver::SimplexOptimizer::MoveToPointConv()
{
	int p, q;
	S->Recompute_x(false);

	Problem::Atom* atom_list = (Problem::Atom*) buf_atom_list;
	for (p=q=0; p<S->atom_num; p++)
	{
		if (S->atoms[p].weight <= 0) continue;
		alpha[q] = S->atoms[p].weight;
		atom_list[q] = S->atoms[p].atom;
		q ++;
	}
	S->P->update_x(S->xi, 0, q, alpha, atom_list);
}



