#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include "LRSolver.h"


LRSolver::LRSolver(int _var_num)
	: var_num(_var_num), 
	parent_subproblem_num(0),
	S_being_contracted(NULL),
	x_global(NULL), mu_global(NULL), 
	n_global(0), n_global_max(0),
	n_global_last_solve(0),
	GAMMA(0),
	cost_const(0),
	D(this),
	buffer(512)
{
	var_counts = new int[var_num];
	memset(var_counts, 0, var_num*sizeof(int));
	nu = new double[var_num];
	memset(nu, 0, var_num*sizeof(double));
}

LRSolver::~LRSolver()
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	for (S=L.First(0, 1); S; S=L.Next())
	{
		if (S->next) 
		{
			delete [] S->next->mapping;
			delete [] S->next->xi;
		}
	}
	for (S=L.First(0, 2); S; S=L.Next())
	{
		S->Delete();
	}
	delete [] var_counts;
	delete [] nu;
	if (x_global) delete [] x_global;
	if (mu_global) delete [] mu_global;
}

void LRSolver::Subproblem::Init(LRSolver* _LR, Problem* _P, int _n, SubproblemOptions* _options)
{
	LR = _LR;
	P = _P;
	n = _n;
	options = _options;
	atom_num = 0;
	mapping = mapping_to_parent = mapping_to_child = NULL;
	N = n;
	next = parent = NULL;
	atom_num_max = 0;
	atoms = NULL;
	_dot_products = NULL;
	xi = NULL;
	P->dot_product_weights = new double[n];
	P->LR = LR;
	P->_S = this;
}

void LRSolver::Subproblem::Delete()
{
	if (atoms)
	{
		int p;
		for (p=0; p<atom_num; p++) P->destroy_atom(atoms[p].atom);
		free(atoms);
		delete [] _dot_products[0];
		delete [] _dot_products;
	}
	if (P->dot_product_weights) delete [] P->dot_product_weights;
	delete P;

	LR->_Slist.Remove(this);
}

void LRSolver::AddSubproblem(Problem *P, int n, int* mapping, SubproblemOptions* options, bool copy_mapping, bool copy_options)
{
	int i;

	Subproblem* S = _Slist.New(1);
	if (copy_options)
	{
		SubproblemOptions* o = (SubproblemOptions*) buffer.Alloc(sizeof(SubproblemOptions));
		memcpy(o, options, sizeof(SubproblemOptions));
		options = o;
	}

	S->Init(this, P, n, options);
	if (copy_mapping)
	{
		S->mapping = (int*) buffer.Alloc(n*sizeof(int));
		memcpy(S->mapping, mapping, n*sizeof(int));
	}
	else S->mapping = mapping;
	for (i=0; i<n; i++) var_counts[mapping[i]] ++;
	n_global += n;
	if (n_global_max < n) n_global_max = n;
	S->P->parent_subproblem_id = parent_subproblem_num ++;
	S->P->child_subproblem_id = -1;
}



int LRSolver::Subproblem::AddAtom(Problem::Atom& a, double cost)
{
	int p, q;
	for (p=0; p<atom_num; p++)
	{
		if (cost == atoms[p].cost && P->compare_atoms(atoms[p].atom, a) == 0)
		{
			P->destroy_atom(a);
			a = atoms[p].atom;
			return p;
		}
	}
	if (atom_num >= atom_num_max)
	{
		if (atom_num_max == 0)
		{
			atom_num_max = 16;
			atoms = (MyAtom*) malloc(atom_num_max*sizeof(MyAtom));
		}
		else
		{
			atom_num_max *= 2;
			atoms = (MyAtom*) realloc(atoms, atom_num_max*sizeof(MyAtom));
		}
		double** D = new double*[atom_num_max];
		D[0] = new double[atom_num_max*atom_num_max];
		for (p=1; p<atom_num_max; p++) D[p] = D[p-1] + atom_num_max;
		if (_dot_products)
		{
			for (p=0; p<atom_num; p++) memcpy(D[p], _dot_products[p], atom_num*sizeof(double));
			delete [] _dot_products[0];
			delete [] _dot_products;
		}
		_dot_products = D;
	}
	p = atom_num ++;
	atoms[p].atom = a;
	atoms[p].cost = cost;
	atoms[p].weight = (options->conv) ? 0 : (options->inactive_atom_iter_max+1);

	for (q=0; q<=p; q++)
	{
		_dot_products[p][q] = 2;
		_dot_products[q][p] = -2;
	}

	return p;
}


void LRSolver::Subproblem::TouchAtom(int p)
{
	if (!options->conv)
	{
		atoms[p].weight = -1;
	}
}

void LRSolver::Subproblem::PruneAtoms(int* keep_index)
{
	int p;
	if (options->conv)
	{
		for (p=atom_num-1; p>=0; p--)
		{
			if (atoms[p].weight <= options->weight_eps && atom_num>=2) RemoveAtom(p, keep_index);
		}
	}
	else
	{
		for (p=0; p<atom_num; p++) atoms[p].weight += 1;
		for (p=atom_num-1; p>=0; p--)
		{
			if (atoms[p].weight > options->inactive_atom_iter_max) RemoveAtom(p, keep_index);
		}
		while ( atom_num > options->max_cache_size )
		{
			int p_max = -1;
			int p_keep = (keep_index) ? (*keep_index) : -1;
			double weight_max = 0;
			for (p=0; p<atom_num; p++)
			{
				if (p == p_keep) continue;
				if (p_max < 0 || weight_max <= atoms[p].weight) { p_max = p; weight_max = atoms[p].weight; }
			}
			if (p_max < 0) break;
			RemoveAtom(p_max, keep_index);
		}
	}
}

void LRSolver::Subproblem::RemoveAtom(int p, int* keep_index)
{
	if (keep_index && *keep_index == p) return;

	P->destroy_atom(atoms[p].atom);
	int q = -- atom_num;

	if (keep_index && *keep_index == q) *keep_index = p;

	memcpy(&atoms[p], &atoms[q], sizeof(MyAtom));

	// p = 1, q = 3. *: keep, .: remove, ABCDX: to copy

	// * . * A
	// . . . .
	// * . * B
	// C . D X
	//			 =>
	// * A *
	// C X D
	// * B *

	int a;
	for (a=0; a<atom_num; a++)
	{
		_dot_products[a][p] = _dot_products[a][q];
		_dot_products[p][a] = _dot_products[q][a];
	}
	_dot_products[p][p] = _dot_products[q][q];
}




void LRSolver::PrintVectors()
{
	int k;
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;

	printf("nu: ");
	for (k=0; k<var_num; k++) printf("%f ", nu[k]);
	printf("\n");

	printf("x:\n");
	for (S=L.First(0, 1); S; S=L.Next())
	{
		printf("\t");
		for (k=0; k<S->n; k++) printf("%f ", S->xi[k]);
		printf("  %f\n", S->xi_cost);
	}

	printf("lambda:\n");
	for (S=L.First(0, 1); S; S=L.Next())
	{
		double* lambda = D.GetLambda(S);
		printf("\t");
		for (k=0; k<S->n; k++) printf("%f ", lambda[k]);
		printf("\n");
	}

	printf("mu:\n");
	for (S=L.First(0, 1); S; S=L.Next())
	{
		printf("\t");
		for (k=0; k<S->n; k++) printf("%f ", S->mui[k]);
		printf("\n");
	}

	printf("=========================\n");
}

void LRSolver::Subproblem::SetDotProductWeights()
{
	MAPPING_FOR(this, k, i,\
	{\
		P->dot_product_weights[k] = 0;\
	},\
	{\
		P->dot_product_weights[k] += 1.0 - 1.0 / LR->var_counts[i];\
	})
}

void LRSolver::InitSolve( )
{
	if (n_global_last_solve == n_global) return;

	D.Alloc(n_global_max);

	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;

	double* x_global_prev = x_global;
	double* mu_global_prev = mu_global;

	x_global = new double[n_global];
	memset(x_global, 0, n_global*sizeof(double));

	mu_global = new double[3*n_global];
	memset(mu_global, 0, n_global*sizeof(double));

	memset(nu, 0, var_num*sizeof(double));
	double* lambda = mu_global+n_global; memset(lambda, 0, n_global_max*sizeof(double));
	H_best = -1e100;

	n_global = 0;
	for (S=L.First(0, 1); S; S=L.Next())
	{
		S->SetDotProductWeights();

		double* xi_prev = S->xi;
		double* mui_prev = S->mui;
		bool copy_flag = (S->xi != NULL);

		S->xi = x_global + n_global;
		S->mui = mu_global + n_global;
		n_global += S->n;

		if (copy_flag)
		{
			memcpy(S->xi, xi_prev, S->n*sizeof(double));
			memcpy(S->mui, mui_prev, S->n*sizeof(double));
		}
		else
		{
			Problem::Atom a = S->P->min_oracle(lambda);
			S->P->init_x(a);
			S->P->atom_to_vector(a, S->xi); S->xi_cost = S->P->compute_atom_cost(a);
			int p = S->AddAtom(a, S->xi_cost);
			if (S->options->conv)
			{
				S->atoms[p].weight = 1;
			}
			else
			{
				S->TouchAtom(p);
			}
		}

		int k;
		for (k=0; k<S->n; k++)
		{
			int i = S->mapping[k];
			nu[i] += S->xi[k] * GAMMA / var_counts[i];
		}
	}
	if (x_global_prev) delete [] x_global_prev;
	if (mu_global_prev) delete [] mu_global_prev;
	n_global_last_solve = n_global;
}

void LRSolver::SetNu()
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	int k;

	memset(nu, 0, var_num*sizeof(double));

	for (S=L.First(0, 1); S; S=L.Next())
	{
		for (k=0; k<S->n; k++) nu[S->mapping[k]] += GAMMA*S->xi[k] + S->mui[k];
	}
	for (k=0; k<var_num; k++) nu[k] /= var_counts[k];
}



void LRSolver::ComputeFullLambda(double* lambda)
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	for (S=L.First(0, 1); S; S=L.Next())
	{
		double* lambdai = D.GetLambda(S);
		memcpy(lambda, lambdai, S->n*sizeof(double));
		lambda += S->n;
	}
}

void LRSolver::Subproblem::Recompute_x(bool renormalize_weights, bool check)
{
	if (!options->conv) { printf("incorrect call to Recompute_x()\n"); exit(1); }

	int p;

	if (renormalize_weights)
	{
		double sum = 0;
		for (p=0; p<atom_num; p++) sum += atoms[p].weight;
		if (sum > 0)
		{
			for (p=0; p<atom_num; p++) atoms[p].weight /= sum;
		}
		else
		{
			for (p=0; p<atom_num; p++) atoms[p].weight = 1.0 / atom_num;
		}
	}

	double* y = LR->D.GetBuf();
	memset(y, 0, n*sizeof(double));
	double y_cost = 0;
	for (p=0; p<atom_num; p++)
	{
		MyAtom* a = &atoms[p];
		if (a->weight <= 0) continue;
		P->add_atom_to_vector(a->weight, a->atom, y);
		y_cost += a->weight*a->cost;
	}

	if (0 && check)
	{
		double* xi_check = new double[n];
		memcpy(xi_check, xi, n*sizeof(double));
		MoveToPoint(y, y_cost);
		int i;
		for (i=0; i<n; i++)
		{
			if (fabs(xi_check[i] - xi[i]) > 1e-6)
			{
				printf("Recompute_x(): vector has changed! (diff=%f)\n", fabs(xi_check[i] - xi[i]));
				exit(1);
			}
		}
		delete [] xi_check;
	}
	else
	{
		MoveToPoint(y, y_cost);
	}
}

/*
double LRSolver::ReadPrimalSolution(double* x, bool compute_cost)
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	int k;
	double* xi = new double[n_global_max];
	memset(x, 0, var_num*sizeof(double));
	for (S=L.First(0, 1); S; S=L.Next())
	{
		for (k=0; k<S->n; k++) x[S->mapping[k]] += S->xi[k];
	}
	for (k=0; k<var_num; k++) x[k] /= var_counts[k];

	double primal_cost = cost_const;
	if (compute_cost)
	{
		for (S=L.First(0, 1); S; S=L.Next())
		{
			for (k=0; k<S->n; k++) xi[k] = x[S->mapping[k]];
			double xi_cost = S->P->compute_x_cost(xi);
			primal_cost += xi_cost;
		}
	}
	delete [] xi;
	return primal_cost;
}
*/
void LRSolver::ReadPrimalSolution(double* x)
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	int k;
	memset(x, 0, var_num * sizeof(double));
	for (S = L.First(0, 1); S; S = L.Next())
	{
		for (k = 0; k < S->n; k++) x[S->mapping[k]] += S->xi[k];
	}
	for (k = 0; k < var_num; k++) x[k] /= var_counts[k];
}

double LRSolver::ComputePrimalSolutionCost(double* x)
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	int k;
	double* xi = new double[n_global_max];

	double primal_cost = cost_const;
	for (S = L.First(0, 1); S; S = L.Next())
	{
		for (k = 0; k < S->n; k++) xi[k] = x[S->mapping[k]];
		double xi_cost = S->P->compute_x_cost(xi);
		primal_cost += xi_cost;
	}
	delete [] xi;
	return primal_cost;
}

double LRSolver::ComputeH(double* _gap, double* lambda_full)
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	double H = cost_const;
	double gap = 0;

	SetNu();

	for (S=L.First(0, 1); S; S=L.Next())
	{
		double gx = 0;
		double* lambda = D.GetLambda(S, &gx);
		if (lambda_full)
		{
			memcpy(lambda_full, lambda, S->n*sizeof(double));
			lambda_full += S->n;
		}
		gx += S->xi_cost;

		Problem::Atom a = S->P->min_oracle(lambda);
		double ga = S->P->dot_product(a, lambda) + S->P->compute_atom_cost(a);
		S->P->destroy_atom(a);

		gap += gx - ga;
		H += ga;
	}

	if (options.verbose >= 1)
	{
		printf("%f secs, lower_bound=%f, prox. gap=%f\n", get_time()-T_start, H, gap);
	}
	*_gap = gap;

	if (H_best < H) H_best = H;

	return H;
}

void LRSolver::Subproblem::PrintAtoms()
{
	int p;
	for (p=0; p<atom_num; p++)
	{
		P->print_atom(atoms[p].atom);
		printf("(%f) ", atoms[p].weight);
	}
}

double LRSolver::ComputeFmu()
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	double sum1 = 0, sum2 = 0, sum3 = 0;
	int k;

	for (S=L.First(0, 1); S; S=L.Next())
	{
		double* xi = S->xi;
		double* mui = S->mui;
		for (k=0; k<S->n; k++)
		{
			sum1 += xi[k]*xi[k];
			sum2 += xi[k]*mui[k];
		}
		sum2 += S->xi_cost;
	}
	for (k=0; k<var_num; k++) sum3 += var_counts[k]*nu[k]*nu[k];

	return cost_const + GAMMA*sum1/2 + sum2 - sum3/(2*GAMMA);
}






void LRSolver::Solve()
{
	T_start = get_time();

	GAMMA = options.gamma;
	InitSolve();

	double* mu1 = mu_global +   n_global;
	double* mu2 = mu_global + 2*n_global;
	int i;
	double gap0;
	int iter;
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;

	double H0 = ComputeH(&gap0);

	if (options.verbose >= 1)
	{
		printf("********************************\n");
		printf("*** GAMMA=%e ***\n", GAMMA);
		printf("********************************\n");
		printf("gap0=%e\n", gap0);
	}

	double H_prev = H0;
	double t1;
	bool restart = true;

	double T_callback_fn = 0;
	double T_last_callback_end = 0;
	double T_adjustx_fn = 0;
	double T_last_adjustx_end = 0;

	for (iter=0; iter<options.mu_update_max; iter++)
	{
		ComputeFullLambda(mu2);

		double mu_dist = 0;
		if (iter > 0)
		{
			int i;
			for (i=0; i<n_global; i++) mu_dist += (mu_global[i]-mu2[i])*(mu_global[i]-mu2[i]);
			mu_dist = sqrt(mu_dist);
			if (mu_dist <= options.mu_eps) break;
		}

		if (options.PPA_flag >= 1)
		{
			if (restart)
			{
				t1 = 1;
				memcpy(mu1,       mu2, n_global*sizeof(double));
				memcpy(mu_global, mu2, n_global*sizeof(double));
				restart = false;
			}
			else
			{
				double t2 = (1 + sqrt(1+4*t1*t1)) / 2;
				double b = - ( t1 - 1 ) / t2, a = 1 - b;
				t1 = t2;

				// set mu1 := mu2
				for (i=0; i<n_global; i++) mu_global[i] = a*mu2[i] + b*mu1[i];
				double* mu_tmp = mu1; mu1 = mu2; mu2 = mu_tmp;
			}
		}
		else
		{
			memcpy(mu_global, mu2, n_global*sizeof(double));
		}

		SetNu();

		double gap_threshold = gap0 / pow((double)(iter+1), options.PPA_alpha);
		if (options.verbose >= 1) { printf("=== starting prox. step #%d with eps=%.2f; ||mu'-mu|=%f\n", iter, gap_threshold, mu_dist); }

		double H = MinimizeFmu(gap_threshold);
		if (options.PPA_flag == 2 && H < H_prev)
		{
			restart = true;
			if (options.verbose >= 1) { printf("RESTARTING A-PPA\n"); }
		}
		H_prev = H;

		if (get_time() - T_start > options.time_max)
		{
			double gap;
			ComputeH(&gap);
			break;
		}
		
		if (options.callback_freq > 0 && options.callback_fn)
		{
			double A = get_time();
			if (T_callback_fn <= 0 || A-T_last_callback_end >= options.callback_freq*T_callback_fn)
			{
				options.callback_fn(options.callback_arg);
				T_last_callback_end = get_time();
				T_callback_fn = T_last_callback_end - A;
				T_start += T_callback_fn;
			}
		}

		if (options.adjustx_freq > 0)
		{
			double A = get_time();
			if (T_adjustx_fn <= 0 || A-T_last_adjustx_end >= options.adjustx_freq*T_adjustx_fn)
			{
				for (S=L.First(0, 1); S; S=L.Next())
				{
					S->xi_cost = S->P->adjust_x(S->xi);
				}
				T_last_adjustx_end = get_time();
				T_adjustx_fn = T_last_adjustx_end - A;
			}
		}
	}


	for (S=L.First(0, 0); S; S=L.Next())
	{
		int c = S->options->contract;
		S->options->contract = 0;
		S->Contract();
		S->options->contract = c;
	}
}

void LRSolver::ReadDualSolution()
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	for (S=L.First(0, 1); S; S=L.Next())
	{
		double gx = 0;
		double* lambda = D.GetLambda(S, &gx);
		S->P->set_final_dual_solution(lambda);
	}
}





