/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2022

    This file is part of LRSolver.

    LRSolver is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    LRSolver is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with LRSolver.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Implementation of the method described in

	Paul Swoboda and Vladimir Kolmogorov. ""MAP inference via Block-Coordinate Frank-Wolfe Algorithm". CVPR 2019.
	Vladimir Kolmogorov and Thomas Pock. "One-sided Frank-Wolfe algorithms for saddle problems". ICML 2021.
	Vladimir Kolmogorov. "Solving relaxations of MAP-MRF problems: Combinatorial in-face Frank-Wolfe directions". arXiv:2010.09567v4, 2022.


Goal: minimize function
f(x) = \sum_{i=0}^{n-1} fi(x[Ai])
over some (possibly discrete) set x\in X.

Ai is some subset of {0,1,...,ni-1}, and x[Ai] is the the restriction of x to the variables in this subset.
Function f depends on 'var_num' variables, term (or subproblem) i depends on ni=|Ai| variables where 1<=ni<=var_num
It is assumed \min_xi [fi(xi)+<lambdai,xi>] can be solved efficiently for any given i and lambdai


To specify subproblem i, the user must create an instance of class 'Problem' (defined in problem.h),
and call LRSolver::AddSubproblem(...) for each subproblem, see problem.h.

*/




#ifndef __LR_SOLVER_H__
#define __LR_SOLVER_H__


#include "problem.h"
#include "block.h"
#include "SList.h"
#include "LRTimer.h"
#include "simplex_opt.h"

// "LR" stands for "Lagrangian Relaxation"

class LRSolver
{
public:
	LRSolver(int var_num);
	~LRSolver();

	struct GlobalOptions
	{
		GlobalOptions() :
			mu_update_max(1000),
			time_max(100.),
			mu_eps(1e-6),

			PPA_flag(2),
			PPA_alpha(2.),
			major_iter_min(0), major_iter_max(1000), major_iter_max_log(0),

			ratio_minor(1.0),
			time_minor_max(5.0),
			ratio_cache(1.0),
			time_cache_max(5.0),
			minor_iter_min(0), minor_iter_max(1000),
			cache_iter_min(0), cache_iter_max(1000),

			compute_gap_freq(3),
			
			callback_freq(0),
			callback_fn(NULL),
			callback_arg(NULL),
			
			adjustx_freq(0),

			gamma(1.0),
			verbose(1)
		{}

		double gamma; // MOST IMPORTANT PARAMETER!!

		double time_max;
		int mu_update_max; // max number of updates of 'mu' (i.e. proximal centers)
		double mu_eps; // stop if ||mu'-mu|| <= mu_eps

		int PPA_flag; // 0: nonaccelerated PPA
		              // 1: accelerated PPA
		              // 2: accelerated PPA + restart when lower bound decreases
		double PPA_alpha; // for accelerated PPA_flag value PPA_alpha>=4 would give optimal convergence rate but seems to be slow (proximal subproblems are then solved very accurately).
		                  // Range PPA_alpha \in [2, 3] may work better.

		// termination criteria of minor iterations & cache loops
		double ratio_minor; // smaller value means sooner termination
		double ratio_cache; // smaller value means sooner termination
		double time_minor_max; // relative to the time of the major loop. Note: if no contractions are use, then should be set to 0 (meaning infty), since the time of the major loop is (almost) 0
		double time_cache_max; // relative to the time of the minor loop.
		int major_iter_min, major_iter_max;
		double major_iter_max_log; // terminate t's proximal step after 'major_iter_max + major_iter_max_log*log_2(t+1)' major iterations

		int minor_iter_min, minor_iter_max;
		int cache_iter_min, cache_iter_max;

		int compute_gap_freq; // compute lower bound and prox. gap after every 'computer_gap_freq' major loops
		
		double callback_freq; // if positive: call 'callback_fn' so that the time spent in this function is less the other time by factor 'callback_freq'
		void (*callback_fn)(void*);
		void* callback_arg;
	
		double adjustx_freq; // if positive: call 'Problem::adjust_x' for all subproblems so that the time spent in thes functions is less the other time by factor 'adjustx_freq'
	
		int verbose;
	} options;

	struct SubproblemOptions
	{
		SubproblemOptions() :
			contract(1),
			//contract_average_eps(1e-3),
			contract_average_eps(0),
			contract_threshold(0.75),
			max_cache_size(10),
			inactive_atom_iter_max(10),
			conv(false),
			weight_eps(1e-8),
			alg(FW),
			simplex_accuracy_before(0.1),
			simplex_accuracy_after(0.1),
			AFW_beta_threshold(0.999)
		{}

		typedef enum
		{
			FW, 
			DICG_PFW, // DICG moves + caching planes
			DICG_AFW, // DICG moves + caching planes
			DICG_CACHE
		} ALG;
		ALG alg;

		bool  conv; // if true, then x is explicitly maintained as a convex combination of atoms (with weights 'MyAtom::weight')
					// otherwise, MyAtom::weight is used as a timestamp
		double weight_eps; // with conv=true: if the weight is smaller than 'weight_eps' then set it to zero
		int   contract; // 0: no contraction
		                // 1: minimal contraction
		                // 2: neighbor-covering contraction
		double contract_average_eps; // if contract == 2 then the code will compute the average vector 'x_average' and pass it to Problem::contract().
		                             // The latter function may then choose to use 'contract_average_eps' as a threshold value.
		double contract_threshold; // Problem::contract may decide to contract only if the size of the contracted problem is smaller than the original size times 'contract_threshold'
		int	  max_cache_size; // >= 0. Max number of atoms in the cache. Used only if conv == false.
		int	  inactive_atom_iter_max; // remove atoms that haven't been used for 'inactive_atom_iter_max' oracle calls
									  // (if conv == false)
		double AFW_beta_threshold; // if (beta > AFW_beta_threshold) in the AFW method then call PFW instead of AFW,
		                           // to avoid numerical errors (AFW works with the number 1/(1-beta)).
		                           // Note, this can happen only if the FW gap is very small.

		double simplex_accuracy_before;
		double simplex_accuracy_after;
	};

	// note, LRSolver::~LRSolver() calls the destructor for all added subproblems.
	// if 'copy_mapping' is true the 'mapping' is copied. Otherwise provided pointer is used,
	// so the user should not be modify / deallocate 'mapping' until 'P' is deallocated.
	// Similarly for 'options'.
	void AddSubproblem(Problem *P, int n, int* mapping, SubproblemOptions* options, bool copy_mapping=false, bool copy_options=false);
	void AddConstant(double c) { cost_const += c; }
	void Solve();

	// For debugging: this function runs some checks to test whether contraction has been implemented correctly.
	// Can be called instead of Solve()
	void SolveTestContraction();

	// can be called after Solve(). x must be an array of size 'var_num'.
	// x[i] is set to be the average of the values of all subproblems that contain this variable.
	void ReadPrimalSolution(double* x);
	// calculates the cost of solution 'x' (by calling functions Problem::compute_x_cost()) and returns the result.
	double ComputePrimalSolutionCost(double* x);

	// this function calls Problem::set_final_dual_solution() for each subproblem
	void ReadDualSolution();

	// this function should be called from Problem::contract(). For description, see problem.h .
	// Q is the child problem of the parent problem P.
	// n: # vars in Q
	// N: size of mapping_to_parent and mapping_to_child (n <= N <= P->n)
	// Must hold:
	//   mapping_to_parent: injective mapping [0,N) => [0,P->n)
	//   mapping_to_child:  nondecreasing surjective mapping [0,N) => [0,n)
	//   mapping_to_child=NULL corresponds to [0,1,...,n-1]
	void AddContractedSubproblem(Problem* Q, int n, double x_cost, int N, int* mapping_to_parent, int* mapping_to_child=NULL);

	// read current atoms for subproblem 'S'.
	int GetAtomNum(Problem* P) { return ((Subproblem*)(P->_S))->atom_num; }
	void GetAtoms(Problem* P, Problem::Atom* atoms, double* weights=NULL) // atoms and weights must be arrays of size GetAtomNum(S)
	{
		Subproblem* S = (Subproblem*)(P->_S);
		int k;
		for (k=0; k<S->atom_num; k++)
		{
			atoms[k] = S->atoms[k].atom;
			if (weights) weights[k] = S->atoms[k].weight;
		}
	}
	SubproblemOptions* GetOptions(Problem* P) { return ((Subproblem*)(P->_S))->options; }




	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////


protected:
	friend class Problem;

	struct MyDouble
	{
		MyDouble() {}
		MyDouble(double _value) : value(_value), _is_valid(true) {}
		operator double() const { assert(_is_valid); return value; }
		void reset() { _is_valid = false; }
		bool is_valid() { return _is_valid; }

	private:
		bool _is_valid;
		double value;
	};

	struct MyAtom
	{
		double weight;
		double c; // for temporary use

		Problem::Atom atom;
		double cost;
	};

	struct SimplexOptimizer;

	struct Subproblem
	{
		void Init(LRSolver* LR, Problem* P, int n, SubproblemOptions* options);
		void Delete();

		int AddAtom(Problem::Atom& a, double cost);
		void RemoveAtom(int p, int* keep_index=NULL);
		void TouchAtom(int p);
		void PruneAtoms(int* keep_index=NULL);

		bool Contract(Problem::Atom a0 = NULL);
		void TransformContractedAtomsConv(int Tnum, Subproblem** Told, Problem::Atom a0, Subproblem* Rfirst);
		void TransformContractedAtomsCache(int Tnum, Subproblem** Told, Problem::Atom a0, Subproblem* Rfirst);

		void TransformContractedAtomsConv10(Problem::Atom a0);
		void TransformContractedAtomsCache10(Problem::Atom a0);

		void SetDotProductWeights();

		void MoveToPoint(double* y, double y_cost); // updates xi:=y, updates nu

		double FW_ORIGstep(int* s_index_ptr);
		double FWstep(int* s_index_ptr);
		double FW2step_nonconv(int* s_index_ptr);
		double FW2step_conv(int* s_index_ptr);
		double DICGstep(int* s_index_ptr);
		double CACHEstep();
		double Step(int* s_index_ptr=NULL)
		{
			if (options->alg == SubproblemOptions::ALG::FW)
			{
				if (options->max_cache_size <= 0 && !options->conv)
				{
					return FW_ORIGstep(s_index_ptr);
				}
				else if (options->simplex_accuracy_before < 1)
				{
					if (options->conv)
					{
						return FW2step_conv(s_index_ptr);
					}
					else
					{
						return FW2step_nonconv(s_index_ptr);
					}
				}
				else
				{
					return FWstep(s_index_ptr);
				}
			}
			else
			{
				return DICGstep(s_index_ptr);
			}
		}


		void Recompute_x(bool renormalize_weights=true, bool check=false); // only if conv = true. check is used only for debugging


		SubproblemOptions* options;
		int type; // 0: parent subproblem, contracted child subproblems have type 2
		          // 1: no contractions
		          // 2: child subproblem, was contracted
		int atom_num, atom_num_max;

		MyAtom* atoms; // # atom_num_max
		double get_dot_product(int i, int j) // compute values on demand
		{
			if ((i == j && _dot_products[i][j]<=-1) || (i != j && _dot_products[i][j]!=_dot_products[j][i]))
			{
				_dot_products[i][j] = _dot_products[j][i] = P->weighted_dot_product(atoms[i].atom, atoms[j].atom);
			}
			return _dot_products[i][j];
		}
		double** _dot_products;  // # atom_num_max * atom_num_max
		Problem* P;
		int n;
		double* xi;
		double xi_cost;
		double* mui;

		LRSolver* LR;

		int* mapping;
		int* mapping_to_parent;
		int* mapping_to_child;
		int N;

#define XMAPPING_FOR(S, r, k, i, code_init, code) {\
		int k; if (S->mapping_to_child == NULL)\
		{\
			for (k=0; k<S->n; k++) { int r=k; int i=S->mapping[k]; code_init; code; }\
		}\
		else\
		{\
			int r, k_prev=-1;\
			for (r=0; r<S->N; r++)\
			{\
				int i = S->mapping[r], k = S->mapping_to_child[r];\
				if (k>k_prev) { code_init; } k_prev = k;\
				code;\
			}\
		}\
	}

#define MAPPING_FOR(S, k, i, code_init, code) { XMAPPING_FOR(S, rrr, k, i, code_init, code); }

		Subproblem* parent;
		Subproblem* next; // parent points to first child, child points to next sibling
		Subproblem* list_prev;
		Subproblem* list_next;

		// for debugging
		void PrintAtoms();
		Problem::Atom DebugAddAtom(double* y);
		void DebugCheckContractedSolutions(double* solution, Subproblem* Sdebug);
		void DebugCheckDotProducts();

//	private:
	};

	struct SimplexOptimizer // data that depends on x,mu (for all subproblems). Will be recomputed each time we need to do an update for this subproblem.
	{
		SimplexOptimizer(LRSolver* LR);
		~SimplexOptimizer();
		void Alloc(int n_global_max);
		void Init(Subproblem* S, int num_max); // called in the beginning of every usage.
		                                       // num_max is the max number of atoms / points

		double* GetLambda(Subproblem* S, double* xlambda_ptr=NULL);
		double* InitAll(); // returns lambda
		void InitWeightedLambda();
		double* GetUpdatedLambda(); // used in FWstep2()

		// There are two ways to use this structure:
		// 1. If S->conv_representation == true: call MinimizeConv(S). (Minimizes over the current set of atoms).
		// 2. If S->conv_representation == false: add points one-by-one, minimize over the added points.

		///////////////////////////////////////
		// 1. S->conv_representation == true //
		///////////////////////////////////////

		void MinimizeConv(double gap_threshold, double gap_reduction_threshold, double* delta_f);
		void MoveToPointConv();

		////////////////////////////////////////
		// 2. S->conv_representation == false //
		////////////////////////////////////////
		// after calling Init(S), one may add/remove atoms from S->atoms. Once atoms in S->atoms are fixed,
		// call ResetAtomProducts(). No other functions involving atoms should be called between Init() and ResetAtomProducts().
		// 'num_max' is the maximum number of atoms (and also points) that can be added.

		// add point of the form y=c0*x + \sum_i ci*S->atoms[atom_id].atom
		void PointStart(double c0);
		void PointAddAtom(double c, int atom_id);
		void PointFinish();

		void ResetPoints();

		// once all points have been added, call Minimize. 
		double* Minimize(double gap_threshold, double gap_reduction_threshold, double* delta_f);
		void MoveToPoint(bool call_update_x=true);

		double xlambda() { return _xlambda; }
		double lambda_atom(int atom_id)
		{
			if (!_lambda_atom[atom_id].is_valid()) _lambda_atom[atom_id] = S->P->dot_product(S->atoms[atom_id].atom, lambda);
			return _lambda_atom[atom_id];
		}
		double weighted_lambda_atom(int atom_id)
		{
			if (!_weighted_lambda_atom[atom_id].is_valid()) _weighted_lambda_atom[atom_id] = S->P->dot_product(S->atoms[atom_id].atom, weighted_lambda);
			return _weighted_lambda_atom[atom_id];
		}

		double* GetBuf() { return n_max_buf; } // gives a pointer to array of size n_max

		double total_time;

	private:
		double _xlambda;
		double _weighted_xx;
		double _weighted_xlambda;
		double* lambda;          // # n_max
		double* weighted_lambda; // # n_max
		double* n_max_buf; // # n_max

		struct Point
		{
			double c0;
			int num;
			double* c; // of size num
			int* atom_id; // of size_num, with atom_id[i] \in [0,atom_num-1]
		};
		MyDouble* _lambda_atom;
		MyDouble* _weighted_lambda_atom;
		LRSolver* LR;
		Subproblem* S;
		Point* points;
		int point_num;

		double* c_buf;
		int* atom_id_buf;
		int point_num_max, buf_num, buf_num_max, atom_num_max;

		// these variables will be used for calling simplex optimization
		double** A;    // # point_num_max * point_num_max
		double* b;     // # point_num_max
		double* alpha; // # point_num_max
		double* buf;   // # point_num_max
		void* buf_atom_list; // # point_num_max. Can be converted either to Atom* or to int*

		void Dealloc();

		SimplexOpt OPT;
	} D;

	int var_num;
	int* var_counts; // of size var_num
	double cost_const;

	Buffer buffer;

	SList<Subproblem> _Slist;
	int parent_subproblem_num, child_subproblem_num;
	Subproblem* S_being_contracted;
	Subproblem** S_being_contracted_list_last;

	double GAMMA;
	int n_global, n_global_max;
	int n_global_last_solve;

	double* x_global;
	double* mu_global;
	double* nu; // of size var_num;
	double H_best;

	double T_start;

	///////////////////////////////////////////////

	void InitSolve();
	void PrimalCallback();
	void SetNu();
	double ComputeFmu();
	double ComputeH(double* gap, double* lambda_full=NULL);
		
	void ComputeFullLambda(double* lambda);
	
	double MinimizeFmu(double& gap_threshold, double H_threshold=1e300);

	double MajorLoop();
	double OracleLoop();
	double CacheLoop();

	void PrintVectors();
};


#endif
