#ifndef __SLIST_H__
#define __SLIST_H__

#include "block.h"

#define SUBPROBLEM_TYPE_NUM 3

// 'Subproblem' must be a data structure with fields
// Subproblem::type
// Subproblem::list_prev
// Subproblem::list_next
// Subproblem::parent (used for debugging)
//
// These fields should not be modified externally (except for parent).
// 'type' is an integer in [0,SUBPROBLEM_TYPE_NUM-1]
//
// SList allows the following:
//  - adding / removing subproblems
//  - iterating over all current subproblems with type \in [type_min,type_max]
//
// Interpretation of 'type':
// type=0: parent of contracted subproblems
// type=1: non-contracted subproblem
// type=2: contracted subproblem (then 'parent' points to the parent subproblem of type 0).


template <class Subproblem> class SList
{
public:

	SList();
	Subproblem* New(int type); // must be type==1 or type==2
	void ChangeType(Subproblem* S, int new_type); // only changes 0 <=> 1 are allowed
	void Remove(Subproblem* S); // only type 2 can be removed
	int GetNum(int type_min, int type_max);
	struct Iterator
	{
		Iterator(SList& Slist) : L(Slist) {}
		Subproblem* First(int type_min, int type_max, bool randomize_order=false);
		Subproblem* Next();

	private:
		SList& L;
		int t, t_min, t_max;
		Subproblem* S;
		Subproblem** S_permutation;
	};
	void PrintId(Subproblem* S0);

private:
	DBlock<Subproblem> subproblems;
	Subproblem* first[SUBPROBLEM_TYPE_NUM];
	Subproblem* last[SUBPROBLEM_TYPE_NUM];
	ReusableBuffer rbuf;
	bool rbuf_locked;
};





template <class Subproblem> inline SList<Subproblem>::SList() : subproblems(64), rbuf_locked(false)
{
	int t;
	for (t=0; t<SUBPROBLEM_TYPE_NUM; t++) first[t] = last[t] = NULL;
}

template <class Subproblem> inline Subproblem* SList<Subproblem>::New(int type)
{
	assert(type==1 || type==2);
	Subproblem* S = subproblems.New();
	S->type = type;
	if (first[type]) { S->list_prev = last[type]; last[type]->list_next = S; }
	else             { S->list_prev = NULL; first[type] = S; }
	last[type] = S;
	S->list_next = NULL;
	return S;
}

template <class Subproblem> inline void SList<Subproblem>::ChangeType(Subproblem* S, int new_type)
{
	assert(S->type==0 || S->type==1);
	assert(new_type==0 || new_type==1);
	S->type = new_type;
}

template <class Subproblem> inline void SList<Subproblem>::Remove(Subproblem* S)
{
	if (S->list_prev) S->list_prev->list_next = S->list_next;
	else              first[S->type]          = S->list_next;
	if (S->list_next) S->list_next->list_prev = S->list_prev;
	else              last[S->type]           = S->list_prev;
	subproblems.Delete(S);
}

inline int RandomInteger(int N) // returns random integer in [0,N-1]
{
	while ( 1 )
	{
		int x = (int) ( (((double)N) * rand()) / RAND_MAX );
		if (x<N) return x;
	}
}

template <class Subproblem> inline Subproblem* SList<Subproblem>::Iterator::First(int type_min, int type_max, bool randomize_order)
{
	if (!randomize_order)
	{
		S_permutation = NULL;
		t_min = t = type_min;
		t_max = type_max;
		if (t == 0) t = 1;
		S = NULL;
		return Next();
	}
	else
	{
		if (L.rbuf_locked) { printf("incorrect call to SList<Subproblem>::First()\n"); exit(1); }
		L.rbuf_locked = true;
		Iterator I(L);
		t = t_max = 0;
		S_permutation = (Subproblem**) L.rbuf.Alloc(1);
		for (S=I.First(type_min, type_max); S; S=I.Next())
		{
			S_permutation = (Subproblem**) L.rbuf.Realloc((t_max+1)*sizeof(Subproblem*));
			S_permutation[t_max ++] = S;
		}

		int i, j;
		for (i=0; i<t_max-1; i++)
		{
			j = i + RandomInteger(t_max-i);
			S = S_permutation[i]; S_permutation[i] = S_permutation[j]; S_permutation[j] = S;
		}

		return Next();
	}
}

template <class Subproblem> inline Subproblem* SList<Subproblem>::Iterator::Next()
{
	if (S_permutation == NULL)
	{
		while ( 1 )
		{
			if (S)
			{
				Subproblem* T = S;
				S = S->list_next;
				if (T->type >= t_min && T->type <= t_max) return T;
			}
			else if ( t == 1 || t <= t_max )
			{
				S = L.first[t ++];
			}
			else return NULL;
		}

		if (S->list_next) { S = S->list_next; return S; }
		return First(t+1, t_max);
	}
	else
	{
		if (t < t_max)
		{
			return S_permutation[t ++];
		}
		else
		{
			L.rbuf_locked = false;
			return NULL;
		}
	}
}


template <class Subproblem> inline int SList<Subproblem>::GetNum(int type_min, int type_max)
{
	Iterator L(*this);
	Subproblem* S;
	int num = 0;
	for (S=L.First(type_min, type_max); S; S=L.Next()) num ++;
	return num;
}

template <class Subproblem> inline void SList<Subproblem>::PrintId(Subproblem* S0)
{
	Iterator L(*this);
	Subproblem* S;
	if (S0->type < 2)
	{
		int id = 0;
		for (S=L.First(S0->type, S0->type); S; S=L.Next(), id ++) { if (S == S0) { printf("%d", id); return; } }
	}
	else
	{
		S = S0->parent;
		PrintId(S);
		Subproblem* T;
		int id = 0;
		for (T=S->next; T; T=T->next, id++) { if (T == S0) { printf("%c", 'a'+id); return; } }
	}
	exit(1);
}


#endif
