#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include "simplex_opt.h"
#include "LRSolver.h"

//#define CHECK_Fmu

double LRSolver::MinimizeFmu(double& gap_threshold, double H_threshold)
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	int major_iter, minor_iter, cache_iter;

	double Fmu;
	if (options.verbose >= 3)
	{
		Fmu = ComputeFmu();
		printf("    Fmu=%f\n", Fmu);
	}

	LRTimer timer;
	double H = 0, gap = 1e100;
	
	double time_max = options.time_max;
	bool should_terminate = false;

	for (major_iter=0; ; major_iter++)
	{
		if (major_iter >= options.major_iter_max + options.major_iter_max_log*log(1.0+major_iter)/log(2.0))
		{
			H = ComputeH(&gap);
			break;
		}
		if (major_iter > 0 && (major_iter % options.compute_gap_freq) == 0)
		{
			H = ComputeH(&gap);
			if ((gap <= gap_threshold || H > H_threshold) && major_iter >= options.major_iter_min) break;
		}
		if (get_time() - T_start > time_max)
		{
			if (should_terminate) break;
			time_max *= 1.1; should_terminate = true; // it's better not to terminate in the middle
		}

		// NOTE: from now on there will be temporary inconsistency between 'x' variables of child and parent subproblems.
		// During this inconsistency do NOT call functions such as ComputeFmu(), ComputeH(), etc.
		D.total_time = 0; // measure time spent in simplex calls
		
#ifdef CHECK_Fmu
		double Fmu_check = ComputeFmu();
#endif

		double T_major_start = timer.get_time();

		////////////////////////////////////////////////////////
		////////////////// loop with contractions //////////////
		////////////////////////////////////////////////////////
		timer.get_time_diff();
		double df_major = MajorLoop();
		double T_major = timer.get_time_diff();
		////////////////////////////////////////////////////////
		////////////////////////////////////////////////////////
		////////////////////////////////////////////////////////

		double T_minor_end = get_time() + ((options.time_minor_max > 0) ? (options.time_minor_max*T_major) : 1.1*options.time_max);
		double df_major_all = df_major;

		int oracle_loop_num = 0, cache_loop_num = 0;
		double oracle_loop_time = 0, cache_loop_time = 0;
		for (minor_iter=0; minor_iter<options.minor_iter_max; minor_iter++)
		{
			////////////////////////////////////////////////////////
			////////////////// loop with oracle calls //////////////
			////////////////////////////////////////////////////////
			timer.get_time_diff();
			double df_oracle = OracleLoop(); oracle_loop_num ++; 
			double T_oracle = timer.get_time_diff(); oracle_loop_time += T_oracle;
			if (df_oracle >= 0) break;
			////////////////////////////////////////////////////////
			////////////////////////////////////////////////////////
			////////////////////////////////////////////////////////

			double df_minor = df_oracle;
			double T_minor = T_oracle;

			double T_cache_end = get_time() + ((options.time_cache_max > 0) ? (options.time_cache_max*T_minor) : 1.1*options.time_max);

			for (cache_iter=0; cache_iter<options.cache_iter_max; cache_iter++)
			{
				////////////////////////////////////////////////////////
				////////////////// loop with only cache ////////////////
				////////////////////////////////////////////////////////
				timer.get_time_diff();
				double df_cache = CacheLoop(); cache_loop_num ++; 
				double T_cache = timer.get_time_diff(); cache_loop_time += T_cache;
				if (df_cache >= 0) break;
				
				df_minor += df_cache;
				T_minor += T_cache;
				////////////////////////////////////////////////////////
				////////////////////////////////////////////////////////
				////////////////////////////////////////////////////////

				// terminate cache loops?
				double T_current = get_time();
				if (cache_iter >= options.cache_iter_min)
				{
					if (-df_cache * T_oracle * options.ratio_cache <= -df_oracle * T_cache) break;
					if (T_current > T_cache_end) break;
				}
				if (T_current - T_start > options.time_max) break;
			}
			df_major_all += df_minor;

			// terminate minor iterations?
			double T_current = get_time();
			if (minor_iter >= options.minor_iter_min)
			{
				if (-df_minor * T_major * options.ratio_minor <= -df_major * T_minor) break;
				if (T_current > T_minor_end) break;
			}
			if (T_current - T_start > time_max)
			{
				if (should_terminate) break;
				time_max *= 1.1; should_terminate = true; // it's better not to terminate in the middle
			}
		}

		if (options.verbose >= 2)
		{
			double T = timer.get_time() - T_major_start;

			int n0_sum = 0, n1_sum = 0, n2_sum = 0, atom_num_max = 0, atom_num_sum = 0, subproblem_num = 0;
			for (S=L.First(0, 0); S; S=L.Next()) n0_sum += S->n;
			for (S=L.First(1, 1); S; S=L.Next()) n1_sum += S->n;
			for (S=L.First(2, 2); S; S=L.Next()) n2_sum += S->n;
			for (S=L.First(1, 2); S; S=L.Next())
			{
				subproblem_num ++;
				atom_num_sum += S->atom_num;
				if (atom_num_max < S->atom_num) atom_num_max = S->atom_num;
			}
			if (subproblem_num == 0) subproblem_num ++; 

			printf("major_iter %d", major_iter);
			if (n0_sum > 0) printf(" [%.2f] (n_sum:%d+%d=>%d)", T_major / T, n1_sum, n0_sum, n2_sum);
			printf(" oracle %d[%.2f]", oracle_loop_num, oracle_loop_time / T);
			printf(" cache %d[%.2f]", cache_loop_num, cache_loop_time / T);
			printf(" simplex[%.2f]", D.total_time / T);
			printf(" #atoms:%.1f(max %d)", ((double)atom_num_sum) / subproblem_num, atom_num_max);
			printf(" df=%.2e", df_major_all);
			printf("\n");
		}

		// the loop above updated 'x' variables in child subproblems, but not in parent subproblems.
		// Restore consistency between child and parent subproblems.
		for (S=L.First(0, 1); S; S=L.Next())
		{
			if (S->type == 0)
			{
				if (S->next) S->xi_cost = 0;
			}
			else
			{
				S->P->reset_contraction();
			}
		}
		for (S=L.First(2, 2); S; S=L.Next())
		{
			XMAPPING_FOR(S, r, k, i, {},\
			{\
				S->parent->xi[S->mapping_to_parent[r]] = S->xi[k];\
			})
			S->parent->xi_cost += S->xi_cost;
		}

#ifdef CHECK_Fmu
		double Fmu_check2 = ComputeFmu();
		double Fmu_diff = (Fmu_check2 - Fmu_check) - df_major_all;
		if (fabs(Fmu_diff) > 1e-5)
		{
			printf(" Fmu_diff=%f ", Fmu_diff);
			exit(1);
		}
#endif

	}
	gap_threshold = gap;
	return H;
}


double LRSolver::MajorLoop()
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	double delta_f = 0;
	for (S=L.First(0, 1, true); S; S=L.Next())
	{
		if (!S->options->contract) continue;
		if (get_time() - T_start > options.time_max) continue;
		if (S->type == 1)
		{
			int s_index;
			delta_f += S->Step(&s_index);
			S->Contract(S->atoms[s_index].atom);
		}
		else
		{
			S->Contract();
			Subproblem* R;
			for (R=S->next; R; R=R->next)
			{
				delta_f += R->Step();
			}
		}
	}

	return delta_f;
}

double LRSolver::OracleLoop()
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	double delta_f = 0;
	for (S=L.First(1, 2, true); S; S=L.Next())
	{
		if (get_time() - T_start > options.time_max) continue;
		delta_f += S->Step();
	}
	return delta_f;
}

double LRSolver::CacheLoop()
{
	SList<Subproblem>::Iterator L(_Slist);
	Subproblem* S;
	double delta_f = 0;
	for (S=L.First(1, 2, true); S; S=L.Next())
	{
		if (get_time() - T_start > options.time_max) continue;
		if (S->options->max_cache_size > 0)
		{
			delta_f += S->CACHEstep();
		}
		else
		{
			delta_f += S->Step();
		}
	}
	return delta_f;
}




inline double dot_product(double* x, double* y, int d)
{
	int i;
	double sum = 0;
	for (i=0; i<d; i++) sum += x[i]*y[i];
	return sum;
}


void LRSolver::Subproblem::MoveToPoint(double* y, double y_cost)
{
	MAPPING_FOR(this, k, i, {}, \
	{\
		LR->nu[i] += LR->GAMMA*(y[k] - xi[k]) / LR->var_counts[i];\
		xi[k] = y[k];
	})

	xi_cost = y_cost;
}

double LRSolver::Subproblem::CACHEstep()
{
	SimplexOptimizer& D = LR->D;
	D.Init(this, atom_num+1);
	double delta_f = 0;
	D.InitAll();

	if (options->conv)
	{
		D.MinimizeConv(0, options->simplex_accuracy_after, &delta_f);
		D.MoveToPointConv();
	}
	else
	{
		D.PointStart(1.); D.PointFinish(); 
		int p;
		for (p=0; p<atom_num; p++)
		{
			D.PointStart(0); D.PointAddAtom(1.0, p); D.PointFinish();
		}
		D.Minimize(0, options->simplex_accuracy_after, &delta_f);
		D.MoveToPoint();
	}

	if (LR->options.verbose >= 4)
	{
		printf("  fw "); P->print_id(); printf(": %d atoms, delta_f=%f\n", atom_num, delta_f);
	}

	PruneAtoms();
	return delta_f;
}

double LRSolver::Subproblem::FW_ORIGstep(int* s_index_ptr)
{
	SimplexOptimizer& D = LR->D;
	D.Init(this, atom_num+2);
	
	double delta_f = 0;
	double* lambda = D.InitAll();
	Problem::Atom s = P->min_oracle(lambda);

	double s_cost = P->compute_atom_cost(s);
	int s_index = AddAtom(s, s_cost); TouchAtom(s_index);
/*
	double gs = D.lambda_atom(s_index) + atoms[s_index].cost;
	double gx = D.xlambda() + xi_cost;
	double gap = gx - gs;
	if (gap <= 0)
	{
		return 0;
	}
*/
	// first point: x
	D.PointStart(1.); D.PointFinish(); 

	// second point: s
	D.PointStart(0.); D.PointAddAtom(1., s_index); D.PointFinish();

	double* gamma = D.Minimize(0, 0, &delta_f);
	D.MoveToPoint();

	if (LR->options.verbose >= 4)
	{
		printf("  fw "); P->print_id(); printf(": %d atoms, delta_f=%f\n", atom_num, delta_f);
	}

	if (s_index_ptr) *s_index_ptr = s_index;
	PruneAtoms(s_index_ptr);
	return delta_f;
}


double LRSolver::Subproblem::FW2step_conv(int* s_index_ptr)
{
	assert (options->conv);

	SimplexOptimizer& D = LR->D;
	D.Init(this, atom_num+2);
	double delta_f = 0;

	D.InitWeightedLambda();
	D.MinimizeConv(0, options->simplex_accuracy_before, &delta_f);
	D.MoveToPointConv();

	double* lambda = D.InitAll();

	Problem::Atom s = P->min_oracle(lambda);

	double s_cost = P->compute_atom_cost(s);
	int s_index = AddAtom(s, s_cost); TouchAtom(s_index);

	D.MinimizeConv(0, options->simplex_accuracy_after, &delta_f);
	D.MoveToPointConv();

	if (LR->options.verbose >= 4)
	{
		printf("  fw "); P->print_id(); printf(": %d atoms, delta_f=%f\n", atom_num, delta_f);
	}

	if (s_index_ptr) *s_index_ptr = s_index;
	PruneAtoms(s_index_ptr);
	return delta_f;
}


double LRSolver::Subproblem::FW2step_nonconv(int* s_index_ptr)
{
	assert (!options->conv);

	SimplexOptimizer& D = LR->D;
	D.Init(this, atom_num+2);
	double delta_f = 0;
	double* lambda = D.InitAll();
	bool points_added = false;

	D.PointStart(1.); D.PointFinish(); 
	int p;
	for (p=0; p<atom_num; p++)
	{
		D.PointStart(0); D.PointAddAtom(1.0, p); D.PointFinish();
	}
	D.Minimize(0, options->simplex_accuracy_before, &delta_f);
	lambda = D.GetUpdatedLambda();

	Problem::Atom s = P->min_oracle(lambda);

	double s_cost = P->compute_atom_cost(s);
	int atom_num_old = atom_num;
	int s_index = AddAtom(s, s_cost); TouchAtom(s_index);
	if (atom_num > atom_num_old)
	{
		D.PointStart(0); D.PointAddAtom(1.0, s_index); D.PointFinish();
	}

	D.Minimize(0, options->simplex_accuracy_after, &delta_f);
	D.MoveToPoint();

	if (LR->options.verbose >= 4)
	{
		printf("  fw "); P->print_id(); printf(": %d atoms, delta_f=%f\n", atom_num, delta_f);
	}

	if (s_index_ptr) *s_index_ptr = s_index;
	PruneAtoms(s_index_ptr);
	return delta_f;
}

double LRSolver::Subproblem::FWstep(int* s_index_ptr)
{
	SimplexOptimizer& D = LR->D;
	D.Init(this, atom_num+2);
	double delta_f = 0;
	double* lambda = D.InitAll();

	Problem::Atom s = P->min_oracle(lambda);

	double s_cost = P->compute_atom_cost(s);
	int s_index = AddAtom(s, s_cost); TouchAtom(s_index);

	if (options->conv)
	{
		D.MinimizeConv(0, options->simplex_accuracy_after, &delta_f);
		D.MoveToPointConv();
	}
	else
	{
		D.PointStart(1.); D.PointFinish(); 
		int p;
		for (p=0; p<atom_num; p++)
		{
			D.PointStart(0); D.PointAddAtom(1.0, p); D.PointFinish();
		}
		D.Minimize(0, options->simplex_accuracy_after, &delta_f);
		D.MoveToPoint();
	}

	if (LR->options.verbose >= 4)
	{
		printf("  fw "); P->print_id(); printf(": %d atoms, delta_f=%f\n", atom_num, delta_f);
	}

	if (s_index_ptr) *s_index_ptr = s_index;
	PruneAtoms(s_index_ptr);
	return delta_f;
}


double LRSolver::Subproblem::DICGstep(int* s_index_ptr)
{
	if (options->conv) { printf("DICGstep cannot be used with 'conv'\n"); exit(1); }

	int alg_type = -1;
	bool hit_boundary=false;
	double delta_f = 0;
	Problem::Atom v;

	SimplexOptimizer& D = LR->D;
	D.Init(this, atom_num+3);
	double* lambda = D.InitAll();
	Problem::Atom s = P->minmax_oracle(lambda, xi, &v);
	if (!v)
	{
		P->destroy_atom(s);
		return FWstep(s_index_ptr);
	}

	double s_cost = P->compute_atom_cost(s);
	double v_cost = P->compute_atom_cost(v);

	int s_index = AddAtom(s, s_cost); TouchAtom(s_index);
	int v_index = AddAtom(v, v_cost);

	double gs = D.lambda_atom(s_index) + atoms[s_index].cost;
	double gv = D.lambda_atom(v_index) + atoms[v_index].cost;
	double gx = D.xlambda() + xi_cost;
	double gap = gx - gs;
	double wgap = gv - gs;
	if (gs == gv || gap <= 0)
	{
		return 0;
	}

	double alpha, beta;
	P->max_move(xi, s, v, &alpha, &beta);

	// first point: x
	D.PointStart(1.); D.PointFinish(); 

	if ( options->alg == SubproblemOptions::ALG::DICG_CACHE )
	{
		if (beta < options->AFW_beta_threshold)
		{
			// second point: (x - beta*v) / (1-beta)
			D.PointStart(1./(1-beta)); D.PointAddAtom(-beta/(1-beta), v_index); D.PointFinish();
		}
		D.PointStart(1.); D.PointAddAtom(alpha, s_index); D.PointAddAtom(-alpha, v_index); D.PointFinish();

		int p;
		for (p=0; p<atom_num; p++)
		{
			if (p == v_index) continue;
			D.PointStart(0); D.PointAddAtom(1, p); D.PointFinish();
		}
		double* gamma = D.Minimize(0, options->simplex_accuracy_after, &delta_f);
		D.MoveToPoint();
		alg_type = 0;
	}
	else if ( options->alg == SubproblemOptions::ALG::DICG_PFW || beta >= options->AFW_beta_threshold)
	{
		// second point: x + alpha*s - alpha*v
		D.PointStart(1.); D.PointAddAtom(alpha, s_index); D.PointAddAtom(-alpha, v_index); D.PointFinish();
		double* gamma = D.Minimize(0, 0, &delta_f);
		D.MoveToPoint();
		alg_type = 1; if (gamma[1] == 1) hit_boundary = true;
	}
	else if ( options->alg == SubproblemOptions::ALG::DICG_AFW )
	{
		if (gap >= gv - gx || beta >= 1)
		{
			// second point: s
			D.PointStart(0.); D.PointAddAtom(1., s_index); D.PointFinish();

			double* gamma = D.Minimize(0, 0, &delta_f);
			D.MoveToPoint();
			alg_type = 2;
		}
		else
		{
			// second point: (x - beta*v) / (1-beta)
			D.PointStart(1./(1-beta)); D.PointAddAtom(-beta/(1-beta), v_index); D.PointFinish();

			double* gamma = D.Minimize(0, 0, &delta_f);
			if (gamma[1] == 1)
			{
				D.MoveToPoint(false);
				P->update_x_AFW(xi, beta, v);
				hit_boundary = true;
			}
			else
			{
				D.MoveToPoint();
			}
			alg_type = 3;
		}
	}

	if (LR->options.verbose >= 4)
	{
		if (alg_type == 0) printf("  dicg_cache");
		else if (alg_type == 1) printf("  dicg_pfw");
		else if (alg_type == 2) printf("  dicg_afw+");
		else if (alg_type == 3) printf("  dicg_afw-");
		if (hit_boundary) printf("*");
		printf(" "); P->print_id(); printf(": wgap=%.2e, %d atoms, ", wgap, atom_num);
		printf("delta_f=%.2e, ", delta_f);
		if (alpha==beta) printf("a=b=%.3e\n", alpha);
		else             printf("(a,b)=(%.3e,%.3e)\n", alpha, beta);
	}

	if (s_index_ptr) *s_index_ptr = s_index;
	PruneAtoms(s_index_ptr);
	return delta_f;
}





