/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2021

    This file is part of LRSolver.

    LRSolver is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with LRSolver.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Class Problem implements function f(x) of n variables over some discrete set X \subset R^n.
(Note, the description in Solver.h uses notation fi and ni, but here for brevity we use f and n).
Elements of X are called ``atoms'', and the convex hull of X (which is a polytope) is denoted as P.

The user must implement functions listed below. One of the main functions is min_oracle(g),
which must compute atom a\in X that minimizes <g,x> + f(x). Usually, atoms can be stored
very compactly. The type Problem::Atom (defined as 'void *') corresponds to atoms stored
in a compact form. The user must allocate the memory for atoms; LRSolver will eventually call
Problem::destroy_atom(a) for every atom 'a' returned by min_oracle() on minmax_oracle().

When adding new problem to LRSolver via LRSolver::AddSubproblem(), the user must specify
some options (LRSolver::SubproblemOptions). The main options are

	contract
	conv
	alg

They affect which of the functions below will be called.
In particular, the flag 'contract' has a major effect on the behaviour of LRSolver.
If it is 1 or 2, then LRSolver calls P->contract() on subproblem P from time to time.
This call can 'fix' certain variables and then add new subproblem(s) Q_1, ..., Q_m (for m>=0) 
obtained after fixing the variables. The original subproblem is called ``parent subproblem'',
and new subproblems are called ''child subproblems''.

*/


#ifndef __PROBLEM_H__
#define __PROBLEM_H__

#include <math.h>
#include <assert.h>


class LRSolver;


class Problem
{
public:
	Problem() : LR(NULL), dot_product_weights(NULL) {}
	virtual ~Problem() {}

	typedef double* Vector; // of size n
	typedef void* Atom;

	////////////////////////////////////////////////////
	// 1. Basic functions that need to be implemented //
	////////////////////////////////////////////////////
	virtual void atom_to_vector(Atom a, Vector x) = 0; // should convert 'a' to its vector representation (and write the result to x)
	virtual void add_atom_to_vector(double gamma, Atom a, Vector x) = 0; // update x += gamma*a
	virtual double dot_product(Atom a, Vector g) = 0; // return < a , x > 
	virtual double weighted_dot_product(Atom a, Atom b) = 0; // return \sum_{i=0}^n dot_product_weights[i]*a[i]*b[i] 
	                                                         // Note, array Problem::dot_product_weights will be set by LRSolver
	virtual double compute_atom_cost(Atom a) = 0; // return f(a)
	virtual int compare_atoms(Atom a, Atom b) = 0; // return -1,0,1 according to some total order on atoms
	virtual Atom min_oracle(Vector g) = 0; // Output: newly allocated atom 's' that minimizes < s, g >
	virtual void destroy_atom(Atom a) = 0;  // the code will eventually call this function for all atoms returned by 'min_oracle' or 'maxmin_oracle'

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// 2. Functions that need to be implemented if alg == DICG*** (Decomposition-Invariant Conditional Gradient)         //
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// This function must compute atom v \in P_x that maximizes <g,v>+f(v),
	// where P_x is the minimal face of P containing x.
	// Due to rounding errors it may not be possible to compute such v; then it is allowed to set v=NULL
	// (in which case LRSolver performs standard FW step).
	//
	// The function must also compute atom s that minimizes <g,s>+f(s), set *_v=v, and return s.
	virtual Atom minmax_oracle(Vector g, Vector x, Atom* _v) { printf("minmax_oracle() not implemented\n"); exit(1); }

	// The function below must do the following:
	// 1. compute maximum alpha s.t. x + alpha*(s-v) \in P, set *_alpha = alpha.
	//                      There is a guarantee that s != v.
	// 2. compute supremum of beta<=1 s.t. (x - beta*v)/(1 - beta) \in P, set *_alpha = beta.
	//                      More precisely, we must have x - beta*v = (1-beta)*y for some y\in P,
	//                      or equivalently x = y + beta*(v-y).
	//                      If x == v then set beta=1.
	//
	// Note, if polytope is defined by constraints Ax=b, Cx\ge 0 then it suffices to check the following inequality constraints:
	//  C*(x + alpha*(s-v)) >= 0
	//  C*(x - beta*v) >= 0
	virtual void max_move(Vector x, Atom s, Atom v, double* _alpha, double* _beta) { printf("max_move() not implemented\n"); exit(1); }

	/////////////////////////////////////////////////////////////////////////
	// 3. Functions that need to be implemented if contract == 1 or 2 //
	/////////////////////////////////////////////////////////////////////////

	/*
	* Note, for functions above we only needed to know costs a_cost of atoms a (computed via a_cost = compute_atom_cost(a)).
	* For functions below we also need to know costs of fractional vectors x \in P.
	* If x is represented as a convex combination of atoms ( x = \sum_i c_i a_i ), then the cost of x
	* is defined as x_cost = \sum_i c_i compute_atom_cost(a_i).
	* 
	* == Example ==
	* 
	* Suppose that problem P depends on 5 variables (P0,...,P4), the current vector is x = [0.2  0.2  0  0.5  0.3]
	* and contract() decides to fix variable P2 to the current value (i.e. P2=0), merge variables P0 and P1 
	* (i.e. enforce P0=P1), and create two new subproblems A, B as follows:
	* 
	*     P0   P1   P2   P3   P4
	*    0.2  0.2   0   0.5   0.3
	*     A0   A0  fix   B0   A1
	* 
	* Thus, A depends on two variables A0,A1 and B depends on one variable B0.
	* Note that A0 corresponds to both P0 and P1 (which is possible since these variables have the same value).
	* Of course, problems A, B must be independent (i.e. the appropriate Cartesian product of A and B must be a subset of P).
	* 
	* Function contract(x, x_cost, a) must do the following:
	* 1. Decide which variables to fix / merge based on 'x' and 'a' (see below).
	* 2a. Allocate problem A, call
	*    LR->AddContractedSubproblem(A, 2, xA_cost, 3, mapping_to_parentA, mapping_to_childA)
	*    where 
	*         2 = # variables in A
	*         xA_cost = cost of vector [0.2 0.3] in A
	*         3 = size of mapping_to_parentA and mapping_to_childA
	*         mapping_to_parentA = [0 1 4]  (corresponds to P0,P1,P4)
	*         mapping_to_childA  = [0 0 1]  (corresponds to A0,A0,A1). This mapping must be monotonic!!!
	* 2b. Allocate problem B, call
	*    LR->AddContractedSubproblem(B, 1, xB_cost, 1, mapping_to_parentB, mapping_to_childB)
	*    where
	*         1 = # variables in B
	*         xB_cost = cost of vector [0.5] in B
	*         1 = size of mapping_to_parentB and mapping_to_childB
	*         mapping_to_parentB = [3]  (corresponds to P3)
	*         mapping_to_childB  = [0]  (corresponds to B0). It is also possible to use mapping_to_child = NULL, this corresponds to array [0 1 ... nB-1]
	* 
	* Note that xA_cost and xB_cost must sum to x_cost.
	* (Exception: if conv == true then LRSolver ignores xA_cost,xB_cost, since it itself can recompute these values).
	* If P->contract() adds just a single child subproblem Q, then it is possible just to copy x_cost.
	* Otherwise the user may need to store some additional information about current 'x' 
	* to correctly decide how to split x_cost between child subproblems.
	* 
	* The user is responsible for allocating / deallocating arrays 'mapping_to_parent' and 'mapping_to_child'.
	* These arrays should not be modified until the corresponding child subproblem is deallocated.
	* 
	********************************************************************
	* WHICH VARIABLES CAN BE FIXED / MERGED BY contract(x, x_cost, a)? *
	********************************************************************
	* Fixing / merging variables means restricting the set of atoms in X to some smaller subset X' \subseteq X.
	* Two conditions must be satisfied:
	* (a) atom a must belong to X'.
	* (b) vector x must belong to conv(X').
	*/

	////////////////////////////////////////////////////////////////////////
	// Input: current vector x, its cost x_cost, atom a (it was obtained via min_oracle() just before calling contract()).
	// If SubproblemOptions::contract == 1 then x_average==NULL.
	// If SubproblemOptions::contract == 2 then x_average is an array of the same size as x obtained by averaging over all subproblems.
	// The user may use this array to NOT fix variables that are active in other subproblems.
	// Note, x_average may contain rounding errors, so it's better to use criterion x_average[i] > eps
	// instead of x_average[i] > 0 for some eps > 0.
	//
	// The user may choose not to contract, then false should be returned.
	virtual bool contract(Vector x, double x_cost, Atom a, Vector x_average) { printf("contract() not implemented\n"); exit(1); }

	// the user must maintain atom 'a_current' for problem P, and implement the following functions:

	// if set_current_atom(a) is called for a parent problem P:
	//     it must update P->a_curent = a.
	// if set_current_atom(a) is called for a child problem Q of P:
	//     it must update the part of P->a_current corresponding to Q to 'a'.
	virtual void set_current_atom(Atom a) { printf("set_current_atom() not implemented\n"); exit(1); }
	// if called for parent subproblem: must return a_current.
	// if called for child subproblem Q: must return the atom corresponding to the restriction of P->a_current to Q.
	//                                  It may happen that this restriction does not belong to Q, then it should return NULL
	virtual Atom get_current_atom() { printf("get_current_atom() not implemented\n"); exit(1); }


	// the user may choose to use the same structure for both contracted and non-contracted problems (if both use the same 'mapping').
	// This function is called to inform the user that 'contract' is no longer active.
	virtual void reset_contraction() {}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// 4. Functions that need to be implemented if the user wants to maintain an internal representation of current vector x. //
	// This may be needed e.g. to implement max_move() or contract().                                                           //
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// 'x' is always initialized with an atom
	virtual void init_x(Atom a) { return; }

	// LRSolver will call the function below every time when x is changed.

	// corresponds to the following update:
	//    x := c0*x + \sum_i c[i]*atoms[i]
	virtual void update_x(Vector x_new, double c0, int atom_num, double* c, Atom* atoms) { return; }

	// special update of the form
	//    x := (x - beta*v) / (1 - beta)
	// where beta is the value returned by the preceding call to max_move() (for atoms s,v).
	// Such update is treated separately in order to avoid rounding errors
	// (to make sure that some components of 'x' are set to zero rather than to something like 1e-10).
	virtual void update_x_AFW(Vector x_new, double beta, Atom v) { return; }


	///////////////////////////////////////////////////////////////////////////////////////////////
	// 6. Functions that need to be implemented if the user calls LRSolver::ReadPrimalSolution() //
	///////////////////////////////////////////////////////////////////////////////////////////////
	virtual double compute_x_cost(Vector x) { printf("compute_x_cost() not implemented\n"); exit(1); }

	///////////////////////////////////////////////////////////////////////////////////////////////
	// 7'. Functions that need to be implemented if GlobalOptions::adjustx_freq > 0
	///////////////////////////////////////////////////////////////////////////////////////////////
	virtual double adjust_x(Vector x) { printf("adjust_x() not implemented\n"); exit(1); }

	// Almost done. The last function 
	virtual void set_final_dual_solution(Vector g) {} // called at the very end
	virtual void print_atom(Atom a) { printf("print_atom() not implemented\n"); exit(1); } // used for debugging




	// these variables will be set by LRSolver::AddSubproblem() or LRSolver::AddContractedSubproblem()
	LRSolver* LR;
	void* _S; // can be cast as ((LRSolver::Subproblem*)_S). (Unfortunately, forward declaration of nested classes not supported by C++...)
	double* dot_product_weights;
	int parent_subproblem_id;
	int child_subproblem_id;
	void print_id()
	{
		printf("%d", parent_subproblem_id);
		if (child_subproblem_id >= 0) printf("%c", 'a'+child_subproblem_id);
	}


};

#endif

