#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "simplex_opt.h"




void SimplexOpt::BestSolution::init(SimplexOpt* _OPT, double* x_init)
{
	OPT = _OPT;
	n = OPT->n;
	int c;

	double* ptr = (double*) rbuf.Alloc((SIMPLEX_OPT_CACHE_NUM+2)*n*sizeof(double));
	x_best  = ptr; ptr += n;
	Ax_best = ptr; ptr += n;
	for (c=0; c<SIMPLEX_OPT_CACHE_NUM; c++)
	{
		x_cache[c] = ptr; ptr += n;
	}
	cache_num = 0;

	memcpy(x_best, x_init, n*sizeof(double));
	cost_init = OPT->evaluate(x_init, &gap_init, Ax_best);
	cost_best = cost_init; gap_best = gap_init;
}

bool SimplexOpt::BestSolution::add(double* x)
{
	if (cache_num > SIMPLEX_OPT_CACHE_NUM)
	{
		printf("ERROR cache_num=%d\n", cache_num);
	}
	if (cache_num < SIMPLEX_OPT_CACHE_NUM)
	{
		memcpy(x_cache[cache_num ++], x, n*sizeof(double));
		return true;
	}

	double* Ax = OPT->local_buf;
	double cost, gap;
	cost = OPT->evaluate(x, &gap, Ax);
	if (cost_best > cost)
	{
		memcpy(x_best, x, n*sizeof(double));
		memcpy(Ax_best, Ax, n*sizeof(double));
		cost_best = cost; gap_best = gap;
		cache_num = 0;
		return true;
	}
	else
	{
		get_best(NULL);
		return false;
	}
}

void SimplexOpt::BestSolution::get_best(double* x, double** _Ax_best)
{
	double cost, gap;
	double* Ax = OPT->local_buf;

	for ( ; cache_num>0; cache_num--)
	{
		cost = OPT->evaluate(x_cache[cache_num-1], &gap, Ax);
		if (cost_best > cost)
		{
			memcpy(x_best, x_cache[cache_num-1], n*sizeof(double));
			memcpy(Ax_best, Ax, n*sizeof(double));
			cost_best = cost; gap_best = gap;
			cache_num = 0;
			break;
		}
	}
	if (x) memcpy(x, x_best, n*sizeof(double));
	if (_Ax_best) *_Ax_best = Ax_best;
}




// find delta such that
// \sum_{i:x[i]>0)(c[i]-delta) + \sum_{i:x[i]=0} \min { c[i]-delta, 0 } = 0

inline void project_grad_to_simplex(int n, double* c, double* x, double* cc) // cc: buffer of size n
{
	int k;

	double sum = 0;
	int m = 0;
	for (k=0; k<n; k++)
	{
		if (x[k] <= 0) cc[m ++] = c[k];
		else sum += c[k];
	}

	std::sort(cc, cc+m);
	for (k=n-m; k<n; k++)
	{
		double delta = cc[k-(n-m)];
		if (sum - k*delta <= 0) break;
		sum += cc[k-(n-m)];
	}

	double delta = sum / k;

	for (k=0; k<n; k++)
	{
		c[k] -= delta;
		if (x[k] <= 0 && c[k] > 0) c[k] = 0;
	}

	/*
	/// check
	sum = 0;
	for (k=0; k<n; k++) sum += c[k];
	if (fabs(sum) > 1e-10)
	{
	printf("Error: sum=%f\n", sum);
	exit(1);
	}
	*/
}


SimplexOpt::SimplexOpt()
{
}

SimplexOpt::~SimplexOpt()
{
}

void SimplexOpt::setup(double** _A, double* _b, int _n)
{
	A = _A;
	b = _b;
	n = _n;

	double* ptr = (double*) rbuf.Alloc(5*n*sizeof(double) + 2*n*sizeof(int));
	local_buf  = ptr; ptr += n;
	buf = ptr; ptr += 4*n;
	buf_int = (int*)(ptr);
}

double SimplexOpt::evaluate(double* x, double* gap, double* Ax_best)
{
	double* Ax = (Ax_best) ? Ax_best : local_buf;
	compute_Ax(x, Ax);

	int i;
	double cost = 0, g_min = 1e100, g_max = -1e100;
	for (i=0; i<n; i++)
	{
		cost += (0.5*Ax[i] + b[i])*x[i];
		double g = Ax[i] + b[i];
		if (g_min > g) g_min = g;
		if (g_max < g && x[i] > 0) g_max = g;
	}
	if (gap) *gap = g_max - g_min;
	return cost;
}


bool SimplexOpt::max_move(double* x, double* d, double alpha)
{
	int i;
	for (i=0; i<n; i++)
	{
		if (x[i] + alpha*d[i] < 0) break;
	}
	if (i == n)
	{
		for (i=0; i<n; i++)
		{
			x[i] += alpha*d[i];
		}
		return true;
	}

	//////////////

	int i_min = -1;
	double eta = 0;
	for (i=0; i<n; i++)
	{
		if (d[i] < 0 && (i_min<0 || eta > - x[i] / d[i]))
		{
			i_min = i;
			eta = - x[i] / d[i];
		}
	}
	if (i_min < 0) return false;
	for (i=0; i<n; i++)
	{
		x[i] += eta*d[i];
		if (x[i] < 0) x[i] = 0;
	}
	x[i_min] = 0;
	return false;
}


bool SimplexOpt::grad_step()
{
	int i, j;
	double* x = buf;
	double* c = buf + n;
	double* Ax;
	best_solution.get_best(x, &Ax);
	for (i=0; i<n; i++)
	{
		c[i] = b[i] + Ax[i];
	}

	project_grad_to_simplex(n, c, x, local_buf);

	j = -1;
	double eta = 0;
	for (i=0; i<n; i++)
	{
		if (c[i] > 0 && (j<0 || eta > x[i] / c[i]))
		{
			j = i;
			eta = x[i] / c[i];
		}
	}

	if (j < 0)
	{
		return false;
	}

	double c_sum = 0;
	for (i=0; i<n; i++)
	{
		if (i == j) c[i] = 0;
		else
		{
			c[i] = x[i] - eta*c[i];
			if (c[i] < 0) c[i] = 0;
			c_sum += c[i];
		}
	}
	if (c_sum <= 0)
	{
		return false;
	}
	for (i=0; i<n; i++) c[i] /= c_sum;

	// find optimal point on [x,c]
	// first, set  c  :=  c-x
	for (i=0; i<n; i++) c[i] -= x[i]; // x' = x + gamma*c

	double AA = 0, BB = 0, gamma;
	// objective: 0.5*AA*gamma*gamma + BB*gamma + const
	// AA = <c, A*c>, BB = <x, A*c> + <b, c>
	for (i=0; i<n; i++)
	{
		double Ac = 0;
		for (j=0; j<n; j++) Ac += A[i][j]*c[j];
		AA += c[i]*Ac;
		BB += x[i]*Ac + b[i]*c[i];
	}
	double delta_f;
	if (AA<=0)
	{
		gamma = (BB >= 0) ? 0 : 1;
		delta_f = BB*gamma;
	}
	else
	{
		gamma = -BB/AA;
		if (gamma < 0) gamma = 0;
		if (gamma > 1) gamma = 1;
		delta_f = (0.5*AA*gamma+BB)*gamma;
	}

	////////////////////////////
	if (gamma < 1) j = -1;
	double x_sum = 0;
	for (i=0; i<n; i++)
	{
		x[i] += gamma*c[i];
		if (x[i] <= 0 || i == j) x[i] = 0;
		x_sum += x[i];
	}

	//if (x_sum<1-1e-6 || x_sum>1+1e-6) { printf("error in simplex_step: x_sum = %f\n", x_sum); exit(1); }
	if (x_sum > 0)
	{
		for (i=0; i<n; i++) x[i] /= x_sum;
	}
	else
	{
		for (i=0; i<n; i++) x[i] = 1.0 / n;
	}

	best_solution.add(x);
	return true;
}










void SimplexOpt::solve_conjgradient(double gap_threshold, bool stop_at_boundary)
{
	int* face = buf_int;
	int* face_prev = buf_int + n;
	double* x  = buf;
	double* Ap = buf + n;
	double* r  = buf + 2*n;
	double* p  = buf + 3*n;
	int i;

	int iter_total = 0;
	bool hit_boundary = true;
	while ( 1 )
	{
		// init conj. gradient
		double* Ax;
		best_solution.get_best(x, &Ax);
		if (best_solution.get_gap() <= gap_threshold) break;
		for (i=0; i<n; i++)
		{
			r[i] = b[i] + Ax[i];
		}
		project_grad_to_simplex(n, r, x, local_buf);
		int face_size = 0;
		for (i=0; i<n; i++)
		{
			r[i] = -r[i];
			if (x[i] > 0 || r[i] > 0) { face[i] = 1; face_size ++; }
			else face[i] = 0;
		}
		if (face_size == 0) break;

		if (!hit_boundary)
		{
			for (i=0; i<n; i++) { if (face_prev[i] != face[i]) break; }
			if (i == n) break;
		}
		hit_boundary = false;
		for (i=0; i<n; i++) face_prev[i] = face[i];
		
		double rr = 0;
		for (i=0; i<n; i++)
		{
			p[i] = r[i];
			rr += r[i]*r[i];
		}

		// conj. gradient iterations
		int iter;
		for (iter=0, iter_total++; iter<face_size-1 && iter<options.iter_inner_max; iter++, iter_total++)
		{
			compute_Ax(p, Ap);

			double pAp = 0;
			for (i=0; i<n; i++)
			{
				Ap[i] *= face[i];
				pAp += p[i]*Ap[i];
			}
			if (pAp <= 0)
			{
				break;
			}

			double alpha = rr / pAp;

			bool result = max_move(x, p, alpha);
			double x_sum = 0;
			for (i=0; i<n; i++)
			{
				if (x[i] < 0) x[i] = 0;
				x_sum += x[i];
			}
			if (x_sum > 0)
			{
				for (i=0; i<n; i++) x[i] /= x_sum;
			}
			else
			{
				for (i=0; i<n; i++) x[i] = 1.0/n;
			}

			bool x_added = best_solution.add(x);
			if (!result)
			{
				hit_boundary = true;
				break;
			}
			if (!x_added || best_solution.get_gap() <= gap_threshold)
			{
				break;
			}

			for (i=0; i<n; i++) r[i] -= alpha*Ap[i];

			double r_sum = 0;
			for (i=0; i<n; i++) r_sum += r[i];
			r_sum /= face_size;
			for (i=0; i<n; i++)
			{
				if (face[i]) r[i] -= r_sum;
			}

			double rr_new = 0;
			for (i=0; i<n; i++) rr_new += r[i]*r[i];

			double beta = (rr > 0) ? (rr_new / rr) : 0;
			rr = rr_new;

			for (i=0; i<n; i++) p[i] = r[i] + beta*p[i];
		}

		if (stop_at_boundary) break;
		if (iter_total > options.iter_total_max) break;
	}
}



double SimplexOpt::solve(double* x, double gap_threshold, double gap_reduction_threshold)
{
	options.init(n);
	best_solution.init(this, x);
	if (n <= 2)
	{
		grad_step();
	}
	else
	{
		double gap = gap_reduction_threshold * best_solution.get_gap_init();
		if (gap_threshold < gap) gap_threshold = gap;

		solve_conjgradient(gap_threshold, false);
		grad_step();
	}
	best_solution.get_best(x);
	return best_solution.get_delta_f();
}

double SimplexOpt::solve_inner(double* x, double gap_threshold, double gap_reduction_threshold)
{
	options.init(n);
	best_solution.init(this, x);
	double gap = gap_reduction_threshold * best_solution.get_gap_init();
	if (gap_threshold < gap) gap_threshold = gap;
	solve_conjgradient(gap_threshold, true);
	grad_step();
	best_solution.get_best(x);
	return best_solution.get_delta_f();
}


























void SimplexOpt::print()
{
	int p, q;
	printf("A=\n");
	for (p=0; p<n; p++)
	{
		printf("    ");
		for (q=0; q<n; q++)
		{
			printf("%f ", A[p][q]);
		}
		printf("\n");
	}
	printf("b=[ ");
	for (p=0; p<n; p++) printf("%f ", b[p]);
	printf("]\n");
}






///////////////////////////////
// #define TEST_SIMPLEX_OPT

#ifdef TEST_SIMPLEX_OPT




void test_simplex_step()
{
	int p, q, r, n = 100;

	double** P = new double*[n];
	P[0] = new double[n*n];
	for (p=1; p<n; p++) P[p] = P[p-1] + n;

	double** A = new double*[n];
	A[0] = new double[n*n + 4*n];
	for (p=1; p<n; p++) A[p] = A[p-1] + n;
	double* b = A[0] + n*n;
	double* x = b + n;
	double* buf = x + n;

	memset(A[0], 0, n*n*sizeof(double));

	for (p=0; p<n; p++)
	for (q=0; q<n; q++)
	{
		P[p][q] = rand()%10;
//		P[p][q] = (p == q) ? 1 : 0;
	}
	for (p=0; p<n; p++)
	for (q=0; q<n; q++)
	{
		A[p][q] = 0;
		for (r=0; r<n; r++) A[p][q] += P[p][r]*P[q][r];
	}

	double* c = NULL;
	if ( 1 )
	{
		double* c = new double[n];
		double c_sum = 0;
		for (p=0; p<n; p++)
		{
			c[p] = 1 + (rand() % 10);
			c_sum += c[p];
		}
		for (p=0; p<n; p++) c[p] /= c_sum;
		for (p=0; p<n; p++)
		{
			b[p] = 0;
			for (q=0; q<n; q++) b[p] -= A[p][q]*c[q];
		}
		printf("c=[ "); for (p=0; p<n; p++) printf("%f ", c[p]); printf("]\n");
	}
	else
	{
		double z = 0.1 * rand() / (RAND_MAX + 1.0);
		for (p=0; p<n; p++) b[p] = z*((rand() / (RAND_MAX + 1.0)) - 0.5);
	}

	for (p=0; p<n; p++) x[p] = 0;
	x[0] = 1;

	SimplexOpt OPT;
	OPT.setup(A, b, n);
	OPT.print();

	//double f0 = OPT.evaluate(c);

	OPT.solve(x);
	//	OPT.solve(x, 0, false);

	delete [] A[0];
	delete [] A;

	delete [] P[0];
	delete [] P;
	if (c) delete [] c;
}

int main()
{
	int seed;
	for (seed=80; ; seed++)
	{
		printf("========== seed = %d ==========\n", seed);
		srand(seed);
		test_simplex_step();
//		break;
		getchar();
	}
}

#endif
