#ifndef __SIMPLEX_OPT__
#define __SIMPLEX_OPT__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include "block.h"

class SimplexOpt
{
public:
	SimplexOpt();
	~SimplexOpt();

	struct Options
	{
		void init(int n)
		{
			iter_inner_max = n; // equivalent to infinity, this limit will never be reached
			iter_total_max = 3*n;
			if (iter_total_max > 100) iter_total_max = 100;
		}

		int iter_inner_max;
		int iter_total_max;
	} options;

	void setup(double** A, double* b, int n);

	// stop if one the following occurs:
	//  - 'gap' becomes smaller than 'gap_threshold'
	//  - 'gap' becomes smaller than 'gap_reduction_threshold*initial_gap'
	//  - if a boundary is hit (in the case of solve_inner()).
	// returns delta_f.

	double solve(double* x, double gap_threshold=0, double gap_reduction_threshold=0);
	double solve_inner(double* x, double gap_threshold=0, double gap_reduction_threshold=0);

	double evaluate(double* x, double* gap=NULL, double* Ax_best=NULL);

	void print();


private:
	double** A; // provided externally
	double* b;  // provided externally
	int n;
	ReusableBuffer rbuf;

	double* local_buf; // of size n, can be used only by lowest-level functions
	double* buf; // of size 4*n
	int* buf_int; // of size 2*n

	struct BestSolution
	{
		void init(SimplexOpt* OPT, double* x_init);
		bool add(double* x); // computes Ax, returns true if the solution improved
		void get_best(double* x, double** Ax_best=NULL);
		double get_delta_f() { return cost_best - cost_init; }
		double get_gap() { return gap_best; }
		double get_gap_init() { return gap_init; }

	private:
		SimplexOpt* OPT;
		int n;
		double* x_best;
		double* Ax_best;
		double cost_best, cost_init;
		double gap_best, gap_init;

#define SIMPLEX_OPT_CACHE_NUM 5
		int cache_num;
		double* x_cache[SIMPLEX_OPT_CACHE_NUM];

		ReusableBuffer rbuf;
	} best_solution;

	void compute_Ax(double* x, double* Ax)
	{
		int i, j;
		for (i=0; i<n; i++)
		{
			double _Ax = 0;
			for (j=0; j<n; j++) _Ax += A[i][j]*x[j];
			Ax[i] = _Ax;
		}
	}


	bool grad_step();    // returns true if x is modified
	void solve_conjgradient(double gap_threshold, bool stop_at_boundary);


	// if x + alpha*d \in P:
	//     update x := x + alpha*d, return true
	// otherwise:
	//     move by the maximum possible distance along d, return false.
	bool max_move(double* x, double* d, double alpha);
};


#endif
