/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2021

    This file is part of RECURSIVE-FW.

    POTTS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __UTILS_H__
#define __UTILS_H__

#include "block.h"

#define FW_INFTY ((double)1e300)

inline double dot_product(double* x, double* y, int d)
{
	int i;
	double sum = 0;
	for (i=0; i<d; i++) sum += x[i]*y[i];
	return sum;
}

inline double dot_product_diff(double* x, double* y, double* z, int d)
{
	int i;
	double sum = 0;
	for (i=0; i<d; i++) sum += (x[i]-y[i])*z[i];
	return sum;
}

inline double norm_squared(double* x, int d)
{
	int i;
	double sum = 0;
	for (i=0; i<d; i++) sum += x[i]*x[i];
	return sum;
}

inline void interpolate(double* x, double* y, double gamma, double* z, int d)
{
	int i;
	for (i=0; i<d; i++) z[i] = (1-gamma)*x[i] + gamma*y[i];
}

#define EVENT_TYPE_NUM (EventType::FLUSH+1)

class Logger
{
public:
	Logger(int _iter_max, double _time_max, double _gap_min, const char* _filename=NULL)
		: iter_max(_iter_max), time_max(_time_max), gap_min(_gap_min), filename(_filename),
		  ptr(NULL), entries(1024), depth(0), min_oracle_num(0), terminate_depth(DEPTH_MAX+1), writing_to_file(false)
	{
		time_start = time_this_level_start = get_time();
		memset(flags, 0, EVENT_TYPE_NUM*sizeof(int));
		memset(times, 0, DEPTH_MAX*sizeof(double));
	}
	~Logger()
	{
		flush();
		if (depth != 0) { printf("error in Logger::~Logger(): depth!=0\n"); exit(1); }
		FILE* fp = stdout;
		if (filename)
		{
			writing_to_file = true;
			fp = fopen(filename, "w");
			if (!fp) return;
			while ( PrintRow(fp) ) {}
		}
		fprintf(fp, "level times: ");
		for (depth=0; depth<DEPTH_MAX; depth++)
		{
			if (times[depth] == 0) break;
			fprintf(fp, "%f ", times[depth]);
		}
		fprintf(fp, "\n");
		if (filename) fclose(fp);
	}



	void log_min_oracle() { Log(EventType::MIN_ORACLE); }
	void log_min_max_oracle() { Log(EventType::MIN_ORACLE); }
	void log_line_search(double gamma) { LogDouble(EventType::LINE_SEARCH, gamma); }
	void log_dicg_alpha(double alpha) { LogDouble(EventType::DICG_ALPHA, alpha); }
	void log_f(double f) { LogDouble(EventType::F, f); }
	void log_gap(double gap) { LogDouble(EventType::GAP, gap); }
	void log_w(double w) { LogDouble(EventType::W, w); }
	void log_wminus(double wminus) { LogDouble(EventType::WMINUS, wminus); }
	void log_Phi(double Phi) { LogDouble(EventType::PHI, Phi); }
	void log_align(double align) { LogDouble(EventType::ALIGN, align); }
	void log_atom_num(int atom_num) { LogInt(EventType::ATOM_NUM, atom_num); }
	void log_str(const char* s) { LogStr(EventType::STRING, s); }
	void log_shrink(int n_shrinked);
	void log_unshrink();
	void start_recursive(int n) { sizes[0] = n; }
	void flush();

	bool keep_going() { return (depth < terminate_depth); }
	void set_terminate() { terminate_depth = -1; }
	void set_terminate_current_level() { terminate_depth = depth; }

private:
	typedef enum {
		MIN_ORACLE=0,
		DICG_ALPHA=1, LINE_SEARCH=2, F=3, GAP=4, W=5, WMINUS=6, PHI=7, ALIGN=8, // double
		ATOM_NUM=9, // int
		STRING=10, // const char*

				   // for internal use
		FLUSH=11,
		SHRINK=12, UNSHRINK=13, 
	} EventType;

	typedef union { double x; int n; const char* s; } Value;
	struct Entry
	{
		EventType type;
		Value v;
	};
	int flags[EVENT_TYPE_NUM];
	Block<Entry> entries;
	Entry* ptr;

	static const int DEPTH_MAX = 100;
	int sizes[DEPTH_MAX];
	double times[DEPTH_MAX];
	int depth, min_oracle_num;
	double time_start, time_this_level_start;

	int iter_max;
	double time_max;
	double gap_min;
	const char* filename;
	bool writing_to_file;

	int terminate_depth; // terminate runs at depth>=terminate_depth
	//bool terminate_this_level, terminate_all_levels;

	void Log(EventType type);
	void LogInt(EventType type, int n);
	void LogDouble(EventType type, double x);
	void LogStr(EventType type, const char* str);
	bool PrintRow(FILE* fp);
};

inline void Logger::Log(EventType type)
{
	if (flags[type]) flush();
	flags[type] = 1;
	Entry* e = entries.New();
	e->type = type;
	if (type == EventType::MIN_ORACLE)
	{
		e->v.n = min_oracle_num ++;
		if (min_oracle_num >= iter_max) terminate_depth = -1;
	}
}

inline void Logger::LogInt(EventType type, int n)
{
	if (flags[type]) flush();
	flags[type] = 1;
	Entry* e = entries.New();
	e->type = type; e->v.n = n;
}

inline void Logger::LogDouble(EventType type, double x)
{
	if (flags[type]) flush();
	flags[type] = 1;
	Entry* e = entries.New();
	e->type = type; e->v.x = x;
	if (type == EventType::GAP && x<=gap_min)
	{
		terminate_depth = depth;
	}
}

inline void Logger::LogStr(EventType type, const char* s)
{
	if (flags[type]) flush();
	flags[type] = 1;
	Entry* e = entries.New();
	e->type = type; e->v.s = s;
}

inline void Logger::flush()
{
	Entry* e = entries.New();
	e->type = EventType::FLUSH;
	e->v.x = get_time() - time_start;
	if (e->v.x >= time_max) terminate_depth = -1;
	memset(flags, 0, EVENT_TYPE_NUM*sizeof(int));
	if (!filename) PrintRow(stdout);
}

inline void Logger::log_shrink(int n_shrinked)
{
	if (depth >= DEPTH_MAX-1) { printf("Error: max_depth exceeded in Logger\n"); exit(1); }
	double T = get_time(); times[depth] += T - time_this_level_start; time_this_level_start = T;
	flush();
	//Log(EventType::MIN_ORACLE);
	Entry* e = entries.New();
	e->type = EventType::SHRINK;
	e->v.n = n_shrinked;
	sizes[++ depth] = n_shrinked;
	flush();
}

inline void Logger::log_unshrink()
{
	double T = get_time(); times[depth] += T - time_this_level_start; time_this_level_start = T;
	flush();
	Entry* e = entries.New();
	e->type = EventType::UNSHRINK;
	if (terminate_depth == depth) terminate_depth = DEPTH_MAX+1;
	depth --;
	flush();
}

inline bool Logger::PrintRow(FILE* fp)
{
	int i;
	if (!ptr) ptr = entries.ScanFirst();
	else ptr = entries.ScanNext();
	if (!ptr) return false;
	for (i=0; i<depth; i++) fprintf(fp, "*");
	for ( ; ptr->type != EventType::FLUSH; ptr = entries.ScanNext() )
	{
		switch (ptr->type)
		{
			case EventType::MIN_ORACLE: fprintf(fp, "oracle=%d ", ptr->v.n); break;
			case EventType::LINE_SEARCH: fprintf(fp, "ls_gamma=%.3e ", ptr->v.x); break;
			case EventType::DICG_ALPHA: fprintf(fp, "dicg_alpha=%.3e ", ptr->v.x); break;
			case EventType::F: fprintf(fp, "f=%.10f ", ptr->v.x); break;
			case EventType::GAP: fprintf(fp, "gap=%.3e ", ptr->v.x); break;
			case EventType::PHI: fprintf(fp, "Phi=%.3e ", ptr->v.x); break;
			case EventType::W: fprintf(fp, "w=%.3e ", ptr->v.x); break;
			case EventType::WMINUS: fprintf(fp, "wminus=%e ", ptr->v.x); break;
			case EventType::ALIGN: fprintf(fp, "align=%.3e ", ptr->v.x); break;
			case EventType::ATOM_NUM: fprintf(fp, "|S|=%d ", ptr->v.n); break;
			case EventType::STRING: fprintf(fp, "%s ", ptr->v.s); break;
			case EventType::SHRINK: 
				if (writing_to_file) sizes[++ depth] = ptr->v.n;
				fprintf(fp, "shrink [ ");
				for (i=0; i<=depth; i++) fprintf(fp, "%d ", sizes[i]);
				fprintf(fp, "] ");
				break;
			case EventType::UNSHRINK:
				if (writing_to_file) depth --;
				fprintf(fp, "unshrink [ ");
				for (i=0; i<=depth; i++) fprintf(fp, "%d ", sizes[i]);
				fprintf(fp, "] ");
				break;
			default: break;
		}
	}
	fprintf(fp, "time=%f\n", ptr->v.x);
	return true;
}

#endif
