#define _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "MRF.h"
#include "../LRSolver/LRSolver.h"
#include "TreeMRF.h"
#include "arboricity/arboricity.h"

MRF::MRF(int _node_num, int* _K, int _edge_num_max)
	: node_num(_node_num), edge_num(0), edge_num_max(_edge_num_max), message_buf(NULL), buffer(1024), node_permutation(NULL), singleton_unary_costs(NULL), fractional_solution(NULL), LR(NULL)
{
	if (node_num <= 0) return;

	int i;
	nodes = new Node[node_num];
	edges = new Edge[edge_num_max];

	int K_sum = 0;
	K_max = 0;
	for (i=0; i<node_num; i++)
	{
		nodes[i].solution_best = -1;
		nodes[i].K = _K[i];
		K_sum += _K[i];
		if (K_max < _K[i]) K_max = _K[i];
		nodes[i].first[0] = nodes[i].first[1] = NULL;
	}
	cost_best = 1e300;
	double* ptr = new double[K_sum];
	memset(ptr, 0, K_sum*sizeof(double));
	for (i=0; i<node_num; i++)
	{
		nodes[i].D = ptr; ptr += nodes[i].K;
	}
	fractional_solution = new double[K_sum];
	memset(fractional_solution, 0, K_sum*sizeof(double));
}

MRF::~MRF()
{
	if (node_num <= 0) return;

	delete [] nodes[0].D;
	delete [] nodes;
	delete [] edges;
	
	if (message_buf) delete [] message_buf;
	if (node_permutation) delete [] node_permutation;
	if (singleton_unary_costs) delete [] singleton_unary_costs;
	if (fractional_solution) delete [] fractional_solution;
}

void MRF::InitMessages()
{
	if (message_buf) return;
	
	int num = 0;
	Edge* e;
	for (e=edges; e<edges+edge_num; e++) num += e->i[0]->K + e->i[1]->K;
	
	double* ptr = message_buf = new double[num];
	memset(message_buf, 0, num*sizeof(double)); 
	for (e=edges; e<edges+edge_num; e++)
	{
		e->message[0] = ptr; ptr += e->i[0]->K;
		e->message[1] = ptr; ptr += e->i[1]->K;
	}
}

void MRF::AddUnary(int i, double* D)
{
	int a;
	for (a=0; a<nodes[i].K; a++) nodes[i].D[a] += D[a];
}

void MRF::MoveAllFromUnaries()
{
	Node* i;
	Edge* e;
	for (i=nodes; i<nodes+node_num; i++)
	{
		int dir, a, degree = 0;
		int K = i->K;
		double* M = (double*) rbuffer.Alloc(K*sizeof(double));
		memset(M, 0, K*sizeof(double));
		for (dir=0; dir<2; dir++)
		for (e=i->first[dir]; e; e=e->next[dir])
		{
			degree ++;
			for (a=0; a<K; a++) M[a] += e->message[dir][a];
		}
		for (a=0; a<K; a++) M[a] = (M[a] + i->D[a]) / degree;
		for (dir=0; dir<2; dir++)
		for (e=i->first[dir]; e; e=e->next[dir])
		{
			for (a=0; a<K; a++) e->message[dir][a] -= M[a];
		}
	}
}

void MRF::AddEdge(int _i, int _j, PairwiseTerm* T)
{
	if (_i == _j || _i<0 || _j<0 || _i>=node_num || _j>=node_num) { printf("Incorrect call to AddEdge()\n"); exit(1); }
	if (edge_num >= edge_num_max) { printf("Too many edges\n"); exit(1); }
	Edge* e = &edges[edge_num ++];
	Node* i[2] = { nodes + _i, nodes + _j };

	int dir;
	for (dir=0; dir<2; dir++)
	{
		e->i[dir] = i[dir];

		e->prev[dir] = NULL;
		if (i[dir]->first[dir])
		{
			i[dir]->first[dir]->prev[dir] = e;
		}
		e->next[dir] = i[dir]->first[dir];
		i[dir]->first[dir] = e;
	}

	e->T = T;
}

void MRF::Edge::Swap(ReusableBuffer* rbuf)
{
	double* mtmp = message[0]; message[0] = message[1]; message[1] = mtmp;
	T->swap(i[0]->K, i[1]->K, rbuf);
	int dir;
	for (dir=0; dir<2; dir++)
	{
		if (next[dir]) next[dir]->prev[dir] = prev[dir];
		if (prev[dir]) prev[dir]->next[dir] = next[dir];
		else i[dir]->first[dir] = next[dir];
	}
	Node* tmp = i[0]; i[0] = i[1]; i[1] = tmp;
	for (dir=0; dir<2; dir++)
	{
		if (i[dir]->first[dir])
		{
			prev[dir] = i[dir]->first[dir];
			i[dir]->first[dir]->prev[dir] = this;
		}
		else prev[dir] = NULL;
		next[dir] = i[dir]->first[dir];
		i[dir]->first[dir] = this;
	}
}

double MRF::ComputeCost(int* x)
{
	Edge* e;
	double cost = 0;

	if (x)
	{
		int i, j;

		for (i=0; i<node_num; i++)
		{
			cost += nodes[i].D[x[i]];
		}

		for (e=edges; e<edges+edge_num; e++)
		{
			i = (int)(e->i[0] - nodes);
			j = (int)(e->i[1] - nodes);
			cost += e->T->get(x[i], x[j], e->i[0]->K, e->i[1]->K);
		}

		if (cost_best > cost && nodes[i].solution_best < 0)
		{
			for (i=0; i<node_num; i++) nodes[i].solution_best = x[i];
			cost_best = cost;
		}
	}
	else
	{
		Node* i;

		for (i=nodes; i<nodes+node_num; i++)
		{
			cost += i->D[i->solution];
		}

		for (e=edges; e<edges+edge_num; e++)
		{
			cost += e->T->get(e->i[0]->solution, e->i[1]->solution, e->i[0]->K, e->i[1]->K);
		}

		if (cost_best > cost && nodes[0].solution_best < 0)
		{
			for (i=nodes; i<nodes+node_num; i++) i->solution_best = i->solution;
			cost_best = cost;
		}
	}

	return cost;
}

double MRF::GetBestSolution(int* x)
{
	int i;
	for (i=0; i<node_num; i++) x[i] = nodes[i].solution_best;
	return cost_best;
}

/*
bool MRF::LR_callback_fn(double* x, void* arg)
{
	MRF* mrf = (MRF*) arg;
	Node* i;

	for (i=mrf->nodes; i<mrf->nodes+mrf->node_num; x+=i->K, i++)
	{
		int a;
		double sum = 0;
		for (a=0; a<i->K; a++)
		{
			if (x[a] < 0) x[a] = 0;
			sum += x[a];
		}
		if (sum <= 0)
		{
			for (a=0; a<i->K; a++) x[a] = 1.0 / i->K;
		}
		else
		{
			for (a=0; a<i->K; a++) x[a] /= sum;
		}
	}

	return true;
}
*/

void MRF::RunLR(LROptions& LRoptions)
{
	RemoveSingletons();
	if (node_num == 0)
	{
		RestoreSingletons();
		return;
	}

	InitMessages();
	MoveAllFromUnaries();

	int i, index;
	Edge* e;

	TreeMRF::Options tree_mrf_options;
	tree_mrf_options.final_callback_fn = Edge::message_update;
	tree_mrf_options.update_x = (LRoptions.subproblem_options.contract && !LRoptions.subproblem_options.conv);
	tree_mrf_options.adjustx_eps = LRoptions.subproblem_options.weight_eps;
	LRoptions.global_options.callback_fn = _LR_primal_callback_fn;
	LRoptions.global_options.callback_arg = this;

	int* node_indices = new int[2*node_num];
	int* tree_nodes = node_indices+node_num;
	Buffer mapping_buf(1);
	ReusableBuffer rbuf_swap;
	ReusableBuffer rbuf_tree_edges;

	for (i=index=0; i<node_num; i++) 
	{
		node_indices[i] = index;
		index += nodes[i].K;
	}

	int var_num = index;
	LR = new LRSolver(var_num);

	GraphDecomposition G(node_num, edge_num);
	for (e=edges; e<edges+edge_num; e++)
	{
		G.AddEdge((int)(e->i[0]-nodes), (int)(e->i[1]-nodes));
	}
	if (LRoptions.global_options.verbose) { printf("decomposing graph into trees... "); fflush(stdout); }
	GraphDecomposition::Method method = GraphDecomposition::Method::NONE;
	if (LRoptions.decomposition_method == 1 || LRoptions.decomposition_method == 2) method = GraphDecomposition::Method::GREEDY;
	if (LRoptions.decomposition_method == 3 || LRoptions.decomposition_method == 4) method = GraphDecomposition::Method::OPT;
	G.Decompose(method);
	if ((LRoptions.decomposition_method & 1) == 0) G.BreakLargeTrees();
	if (LRoptions.global_options.verbose)
	{
		printf("done\n"); fflush(stdout);
		G.PrintStats(LRoptions.global_options.verbose > 2);
	}
	while ( true )
	{
		int r, i, j, e;
		int tree_size;
		r = G.GetNextRoot(tree_size);
		if (r < 0) break;

		int* tree_edges_list = (int*) rbuf_tree_edges.Alloc((tree_size-1)*sizeof(int));
		int c = 0;
		TreeMRF* P = new TreeMRF(tree_size, &tree_mrf_options, &mapping_buf, &rbuffer);
		tree_nodes[r] = P->AddRoot(nodes[r].K, node_indices[r]);
		while ( (e=G.GetNextEdge(i, j)) >= 0 )
		{
			tree_edges_list[c ++] = e;
			if (edges[e].i[0] == nodes+i) edges[e].Swap(&rbuf_swap);
			tree_nodes[j] = P->AddChild(nodes[j].K, node_indices[j], tree_nodes[i], edges[e].T, &edges[e]);
		}
		int* mapping = P->FinalizeTree();
		for ( ; c>0; tree_edges_list++, c--)
		{
			e = *tree_edges_list;
			i = tree_nodes[(int) (edges[e].i[0] - nodes) ];
			j = tree_nodes[(int) (edges[e].i[1] - nodes) ];
			P->SubtractUnary(i, edges[e].message[0]);
			P->SubtractUnary(j, edges[e].message[1]);
		}
		LR->AddSubproblem(P, P->n, mapping, &LRoptions.subproblem_options);
	}

	delete [] node_indices;

	memcpy(&LR->options, &LRoptions.global_options, sizeof(LRSolver::GlobalOptions));
	
	//LR->SolveTestContraction();

	LR->Solve();
	
	LR_ComputePrimalCost();

	LR->ReadDualSolution();

	delete LR;
	LR = NULL;

	RestoreSingletons();
}

void MRF::LR_ComputePrimalCost()
{
	int i, k, index;
	LR->ReadPrimalSolution(fractional_solution);
	double sum_min = 1, sum_max = 1;
	for (i=index=0; i<node_num; i++) 
	{
		double sum = 0;
		nodes[i].solution = 0;
		for (k=0; k<nodes[i].K; k++)
		{
			if (fractional_solution[index + k] < 0) fractional_solution[index + k] = 0;
			sum += fractional_solution[index + k];
			if (fractional_solution[index + nodes[i].solution] < fractional_solution[index + k]) nodes[i].solution = k;
		}
		if (sum_min > sum) sum_min = sum;
		if (sum_max < sum) sum_max = sum;

		if (sum <= 0)
		{
			for (k = 0; k < nodes[i].K; k++)
			{
				fractional_solution[index + k] = 1.0 / nodes[i].K;
			}
		}
		else
		{
			for (k = 0; k < nodes[i].K; k++)
			{
				fractional_solution[index + k] /= sum;
			}
		}
		index += nodes[i].K;
	}

	if (sum_min < 1-1e-6 || sum_max > 1+1e-6)
	{
		printf("normalization=%f:%f (corrected) ", sum_min, sum_max);
	}

	double primal_cost = LR->ComputePrimalSolutionCost(fractional_solution);
	double primal_cost_int = ComputeCost(NULL);
	printf("primal cost=%f [integer primal cost=%f]\n", primal_cost, primal_cost_int);
}



void MRF::Print()
{
	int k;
	Node* i;
	Edge* e;
	for (i=nodes; i<nodes+node_num; i++)
	{
		printf("node %d: ", (int)(i-nodes));
		for (k=0; k<i->K; k++) printf(" %f", i->D[k]);
		printf("\n");
	}
	for (e=edges; e<edges+edge_num; e++)
	{
		printf("edge (%d,%d):\n", (int)(e->i[0]-nodes), (int)(e->i[1]-nodes));
		e->T->print(e->i[0]->K, e->i[1]->K);
	}
}


void MRF::PrintStats()
{
	Node* i;
	Edge* e;
	int K_min = nodes[0].K, K_max = nodes[0].K;
	for (i=nodes; i<nodes+node_num; i++)
	{
		if (K_min > i->K) K_min = i->K;
		if (K_max < i->K) K_max = i->K;
	}
	int n_general = 0, n_potts = 0;
	for (e=edges; e<edges+edge_num; e++)
	{
		if      (e->T->type == PairwiseTerm::Type::GENERAL) n_general ++;
		else if (e->T->type == PairwiseTerm::Type::POTTS) n_potts ++;
	}
	printf("%d nodes, %d edges (%d general, %d potts), #labels: [%d,%d]\n", node_num, edge_num, n_general, n_potts, K_min, K_max);
}














/*
MRF* generate_test(Buffer* buf, int seed)
{
	srand(seed);
	rand(); rand();

	int i, j, e, k;
	const int node_num = 2;
	int* K = (int*) buf->Alloc(node_num*sizeof(int));
	for (i=0; i<node_num; i++) K[i] = 2;// + (rand()%2);
	MRF* mrf = new MRF(node_num, K, 2*node_num);
	for (i=0; i<node_num-1; i++)
	{
		j = i+1;
		for (e=0; e<2; e++)
		{
			PairwiseTermGeneral* T = AllocPairwiseTermGeneral(K[i], K[j], NULL, buf);
			for (k=0; k<K[i]*K[j]; k++) T->V[k] = (double)(rand() % 10);
			mrf->AddEdge(i, j, T);
		}
	}

	return mrf;
}
*/



MRF::MRF(const char* filename, Buffer* buf)
	: edge_num(0), message_buf(NULL), buffer(1024)
{
	FILE* fp = fopen(filename, "r");
	if (!fp) { printf("Cannot open %s\n", filename); exit(1); }
	if (fscanf(fp, "%d %d\n", &node_num, &edge_num_max) != 2 || node_num < 1 || edge_num_max < 0) { printf("Incorrect file format %s\n", filename); exit(1); }

	int* _K = new int[node_num];
	int i, e;
	for (i=0; i<node_num; i++)
	{
		if (fscanf(fp, "%d", &_K[i]) != 1 || _K[i]<1)  { printf("Incorrect file format %s\n", filename); exit(1); }
	}

	nodes = new Node[node_num];
	edges = new Edge[edge_num_max];

	int K_sum = 0;
	K_max = 0;
	for (i=0; i<node_num; i++)
	{
		nodes[i].solution_best = -1;
		nodes[i].K = _K[i];
		K_sum += _K[i];
		if (K_max < _K[i]) K_max = _K[i];
		nodes[i].first[0] = nodes[i].first[1] = NULL;
	}
	cost_best = 1e300;
	double* ptr = new double[K_sum];
	memset(ptr, 0, K_sum*sizeof(double));
	for (i=0; i<node_num; i++)
	{
		nodes[i].D = ptr; ptr += nodes[i].K;
	}
	
	for (i=0; i<node_num; i++)
	{
		int _i, k;
		if (fscanf(fp, "%d:", &_i) != 1 || i != _i) { printf("Incorrect file format %s\n", filename); exit(1); }
		for (k=0; k<_K[i]; k++)
		{
			if (fscanf(fp, "%lf", &nodes[i].D[k]) != 1) { printf("Incorrect file format %s\n", filename); exit(1); }
		}
	}
	for (e=0; e<edge_num_max; e++)
	{
		int _i, _j, a, b;
		if (fscanf(fp, "%d %d:", &_i, &_j) != 2 || _i < 0 || _j < 0 || _i >= node_num || _j >= node_num || _i == _j ) { printf("Incorrect file format %s\n", filename); exit(1); }
		int A = _K[_i], B = _K[_j];
		PairwiseTermGeneral* T = AllocPairwiseTermGeneral(A, B, NULL, buf);
		double* V = T->V;
		for (a=0; a<A; a++)
		for (b=0; b<B; b++)
		{
			if (fscanf(fp, "%lf", V ++) != 1) { printf("Incorrect file format %s\n", filename); exit(1); }
		}
		AddEdge(_i, _j, T);
	}
	
	delete [] _K;
	fclose(fp);

	fractional_solution = new double[K_sum];
	memset(fractional_solution, 0, K_sum*sizeof(double));
}

void MRF::Save(const char* filename)
{
	int dir, k;
	Node* i;
	Edge* e;

	FILE* fp = fopen(filename, "w");
	fprintf(fp, "%d %d\n", node_num, edge_num);
	for (i=nodes; i<nodes+node_num; i++)
	{
		fprintf(fp, "%d ", i->K);
	}
	fprintf(fp, "\n");
	for (i=nodes; i<nodes+node_num; i++)
	{
		double* D = new double[i->K];
		memcpy(D, i->D, i->K*sizeof(double));
		for (dir=0; dir<2; dir++)
		for (e=i->first[dir]; e; e=e->next[dir])
		{
			for (k=0; k<i->K; k++) D[k] += e->message[dir][k];
		}
		fprintf(fp, "%d:", (int)(i-nodes));
		for (k=0; k<i->K; k++) fprintf(fp, " %f", D[k]);
		fprintf(fp, "\n");
		delete [] D;
	}
	for (e=edges; e<edges+edge_num; e++)
	{
		fprintf(fp, "%d %d:", (int)(e->i[0]-nodes), (int)(e->i[1]-nodes));
		int a, b;
		for (a=0; a<e->i[0]->K; a++)
		for (b=0; b<e->i[1]->K; b++)
		{
			fprintf(fp, " %f", e->T->get(a, b, e->i[0]->K, e->i[1]->K) - e->message[0][a] - e->message[1][b]);
		}
		fprintf(fp, "\n");
	}
	fclose(fp);
}


void MRF::RemoveSingletons()
{
	if (node_permutation) { printf("error in Permute()\n"); exit(1); }

	int k;
	Node* i;
	int nonsingleton_num = 0;
	for (i=nodes; i<nodes+node_num; i++)
	{
		if (i->first[0] || i->first[1]) nonsingleton_num ++;
	}

	if (nonsingleton_num == node_num) return;

	node_permutation = new int[node_num];
	int i0 = 0, i1 = nonsingleton_num;
	for (i=nodes; i<nodes+node_num; i++)
	{
		if (i->first[0] || i->first[1]) node_permutation[(int)(i-nodes)] = i0 ++;
		else                            node_permutation[(int)(i-nodes)] = i1 ++;
	}

	Node* nodes_old = nodes;
	nodes = new Node[node_num];
	for (i=nodes_old; i<nodes_old+node_num; i++)
	{
		memcpy(nodes+node_permutation[(int)(i-nodes_old)], i, sizeof(Node));
	}
	Edge* e;
	for (e=edges; e<edges+edge_num; e++)
	{
		e->i[0] = nodes + node_permutation[(int)(e->i[0]-nodes_old)];
		e->i[1] = nodes + node_permutation[(int)(e->i[1]-nodes_old)];
	}
	delete [] nodes_old;

	node_num_orig = node_num;
	node_num = nonsingleton_num;

	double singleton_cost = 0;
	for (i=nodes+node_num; i<nodes+node_num_orig; i++)
	{
		i->solution_best = 0;
		for (k=1; k<i->K; k++) { if (i->D[i->solution_best] > i->D[k]) i->solution_best = k; }
		singleton_cost += i->D[i->solution_best];
	}
	if (node_num > 0)
	{
		i = nodes;
		singleton_unary_costs = new double[i->K];
		memcpy(singleton_unary_costs, i->D, i->K*sizeof(double));
		for (k=0; k<i->K; k++) i->D[k] += singleton_cost;
	}
}

void MRF::RestoreSingletons()
{
	if (!node_permutation) return;

	Node* i;
	int k;

	if (node_num > 0)
	{
		i = nodes;
		memcpy(i->D, singleton_unary_costs, i->K*sizeof(double));
		delete [] singleton_unary_costs;
		singleton_unary_costs = NULL;
	}

	int* node_permutation_rev = new int[node_num_orig];
	for (k=0; k<node_num_orig; k++) node_permutation_rev[node_permutation[k]] = k;
	delete [] node_permutation;
	node_permutation = node_permutation_rev;

	Node* nodes_old = nodes;
	node_num = node_num_orig;
	nodes = new Node[node_num];
	for (i=nodes_old; i<nodes_old+node_num; i++)
	{
		memcpy(nodes+node_permutation[(int)(i-nodes_old)], i, sizeof(Node));
	}
	Edge* e;
	for (e=edges; e<edges+edge_num; e++)
	{
		e->i[0] = nodes + node_permutation[(int)(e->i[0]-nodes_old)];
		e->i[1] = nodes + node_permutation[(int)(e->i[1]-nodes_old)];
	}
	delete [] nodes_old;

	delete [] node_permutation;
	node_permutation = NULL;
}
