#ifndef __MRF_H__
#define __MRF_H__

#include "../LRSolver/LRSolver.h"
#include "../LRSolver/block.h"
#include "pairwise_term.h"
#include "TreeMRF.h"
#include "graph_decomposition.h"

///////////////////////////////////////////////////////////////


class MRF
{
public:

	// K must be an array of size 'node_num', where K[i]>=1 is the number of labels for node i
	MRF(int node_num, int* K, int edge_num_max);
	~MRF();

	// D is added to internally storied unaries
	void AddUnary(int i, double* D);

	// the user is responsible for allocating / deallocating T.
	// Note: the code may call T->swap()
	void AddEdge(int i, int j, PairwiseTerm* T);

	// can be called after calling one of the algorithms
	double GetBestSolution(int* x); // solution is written to array 'x' (of size node_num), its cost is returned
	double ComputeCost(int* x); // x - array of size node_num. Returns cost of labeling x

	// can be called after RunLR(). Returns pointer to array of size nodes[0].K + nodes[1].K + ... + nodes[node_num-1].K.
	double* GetFractionalSolution() { return fractional_solution; }

	///////////////////
	//   ALGORITHMS  //
	///////////////////

	/////////////////////////////////
	struct SRMPOptions
	{
		SRMPOptions() : iter_max(0), time_max(100000000), eps(1e-8), compute_solution_freq(5), verbose(1) {}

		int iter_max;
		double time_max;
		double eps; // stop if the increase of the lower during one iteration is less than 'eps'
		int compute_solution_freq; // extract solution after every 'compute_solution_freq' iterations
		int verbose;
	};
	void RunSRMP(SRMPOptions& SRMPoptions);
	/////////////////////////////////

	/////////////////////////////////
	struct LROptions
	{
		LROptions() : decomposition_method(0) {}

		LRSolver::GlobalOptions global_options;
		LRSolver::SubproblemOptions subproblem_options;
		int decomposition_method; // 0: first try grid, then OPT + break large trees
		                          // 1: greedy
		                          // 2: greedy + break large trees
		                          // 3: OPT (find decomposition that minimizes # forests - can be slow!)
		                          // 4: OPT + break large trees
	};
	void RunLR(LROptions& LRoptions);
	/////////////////////////////////




	int GetNodeNum() { return node_num; }
	int GetEdgeNum() { return edge_num; }
	int GetLabelNum(int i) { return nodes[i].K; }
	double* GetUnary(int i) { return nodes[i].D; }
	PairwiseTerm* GetEdge(int _e, int& i, int& j) { Edge* e = edges+_e; i = (int)(e->i[0]-nodes); j = (int)(e->i[1]-nodes); return e->T; }
	
	void Print();
	void PrintStats();
	void PrintAllMessages();

	


	void Save(const char* filename);
	MRF(const char* filename, Buffer* buf); // load




private:
	struct Node;
	struct Edge
	{
		Node* i[2];
		Edge* prev[2];
		Edge* next[2];

		PairwiseTerm* T;
		double* message[2]; // message[0] : message to i[0], message[1]: message to i[1]

		void Swap(ReusableBuffer* rbuf);

		static void message_update(int dir, double* m, void* arg)
		{
			Edge* e = (Edge*) arg;
			int k;
			for (k=0; k<e->i[dir]->K; k++) e->message[dir][k] += m[k];
		}
	};
	struct Node
	{
		int K; // # labels
		double* D;
		Edge* first[2];

		int solution;
		int solution_best;
	};

	Node* nodes;
	Edge* edges;
	int node_num;
	int edge_num, edge_num_max, K_max;
	Buffer buffer;
	ReusableBuffer rbuffer;

	double* fractional_solution;

	double* message_buf;
	void InitMessages();
	
	void MoveAllFromUnaries();
	double ComputeLowerBound(); // from messages

	//static bool LR_callback_fn(double* x, void* arg);

	static void _LR_primal_callback_fn(void* arg) { ((MRF*)arg)->LR_ComputePrimalCost(); }
	void LR_ComputePrimalCost();
	LRSolver* LR;

	int node_num_orig;
	int* node_permutation;
	double* singleton_unary_costs;
	void RemoveSingletons();
	void RestoreSingletons();
		
	double cost_best;
	friend GraphDecomposition;
};



#endif
