#ifndef __TREE_MRF_H__
#define __TREE_MRF_H__

#include <stdio.h>
#include <stdlib.h>
#include "../LRSolver/problem.h"
#include "pairwise_term.h"


class TreeMRF : public Problem
{
public:
	struct Options
	{
		Options() : final_callback_fn(NULL), update_x(false), adjustx_eps(0) {}

		bool update_x;
		double adjustx_eps;
		void (*final_callback_fn)(int dir, double* messages, void* edge_arg);
	};

	TreeMRF(int node_num_max, Options* tree_mrf_options, Buffer* mapping_buf, ReusableBuffer* rbuf);
	~TreeMRF();

	int AddRoot(int K, int global_index);
	int AddChild(int K, int global_index, int parent, PairwiseTerm* T, void* edge_arg);
	int* FinalizeTree();
	void SubtractUnary(int i, double* D); // can be called only after FinalizeTree()







	void atom_to_vector(Atom _a, Vector x) override;
	void add_atom_to_vector(double gamma, Atom _a, Vector x) override;
	double dot_product(Atom _a, Vector g) override;
	double weighted_dot_product(Atom _a, Atom _b) override;
	int compare_atoms(Atom _a, Atom _b) override;
	Atom min_oracle(Vector g) override;
	double compute_atom_cost(Atom _a) override;
	void destroy_atom(Atom a) override;

	void init_x(Atom atom) override;
	void update_x(Vector x_new, double c0, int atom_num, double* c, Atom* atoms) override;
	void update_x_AFW(Vector x_new, double beta, Atom v) override;

	bool contract(Vector x, double x_cost, Atom a, Vector x_average) override;

	void set_current_atom(Atom a) override;
	Atom get_current_atom() override;

	double compute_x_cost(Vector x) override;
	double adjust_x(Vector x) override;

	void set_final_dual_solution(Vector g) override;
	void print_atom(Atom a) override;


	struct Node;
	struct Edge;

	struct Node
	{
		int K, global_index, D_index, back_ptr_index;
		Node* parent;

		bool is_original_T;
		PairwiseTerm* T;
		Node* i_child; // used only for parent subproblems. Corresponding node in the child subproblem (or NULL)
		double x_cost;

		int current_solution;
		void* edge_arg; // used only for parent subproblems
	};

	Node* nodes;
	int n, node_num, node_num_max, back_ptr_size;
	double* unary_costs;
	double const_term;
	Options* options;
	Buffer* mapping_buf; // used for 'mapping' for parent subproblems
	ReusableBuffer* rbuf; // used by min_oracle, minmax_oracle, contract
	int* mapping_to_parent; // != NULL for contracted subproblems

	int* index; // of size node_num+1 (for parent subproblems)
	int* rev_mapping; // of size <= n (for parent subproblems)
	char* pairwise_terms_buf; // for child subproblems, stored contracted pairwise terms

	Atom a_prev; // solution at previous contraction

	TreeMRF* parent_tree;
};

inline TreeMRF::TreeMRF(int _node_num_max, Options* _options, Buffer* _buf, ReusableBuffer* _rbuf)
	: options(_options), n(0), node_num(0), node_num_max(_node_num_max), mapping_buf(_buf), rbuf(_rbuf), 
	mapping_to_parent(NULL), index(NULL), rev_mapping(NULL), pairwise_terms_buf(NULL),
	parent_tree(NULL), const_term(0), unary_costs(NULL), a_prev(NULL)
{
	nodes = new Node[node_num_max];
}

inline TreeMRF::~TreeMRF()
{
	delete [] nodes;
	if (mapping_to_parent) delete [] mapping_to_parent;
	if (index) delete [] index;
	if (rev_mapping) delete [] rev_mapping;
	if (pairwise_terms_buf) free(pairwise_terms_buf);
	if (unary_costs) delete [] unary_costs;
	if (a_prev) destroy_atom(a_prev);
}

inline int TreeMRF::AddRoot(int K, int global_index)
{
	if (node_num != 0 || node_num>=node_num_max) { printf("incorrect call to AddRoot()\n"); exit(1); }
	nodes[0].K = K;
	nodes[0].global_index = global_index;
	nodes[0].parent = NULL;
	return node_num ++;
}

inline int TreeMRF::AddChild(int K, int global_index, int parent, PairwiseTerm* T, void* _edge_arg)
{
	if (parent<0 || parent>=node_num || node_num>=node_num_max) { printf("incorrect call to AddChild()\n"); exit(1); }
	Node* i = nodes + node_num;
	i->K = K;
	i->global_index = global_index;
	i->parent = nodes + parent;
	i->T = T;
	i->edge_arg = _edge_arg;

	return node_num ++;
}

#define REVERSE_TREE_MRF_NODE(j) { if (j) j = nodes + (node_num-1-(int)((j)-nodes_old)); }

inline int* TreeMRF::FinalizeTree()
{
	Node* nodes_old = nodes;
	nodes = new Node[node_num];

	int _i, k, m;
	Node* i;
	for (_i=0, i=nodes; _i<node_num; _i++, i++)
	{
		memcpy(i, &nodes_old[node_num-_i-1], sizeof(Node));
		REVERSE_TREE_MRF_NODE(i->parent);
		i->is_original_T = true;
		i->i_child = NULL;
		i->x_cost = 0;
	}
	delete [] nodes_old;

	// set n, Node::D_index, Node::back_ptr_index, mapping
	for (i=nodes, m=0; i<nodes+node_num-1; i++) { i->back_ptr_index = m; m += i->parent->K; }
	back_ptr_size = m;
	for (i=nodes, n=0; i<nodes+node_num; i++) { i->D_index = n; n += i->K; }
	int* mapping = (int*) mapping_buf->Alloc(n*sizeof(int));
	for (i=nodes; i<nodes+node_num; i++)
	{
		for (k=0; k<i->K; k++) mapping[i->D_index + k] = i->global_index + k;
	}

	index = new int[node_num+1];
	rev_mapping = new int[n];
	memset(index, 0, (node_num+1)*sizeof(int));
	
	unary_costs = new double[n];
	memset(unary_costs, 0, n*sizeof(double));

	return mapping;
}

inline void TreeMRF::SubtractUnary(int _i, double* D)
{
	int a;
	Node* i = &nodes[node_num-_i-1];
	for (a=0; a<i->K; a++) unary_costs[i->D_index+a] -= D[a];
}

inline void TreeMRF::atom_to_vector(Atom _a, Vector x)
{
	int* solution = (int*)_a;
	int i, index;

	memset(x, 0, n*sizeof(double));
	for (i=index=0; i<node_num; i++)
	{
		x[index+solution[i]] = 1;
		index += nodes[i].K;
	}
}

inline void TreeMRF::add_atom_to_vector(double gamma, Atom _a, Vector x)
{
	int* solution = (int*)_a;
	int i, index;

	for (i=index=0; i<node_num; i++)
	{
		x[index+solution[i]] += gamma;
		index += nodes[i].K;
	}
}


inline double TreeMRF::dot_product(Atom _a, Vector g)
{
	int* solution = (int*)_a;
	int i, index;
	double sum = 0;

	for (i=index=0; i<node_num; i++)
	{
		sum += g[index + solution[i]];
		index += nodes[i].K;
	}
	return sum;
}

inline double TreeMRF::weighted_dot_product(Atom _a, Atom _b)
{
	int* a = (int*)_a;
	int* b = (int*)_b;
	int i, index;
	double sum = 0;

	for (i=index=0; i<node_num; i++)
	{
		if (a[i] == b[i]) sum += dot_product_weights[a[i] + index];
		index += nodes[i].K;
	}
	return sum;
}


inline int TreeMRF::compare_atoms(Atom _a, Atom _b)
{
	int* a = (int*) _a;
	int* b = (int*) _b;
	int i;

	for (i=0; i<node_num; i++)
	{
		if (a[i] != b[i]) return (a[i] < b[i]) ? -1 : +1;
	}
	return 0;
}

inline TreeMRF::Atom TreeMRF::min_oracle(Vector g)
{
	int* solution = (int*) malloc(node_num*sizeof(int));
	Node* i;
	int a;
	double* D = (double*) rbuf->Alloc(n*sizeof(double) + back_ptr_size*sizeof(int));
	int* back_ptr = (int*)(D + n);
	if (unary_costs)
	{
		for (a=0; a<n; a++) D[a] = g[a] + unary_costs[a];
	}
	else
	{
		memcpy(D, g, n*sizeof(double));
	}

	for (i=nodes; i<nodes+node_num-1; i++)
	{
		Node* p = i->parent;
		double* Di = D+i->D_index;
		double* Dp = D+p->D_index;
		if (i->is_original_T)
		{
			i->T->pass_message_fw(Di, Dp, back_ptr+i->back_ptr_index, i->K, p->K);
		}
		else
		{
			Node* ii = &parent_tree->nodes[i->global_index];
			Node* pp = &parent_tree->nodes[p->global_index];
			i->T->pass_message_fw(Di, Dp, back_ptr+i->back_ptr_index, 
				ii->K, pp->K, 
				i->K, p->K,
				mapping_to_parent+i->D_index, mapping_to_parent+p->D_index,
				ii->D_index, pp->D_index);
		}
	}

	int a_best = 0;
	double cost_best = D[i->D_index];
	for (a=1; a<i->K; a++)
	{
		if (cost_best > D[i->D_index+a]) { cost_best = D[i->D_index+a]; a_best = a; }
	}
	solution[node_num-1] = a_best;
	for (i--; i>=nodes; i--) solution[(int)(i-nodes)] = back_ptr[i->back_ptr_index + solution[(int)(i->parent-nodes)]];

	return solution;
}

inline double TreeMRF::compute_atom_cost(Atom _a)
{
	int* solution = (int*) _a;
	int _i;
	Node* i;

	double cost = const_term;
	for (_i=0, i=nodes; _i<node_num-1; _i++, i++)
	{
		Node* p = i->parent;
		int _p = (int)(p - nodes);
		int a = solution[_i];
		int b = solution[_p];
		if (i->is_original_T)
		{
			cost += i->T->get(a, b, i->K, p->K);
		}
		else
		{
			Node* ii = &parent_tree->nodes[i->global_index];
			Node* pp = &parent_tree->nodes[p->global_index];

			a = mapping_to_parent[i->D_index + a] - ii->D_index;
			b = mapping_to_parent[p->D_index + b] - pp->D_index;
			cost += i->T->get(a, b, ii->K, pp->K); 
		}
	}
	if (unary_costs)
	{
		for (_i=0, i=nodes; _i<node_num; _i++, i++) cost += unary_costs[i->D_index + solution[_i]];
	}
	return cost;
}

inline void TreeMRF::init_x(Atom a)
{
	if (!options->update_x) return;

	int _i, _p;
	Node* i;

	for (_i = 0, i = nodes; _i < node_num - 1; _i++, i++)
	{
		Node* p = i->parent;
		_p = (int)(i->parent - nodes);

		i->x_cost = i->T->get(((int*)(a))[_i], ((int*)(a))[_p], i->K, p->K);
	}
}

inline void TreeMRF::update_x(Vector x_new, double c0, int atom_num, double* c, Atom* atoms)
{
	if (!options->update_x) return;

	int _i, _p, r;
	Node* i;

	if (!parent_tree)
	{
		for (_i = 0, i = nodes; _i < node_num - 1; _i++, i++)
		{
			Node* p = i->parent;
			_p = (int)(i->parent - nodes);

			i->x_cost *= c0;
			for (r = 0; r < atom_num; r++)
			{
				double V = i->T->get(((int*)(atoms[r]))[_i], ((int*)(atoms[r]))[_p], i->K, p->K);
				i->x_cost += c[r] * V;
			}
		}
	}
	else
	{
		for (_i = 0, i = nodes; _i < node_num - 1; _i++, i++)
		{
			Node* p = i->parent;
			_p = (int)(i->parent - nodes);

			i->x_cost *= c0;
			if (i->is_original_T)
			{
				for (r = 0; r < atom_num; r++)
				{
					double V = i->T->get(((int*)(atoms[r]))[_i], ((int*)(atoms[r]))[_p], i->K, p->K);
					i->x_cost += c[r] * V;
				}
			}
			else
			{
				Node* ii = &parent_tree->nodes[i->global_index];
				Node* pp = &parent_tree->nodes[p->global_index];
				for (r = 0; r < atom_num; r++)
				{
					int a = mapping_to_parent[i->D_index + ((int*)(atoms[r]))[_i]] - ii->D_index;
					int b = mapping_to_parent[p->D_index + ((int*)(atoms[r]))[_p]] - pp->D_index;
					double V = i->T->get(a, b, ii->K, pp->K);
					i->x_cost += c[r] * V;
				}
			}

			//i->Tr.test_consistency(A, B, x_new + i->D_index, x_new + i->parent->D_index);
		}
	}

	if ( 0 )
	{
		static int counter = 0;
		counter ++;
		int k;
		for (_i = 0, i = nodes; _i < node_num - 1; _i++, i++)
		{
			double sum = 0;
			for (k=0; k<nodes[_i].K; k++)
			{
				if (x_new[k] < -1e-6) { printf("negative x: %f\n", x_new[k]); getchar(); }
				sum += x_new[k];
			}
			if (sum < 1-1e-6 || sum > 1+1e-6)
			{
				printf("unnormalized x: %f [%d]\n", sum, counter); getchar();
			}
			x_new += nodes[_i].K;
		}
	}
}

inline void TreeMRF::update_x_AFW(Vector x_new, double beta, Atom _v)
{
	printf("AFW cannot be used with TreeMRF\n");
	exit(1);
}

inline void TreeMRF::destroy_atom(Atom a)
{
	free((int*)a);
}


inline void TreeMRF::set_current_atom(Atom _a)
{
	int k;
	int* solution = (int*)_a;
	if (parent_tree == NULL)
	{
		for (k=0; k<node_num; k++) nodes[k].current_solution = solution[k];
	}
	else
	{
		for (k=0; k<node_num; k++)
		{
			Node* I = &nodes[k];
			Node* i = &parent_tree->nodes[I->global_index];
			i->current_solution = mapping_to_parent[I->D_index + solution[k]] - i->D_index;
		}
	}
}

inline Problem::Atom TreeMRF::get_current_atom()
{
	int k, a;
	int* solution = (int*) malloc(node_num*sizeof(int));

	if (!parent_tree)
	{
		for (k=0; k<node_num; k++) solution[k] = nodes[k].current_solution;
	}
	else
	{
		for (k=0; k<node_num; k++)
		{
			Node* I = &nodes[k];
			Node* i = &parent_tree->nodes[I->global_index];
			for (a=0; a<I->K; a++)
			{
				if (mapping_to_parent[I->D_index + a] - i->D_index == i->current_solution)
				{
					solution[k] = a;
					break;
				}
			}
			if (a == I->K)
			{
				free(solution);
				return NULL;
			}
		}
	}
	return solution;
}

inline bool TreeMRF::contract(Vector x, double x_cost, Atom _a, Vector x_average)
{
	LRSolver::SubproblemOptions* Soptions = LR->GetOptions(this);

	int* solution = (int*) _a;

	Node* i;
	int _i, k, m;

//	int* tmp_buf = new int[ (node_num+1) + n + 3*node_num ];
//	int* index_old = tmp_buf; // # node_num+1
//	int* rev_mapping_old = tmp_buf + (node_num+1); // n
//	int* tree_ids = rev_mapping_old + n; // # node_num

	int* tmp_buf = new int[ 3*node_num ];
	int* tree_ids = tmp_buf; // # node_num

	int* tree_sizes = tree_ids + node_num; // # node_num
	int* tree_label_sizes = tree_sizes + node_num; // # node_num
	int s, tree_num = 0;
	double cost0 = 0;
	if (pairwise_terms_buf) free(pairwise_terms_buf);
	pairwise_terms_buf = NULL;

//	memcpy(index_old, index, (node_num+1)*sizeof(int));
//	memcpy(rev_mapping_old, rev_mapping, index[node_num]*sizeof(int));

	int rev_mapping_size = 0;
	for (_i=0, i=nodes; _i<node_num; _i++, i++)
	{
		index[_i] = rev_mapping_size;
		for (k=0; k<i->K; k++)
		{
			////////////////////////////////////////////////////////////////////////
			////////////////////////////////////////////////////////////////////////
			////////////////////////////////////////////////////////////////////////

			if ((x[i->D_index+k] > 0) || (solution[_i] == k)
			  || (x_average && x_average[i->D_index+k] > Soptions->contract_average_eps) )
				rev_mapping[rev_mapping_size ++] = k;

			////////////////////////////////////////////////////////////////////////
			////////////////////////////////////////////////////////////////////////
			////////////////////////////////////////////////////////////////////////
		}
		if (rev_mapping_size == index[_i] + 1) rev_mapping_size = index[_i];

		if (a_prev && i->parent)
		{
			// set i->x_cost
			Node* p = i->parent;
			int _p = (int)(i->parent-nodes), a, b;
			if (i->i_child == NULL && p->i_child != NULL)
			{
				i->x_cost = 0;
				for (b=0; b<p->K; b++) i->x_cost += i->T->get(((int*)a_prev)[_i], b, i->K, p->K)*x[p->D_index+b];
			}
			else if (i->i_child != NULL && p->i_child == NULL)
			{
				i->x_cost = 0;
				for (a=0; a<i->K; a++) i->x_cost += i->T->get(a, ((int*)a_prev)[_p], i->K, p->K)*x[i->D_index+a];
			}
			else if (i->i_child != NULL && p->i_child != NULL)
			{
				i->x_cost = i->i_child->x_cost;
			}
		}
	}
	index[_i] = rev_mapping_size;

	if (rev_mapping_size >= n * Soptions->contract_threshold)
	{
		delete[] tmp_buf;
		if (a_prev)
		{
			Node* i;
			for (i = nodes; i < nodes + node_num; i++) i->i_child = NULL;
			destroy_atom(a_prev);
			a_prev = NULL;
		}
		return false;
	}

	for (_i=node_num-1, i=nodes+_i; _i>=0; _i--, i--)
	{
		if (index[_i] == index[_i+1])
		{
			tree_ids[_i] = -1;
			continue;
		}
		if (i->parent && tree_ids[(int)(i->parent-nodes)] >= 0)
		{
			int s = tree_ids[_i] = tree_ids[(int)(i->parent-nodes)];
			tree_sizes[s] ++;
			tree_label_sizes[s] += index[_i+1] - index[_i];
		}
		else
		{
			tree_sizes[tree_num] = 1;
			tree_ids[_i] = tree_num;
			tree_label_sizes[tree_num] = index[_i+1] - index[_i];
			tree_num ++;
		}
	}

	TreeMRF** trees = new TreeMRF*[tree_num];
	for (s=0; s<tree_num; s++)
	{
		trees[s] = new TreeMRF(tree_sizes[s], options, NULL, rbuf);
		trees[s]->mapping_to_parent = new int[tree_label_sizes[s]];
		trees[s]->unary_costs = new double[tree_label_sizes[s]];
		memset(trees[s]->unary_costs, 0, tree_label_sizes[s]*sizeof(double));
		trees[s]->parent_tree = this;
	}
	int pairwise_terms_size = 0;
	for (_i=0, i=nodes; _i<node_num; _i++, i++)
	{
		s = tree_ids[_i];
		if (s >= 0)
		{
			TreeMRF* t = trees[s];
			Node* I = i->i_child = &t->nodes[t->node_num ++];
			I->x_cost = i->x_cost;
			I->global_index = _i;
			I->K = index[_i+1] - index[_i];
			I->D_index = t->n;
			for (k=0; k<I->K; k++)
			{
				int ind = i->D_index + rev_mapping[index[_i] + k];
				if (unary_costs) t->unary_costs[t->n] = unary_costs[ind];
				t->mapping_to_parent[t->n ++] = ind;
			}
			if (i->parent)
			{
				int _p = (int)(i->parent-nodes);
				if (tree_ids[_p] >= 0)
				{
					int AA = I->K, BB = index[_p+1] - index[_p];
					i->is_original_T = i->T->explicit_contraction(i->K, i->parent->K, AA, BB);
					if (i->is_original_T) pairwise_terms_size += PairwiseTermGeneralSize(AA, BB);
				}
			}
		}
		else
		{
			i->i_child = NULL;
		}
	}
	if (pairwise_terms_size) pairwise_terms_buf = (char*) malloc(pairwise_terms_size);
	char* pairwise_terms_ptr = pairwise_terms_buf;
	for (_i=0, i=nodes; _i<node_num; _i++, i++)
	{
		s = tree_ids[_i];
		Node* p = i->parent;
		int _p = (p) ? (int)(p - nodes) : -1;
		if (s < 0)
		{
			if (unary_costs)
			{
				cost0 += unary_costs[i->D_index + solution[_i]];
			}
			if (p)
			{
				if (tree_ids[_p]>=0)
				{
					TreeMRF* t = trees[tree_ids[_p]];
					Node* P = p->i_child;
					for (k=0; k<P->K; k++)
					{
						t->unary_costs[P->D_index+k] += i->T->get(solution[_i], rev_mapping[index[_p] + k], i->K, p->K);
					}
				}
				else
				{
					cost0 += i->T->get(solution[_i], solution[_p], i->K, p->K);
				}
			}
			continue;
		}
		TreeMRF* t = trees[s];
		Node* I = i->i_child;
		I->parent = NULL;
		if (p)
		{
			if (tree_ids[_p] < s)
			{
				for (k=0; k<I->K; k++)
				{
					t->unary_costs[I->D_index+k] += i->T->get(rev_mapping[index[_i] + k], solution[_p], i->K, p->K);
				}
			}
			else
			{
				I->parent = p->i_child;
				int A = i->K, B = p->K, AA = I->K, BB = I->parent->K;
				int* list_a = &rev_mapping[index[_i]];
				int* list_b = &rev_mapping[index[_p]];

//				int AA_prev = index_old[_i+1] - index_old[_i], BB_prev = index_old[_p+1] - index_old[_p];

				if (i->is_original_T)
				{
					PairwiseTermGeneral* TT = (PairwiseTermGeneral*) pairwise_terms_ptr;
					pairwise_terms_ptr += PairwiseTermGeneralSize(AA, BB);
					TT->type = PairwiseTerm::Type::GENERAL;
					TT->init(i->T, A, B, AA, BB, list_a, list_b);
					I->T = TT;
					I->is_original_T = true;
				}
				else
				{
					i->is_original_T = true;
					I->T = i->T;
					I->is_original_T = false;
				}
				t->const_term += i->x_cost;
			}
		}
		for (k=0; k<I->K; k++)
		{
			t->const_term += t->unary_costs[I->D_index+k] * x[i->D_index + rev_mapping[index[_i] + k]];
		}
	}

	for (s=0; s<tree_num; s++)
	{
		TreeMRF* t = trees[s];
		for (i=t->nodes, m=0; i<t->nodes+t->node_num-1; i++) { i->back_ptr_index = m; m += i->parent->K; }
		t->back_ptr_size = m;
		double cost = t->const_term;
		t->const_term = 0;
		if (s == 0) { t->const_term = cost0; cost += cost0; }
		LR->AddContractedSubproblem(t, t->n, cost, t->n, t->mapping_to_parent);
	}

	delete [] trees;
	delete [] tmp_buf;

	if (!a_prev)
	{
		a_prev = malloc(node_num * sizeof(int));
	}
	memcpy(a_prev, _a, node_num*sizeof(int));

	return true;
}

inline double TreeMRF::compute_x_cost(Vector x)
{
	int _i;
	Node* i;

	double total_cost = const_term;
	for (_i=0, i=nodes; _i<node_num-1; _i++, i++)
	{
		double cost;
		Node* p = i->parent;
		int _p = (int)(p - nodes);
		if (i->is_original_T)
		{
			cost = i->T->opt_transport(x+i->D_index, x+p->D_index, i->K, p->K);
		}
		else
		{
			Node* ii = &parent_tree->nodes[i->global_index];
			Node* pp = &parent_tree->nodes[p->global_index];

			cost = i->T->opt_transport(x+i->D_index, x+p->D_index, ii->K, pp->K,
				i->K, p->K, mapping_to_parent + i->D_index, mapping_to_parent + p->D_index, ii->D_index, pp->D_index);
		}
		total_cost += cost;
	}
	if (unary_costs)
	{
		for (_i=0; _i<n; _i++) total_cost += unary_costs[_i] * x[_i];
	}
	return total_cost;
}

inline double TreeMRF::adjust_x(Vector x)
{
	int _i, k;
	Node* i;

	for (_i=0, i=nodes; _i<node_num; _i++, i++)
	{
		double sum = 0;
		for (k=0; k<i->K; k++)
		{
			if (x[i->D_index+k] < options->adjustx_eps) x[i->D_index+k] = 0;
			sum += x[i->D_index+k];
		}
		if (sum <= 0)
		{
			for (k=0; k<i->K; k++) x[i->D_index+k] = 1.0 / i->K;
		}
		else
		{
			for (k=0; k<i->K; k++) x[i->D_index+k] /= sum;
		}
	}

	double total_cost = const_term;
	for (_i=0, i=nodes; _i<node_num-1; _i++, i++)
	{
		double cost;
		Node* p = i->parent;
		int _p = (int)(p - nodes);
		
		if (i->is_original_T)
		{
			cost = i->T->opt_transport(x+i->D_index, x+p->D_index, i->K, p->K);
			i->x_cost = cost;
		}
		else
		{
			Node* ii = &parent_tree->nodes[i->global_index];
			Node* pp = &parent_tree->nodes[p->global_index];

			cost = i->T->opt_transport(x+i->D_index, x+p->D_index, ii->K, pp->K,
				i->K, p->K, mapping_to_parent + i->D_index, mapping_to_parent + p->D_index, ii->D_index, pp->D_index);
			i->x_cost = ii->x_cost = cost;
		}
		total_cost += cost;
	}
	if (unary_costs)
	{
		for (_i=0; _i<n; _i++) total_cost += unary_costs[_i] * x[_i];
	}
	return total_cost;
}

inline void TreeMRF::set_final_dual_solution(Vector g)
{
return; // not implemented
/*
	if (!options->final_callback_fn) return;
	if (!unary_costs) { printf("error in set_final_dual_solution()\n"); exit(1); }

	Node* i;
	int a, b, K_max = 0;

	for (i=nodes; i<nodes+node_num; i++) { if (K_max < i->K) K_max = i->K; }

	double* D = (double*) rbuf->Alloc((n+2*K_max)*sizeof(double));
	double* A = D + n;
	double* B = A + K_max;
	for (a=0; a<n; a++) D[a] = g[a] + unary_costs[a];

	for (i=nodes; i<nodes+node_num-1; i++)
	{
		Node* p = i->parent;
		double* Di = D+i->D_index;
		double* Dp = D+p->D_index;
		for (a=0; a<i->K; a++) A[a] = -Di[a];
		i->T->pass_message_srmp_fw(A, B, i->K, p->K);
		options->final_callback_fn(1, B, i->edge_arg);
		for (b=0; b<p->K; b++) Dp[b] += B[b];
	}

	double rho = 0.75;

//	double* Di = D+i->D_index;
//	for (a=0; a<i->K; a++) Di[a] *= -rho;

	for (i--; i>=nodes+node_num-1; i--)
	{
		Node* p = i->parent;
		double* Di = D+i->D_index;
		double* Dp = D+p->D_index;
		i->T->pass_message_srmp_bw(Di, Dp, i->K, p->K);
		options->final_callback_fn(0, Di, i->edge_arg);
//		for (a=0; a<i->K; a++) Di[a] *= -rho;
	}
*/
}

inline void TreeMRF::print_atom(Atom a)
{
	int i;
	int* solution = (int*) a;

	if (!parent_tree)
	{
		for (i=0; i<node_num; i++) printf("%d", solution[i]);
	}
	else
	{
		int* x = new int[parent_tree->node_num];
		for (i=0; i<parent_tree->node_num; i++) x[i] = -1;
		for (i=0; i<node_num; i++)
		{
			int p = nodes[i].global_index;
			x[p] = mapping_to_parent[nodes[i].D_index + solution[i]] - parent_tree->nodes[p].D_index;	
		}
		for (i=0; i<parent_tree->node_num; i++) { if (x[i]<0) printf("*"); else printf("%d", x[i]); }
		delete [] x;
	}
}

#endif
