#include <stdio.h>
#include <stdlib.h>
#include "arboricity/arboricity.h"
#include "graph_decomposition.h"

struct RandomPermutation
{
	RandomPermutation(int _n) : n(_n) { perm = new int[n]; i = n; }
	~RandomPermutation() { delete [] perm; }

	void Start()
	{
		if (3*i > n)
		{
			for (i=0; i<n; i++) perm[i] = i;
		}
		else
		{
			int j;
			for (j=0; j<=i; j++)
			{
				int k = perm[j];
				if (k > i) perm[k] = k;
				perm[j] = j;
			}
		}
		i = 0;
	}
	int GetNext() // can be called at most n times
	{
		int j = i + RandomInteger(n-i);
		int k = perm[j]; perm[j] = perm[i]; perm[i] = k;
		i ++;
		return k;
	}

private:
	int RandomInteger(int N) // returns random integer in [0,N-1]
	{
		while ( 1 )
		{
			int x = (int) ( (((double)N) * rand()) / RAND_MAX );
			if (x<N) return x;
		}
	}

	int i, n;
	int* perm;
};

GraphDecomposition::GraphDecomposition(int _node_num, int _edge_num_max)
	: node_num(_node_num), edge_num(0), edge_num_max(_edge_num_max),
	  method(NONE), grid_edges(NULL)

{
	Node* i;
	nodes = new Node[node_num];
	edges = new Edge[edge_num_max];
	for (i=nodes; i<nodes+node_num; i++)
	{
		i->first[0] = i->first[1] = NULL;
	}
	forest_num = 0; forest_num_max = 32;
	forests = (Forest**) malloc(forest_num_max*sizeof(Forest*));
}

GraphDecomposition::~GraphDecomposition()
{
	delete [] nodes;
	delete [] edges;
	if (grid_edges) delete [] grid_edges;
	int f;
	for (f=0; f<forest_num; f++) delete forests[f];
	free(forests);
}

GraphDecomposition::Forest* GraphDecomposition::AllocForest()
{
	if (forest_num >= forest_num_max)
	{
		forest_num_max = 2*forest_num;
		forests = (Forest**) realloc(forests, forest_num_max*sizeof(Forest*));
	}
	forests[forest_num] = new Forest(node_num);
	return forests[forest_num ++];
}

GraphDecomposition::EdgeId GraphDecomposition::AddEdge(NodeId _i, NodeId _j)
{
	if (_i > _j) { int k = _i; _i = _j; _j = k; }
	if (_i == _j || _i<0 || _j<0 || _i>=node_num || _j>=node_num) { printf("Incorrect call to AddEdge()\n"); exit(1); }
	if (edge_num >= edge_num_max) { printf("Too many edges\n"); exit(1); }
	Edge* e = &edges[edge_num];

	e->i[0] = _i; e->i[1] = _j;
//	e->next[0] = nodes[_i].first[0]; e->next[1] = nodes[_j].first[1];
//	nodes[_i].first[0] = nodes[_j].first[1] = e;

	return edge_num ++;
}

void GraphDecomposition::DecomposeGreedy()
{
	int e;
	int f;
	Forest* F = AllocForest();
	F->Reset();
	RandomPermutation* R = new RandomPermutation(forest_num);
	for (e=0; e<edge_num; e++)
	{
		R->Start();
		for (f=0; f<forest_num; f++)
		{
			if (forests[R->GetNext()]->AddEdge(edges+e, node_num)) break;
		}
		if (f == forest_num)
		{
			Forest* F = AllocForest();
			F->Reset();
			F->AddEdge(edges+e, node_num);
			delete R;
			R = new RandomPermutation(forest_num);
		}
	}
	delete R;
}

GraphDecomposition::NodeId GraphDecomposition::GetNextRootGeneral(int& tree_size)
{
	while ( 1 )
	{
		if (!F_current)
		{
			if (f_next >= forest_num) return -1;
			F_current = forests[f_next ++];
			int i, j;
			Edge* e;
			for (i=0; i<node_num; i++) nodes[i].first[0] = nodes[i].first[1] = NULL;
			for (e=F_current->edge_first; e; e=e->list_next)
			{
				i = e->i[0]; j = e->i[1];
				e->next[0] = nodes[i].first[0]; e->next[1] = nodes[j].first[1];
				nodes[i].first[0] = nodes[j].first[1] = e;
				e->F = NULL;
			}
			root_current = -1;
		}
		for (root_current ++; root_current<node_num; root_current ++)
		{
			if (F_current->union_find_ptr[root_current] < -1)
			{
				tree_size = -F_current->union_find_ptr[root_current];
				current_list = root_current;
				nodes[root_current].tmp = -1;
				return root_current;
			}
		}
		F_current = NULL;
	}
}

GraphDecomposition::EdgeId GraphDecomposition::GetNextEdgeGeneral(NodeId& parent, NodeId& child)
{
	while ( 1 )
	{
		if (current_list < 0) return -1;
		Node* i = nodes + current_list;
		int dir;
		for (dir=0; dir<2; dir++)
		{
			Edge* e;
			while ((e=i->first[dir]))
			{
				i->first[dir] = e->next[dir];
				if (e->F == NULL)
				{
					e->F = F_current;
					parent = current_list;
					child = (e->i[0] == current_list) ? e->i[1] : e->i[0];
					nodes[child].tmp = nodes[current_list].tmp;
					nodes[current_list].tmp = child;
					return (int)(e-edges);
				}
			}
		}
		current_list = i->tmp;
	}
}

void GraphDecomposition::Decompose(Method _method)
{
	method = _method;
	if (method == Method::NONE)
	{
		if (TryGrid())
		{
			InitScan();
			return;
		}
		method = Method::OPT;
	}
	if (method == Method::GRID4 || method == Method::GRID8)
	{
		method = Method::NONE;
		TryGrid();
		if (method != _method && (method != Method::GRID4 || sizeY != 1))
		{
			printf("error: non-grid graph!\n");
			exit(1);
		}
	}
	if (method == Method::GREEDY)
	{
		DecomposeGreedy();
	}
	else if (method == Method::OPT)
	{
		DecomposeOPT();
	}
	else
	{
		printf("not implemented\n");
		exit(1);
	}
	InitScan();
}

void GraphDecomposition::InitScan()
{
	F_current = NULL;
	f_next = 0;

	grid_next_chain = 0;
}

#define GET_SORTED_ENDPOINTS(e, p, q) { p = e->i[0]; q = e->i[1]; if (p > q) { int r = p; p = q; q = r; } }
#define GRID_NODE(x, y) ((x) + (y)*sizeX)
#define GRID8_DIAG_SIZE(x, sizeX, sizeY) ( (x)<(sizeY) ? ((x)+1) : (sizeY) )

bool GraphDecomposition::TryGrid()
{
	int p, q;
	Node* i;
	Edge* e;
	for (i=nodes; i<nodes+node_num; i++) i->tmp = 0;
	for (e=edges; e<edges+edge_num; e++)
	{
		nodes[e->i[0]].tmp ++;
		nodes[e->i[1]].tmp ++;
	}

	if (nodes[0].tmp == 1 && edge_num == node_num-1) // try chain
	{
		grid_edges = new EdgeId[edge_num];
		memset(grid_edges, 0, edge_num*sizeof(EdgeId));
		for (e=edges; e<edges+edge_num; e++)
		{
			GET_SORTED_ENDPOINTS(e, p, q);
			if (p < 0 || p >= node_num-1 || q != p+1) return false;
			if (grid_edges[p]) return false;
			grid_edges[p] = (int)(e-edges);
		}
		method = Method::GRID4;
		sizeX = node_num;
		sizeY = 1;
		printf("chain %d\n", sizeX);
		return true;
	}

	if (nodes[0].tmp == 2) // try GRID4
	{
		for (sizeX=2; sizeX<=node_num; sizeX++) { if (nodes[sizeX-1].tmp == 2) break; if (nodes[sizeX-1].tmp != 3) return false;  }
		sizeY = node_num / sizeX;
		if (node_num != sizeX*sizeY) return false;
		if (edge_num != sizeX*(sizeY-1) + sizeY*(sizeX-1)) return false;
		grid_edges = new EdgeId[2*node_num];
		memset(grid_edges, 0, 2*node_num*sizeof(EdgeId));
		for (e=edges; e<edges+edge_num; e++)
		{
			int k;
			GET_SORTED_ENDPOINTS(e, p, q);
			int x = p % sizeX, y = p / sizeX;
			if      (x < sizeX && q == p+1)   k = p;
			else if (y < sizeY && q==p+sizeX) k = p + node_num;
			else return false;
			if (grid_edges[k]) return false;
			grid_edges[k] = (int)(e-edges);
		}
		method = Method::GRID4;
		printf("grid4 %dx%d\n", sizeX, sizeY);
		return true;
	}
	
	if (nodes[0].tmp == 3) // try GRID8
	{
		for (sizeX=2; sizeX<=node_num; sizeX++) { if (nodes[sizeX-1].tmp == 3) break; if (nodes[sizeX-1].tmp != 5) return false; }
		sizeY = node_num / sizeX;
		if (node_num != sizeX*sizeY) return false;
		if (edge_num != sizeX*(sizeY-1) + sizeY*(sizeX-1) + 2*(sizeX-1)*(sizeY-1)) return false;
		grid_edges = new EdgeId[4*node_num];
		memset(grid_edges, 0, 4*node_num*sizeof(EdgeId));
		for (e=edges; e<edges+edge_num; e++)
		{
			int k;
			GET_SORTED_ENDPOINTS(e, p, q);
			int x = p % sizeX, y = p / sizeX;
			if      (x < sizeX-1 && q == p+1)                    k = p;
			else if (x < sizeX-1 && y < sizeY-1 && q==p+sizeX+1) k = p + node_num;
			else if (y < sizeY-1 && q==p+sizeX)                  k = p + 2*node_num;
			else if (x > 0 && y < sizeY-1 && q==p+sizeX-1)       k = p + 3*node_num;
			else return false;
			if (grid_edges[k]) return false;
			grid_edges[k] = (int)(e-edges);
		}
		method = Method::GRID8;
		printf("grid8 %dx%d\n", sizeX, sizeY);
		return true;
	}
	return false;
}

GraphDecomposition::NodeId GraphDecomposition::GetNextRoot(int& tree_size)
{
	switch (method)
	{
		case Method::GRID4: return GetNextRootGRID4(tree_size);
		case Method::GRID8: return GetNextRootGRID8(tree_size);
		default: return GetNextRootGeneral(tree_size);
	}
}

GraphDecomposition::EdgeId GraphDecomposition::GetNextEdge(NodeId& parent, NodeId& child)
{
	EdgeId e;
	switch (method)
	{
		case Method::GRID4: 
		case Method::GRID8: e = GetNextEdgeGRID(parent, child); break;
		default: e = GetNextEdgeGeneral(parent, child); break;
	}
	if (e >= 0)
	if (edges[e].i[0] != parent || edges[e].i[1] != child)
	if (edges[e].i[1] != parent || edges[e].i[0] != child)
	{
		printf("?"); exit(1);
	}
	return e;
}

GraphDecomposition::NodeId GraphDecomposition::GetNextRootGRID4(int& tree_size)
{
	if (grid_next_chain < sizeY)
	{
		tree_size = sizeX;
		grid_p = GRID_NODE(sizeX-1, grid_next_chain);
		grid_e = grid_p;
		grid_shift = -1;
	}
	else if (grid_next_chain < sizeX + sizeY && sizeY > 1)
	{
		tree_size = sizeY;
		grid_p = GRID_NODE(grid_next_chain - sizeY, sizeY-1);
		grid_e = node_num + grid_p;
		grid_shift = -sizeX;
	}
	else return -1;

	grid_next_chain ++;
	grid_nodes_left = tree_size - 1;
	return grid_p;
}

GraphDecomposition::NodeId GraphDecomposition::GetNextRootGRID8(int& tree_size)
{
	int c = grid_next_chain;
	while ( 1 )
	{
		if (c < sizeY)
		{
			tree_size = sizeX;
			grid_p = GRID_NODE(sizeX-1, c);
			grid_e = grid_p;
			grid_shift = -1;
			break;
		}
		c -= sizeY;

		if (c < sizeX+sizeY-3)
		{
			if (c < sizeX-1)
			{
				tree_size = GRID8_DIAG_SIZE(1+c, sizeX, sizeY);
				grid_p = GRID_NODE(1+c, sizeY-1);
			}
			else
			{
				int y = sizeY-2 - (c-sizeX+1);
				tree_size = GRID8_DIAG_SIZE(y, sizeY, sizeX);
				grid_p = GRID_NODE(sizeX-1, y);
			}
			grid_e = node_num + grid_p;
			grid_shift = -sizeX-1;
			break;
		}
		c -= sizeX+sizeY-3;

		if (c < sizeX)
		{
			tree_size = sizeY;
			grid_p = GRID_NODE(c, sizeY-1);
			grid_e = 2*node_num + grid_p;
			grid_shift = -sizeX;
			break;
		}
		c -= sizeX;

		if (c < sizeX+sizeY-3)
		{
			if (c < sizeY-1)
			{
				tree_size = GRID8_DIAG_SIZE(1+c, sizeY, sizeX);
				grid_p = GRID_NODE(0, 1+c);
			}
			else
			{
				int x = c - sizeY + 2;
				tree_size = GRID8_DIAG_SIZE(sizeX-1-x, sizeX, sizeY);
				grid_p = GRID_NODE(x, sizeY-1);
			}
			grid_e = 3*node_num + grid_p;
			grid_shift = -sizeX+1;
			break;
		}
		c -= sizeX+sizeY-3;

		return -1;
	}

	grid_next_chain ++;
	grid_nodes_left = tree_size - 1;
	return grid_p;
}

GraphDecomposition::EdgeId GraphDecomposition::GetNextEdgeGRID(NodeId& parent, NodeId& child)
{
	if ((grid_nodes_left --) <= 0) return -1;
	parent = grid_p;
	grid_p += grid_shift;
	child = grid_p;
	grid_e += grid_shift;
	return grid_edges[grid_e];
}

void GraphDecomposition::DecomposeOPT()
{
	Edge* e;
	UndirectedGraph G(node_num, edge_num);
	for (e=edges; e<edges+edge_num; e++)
	{
		G.AddEdge(e->i[0], e->i[1], 1);
	}
	int* A = new int[node_num];
	int f, i;
	int num = G.Solve();
	for (f=0; f<num; f++)
	{
		G.GetForestEdges(f, A);
		Forest* F = AllocForest();
		F->Reset();
		for (i=0; i<node_num; i++)
		{
			if (A[i] >= 0) F->AddEdge(edges+A[i], node_num);
		}
	}
	delete [] A;
}

void GraphDecomposition::PrintStats(bool print_tree_sizes)
{
	if (method == Method::GRID4)
	{
		printf("Grid4 %dx%d, split into horizontal chains / vertical chains\n", sizeX, sizeY);
	}
	else if (method == Method::GRID8)
	{
		printf("Grid4 %dx%d, split into horizontal chains / vertical / diagonal chains\n", sizeX, sizeY);
	}
	else if (!print_tree_sizes)
	{
		int i, f, tree_num = 0;
		int size_min = node_num, size_max = 0;
		for (f=0; f<forest_num; f++)
		{
			for (i=0; i<node_num; i++)
			{
				int s = -forests[f]->union_find_ptr[i];
				if (s < 2) continue;
				tree_num ++;
				if (size_min > s) size_min = s;
				if (size_max < s) size_max = s;
			}
		}
		
		printf("%d forests, %d trees with sizes in [%d,%d]\n", forest_num, tree_num, size_min, size_max);
	}
	else
	{
		printf("%d forests with sizes", forest_num);
		int i, f;
		for (f=0; f<forest_num; f++)
		{
			for (i=0; i<node_num; i++)
			{
				int s = -forests[f]->union_find_ptr[i];
				if (s < 2) continue;
				printf(" %d", s);
			}
		}
		printf("\n");
	}

}

void GraphDecomposition::PrintTrees()
{
	InitScan();
	while ( 1 )
	{
		int tree_size;
		int r = GetNextRoot(tree_size);
		if (r < 0) break;
		printf("tree (%d nodes): %d", tree_size, r);
		while ( 1 )
		{
			int i, j, e = GetNextEdge(i, j);
			if (e < 0) break;
			printf(" %d->%d", j, i);
		}
		printf("\n");
	}
	InitScan();
}

void GraphDecomposition::BreakLargeTrees()
{
	if (method == Method::GRID4 || method == Method::GRID8) return;

	int i, f0, f, tree_num = 0;
	int* size_max_array = new int[forest_num];
	for (f=0; f<forest_num; f++)
	{
		size_max_array[f] = 0;
		for (i=0; i<node_num; i++)
		{
			int s = -forests[f]->union_find_ptr[i];
			if (s < 2) continue;
			tree_num ++;
			if (size_max_array[f] < s) size_max_array[f] = s;
		}
	}

	int size_max = (int)(1.5*edge_num/tree_num);
	for (f0=0; f0<forest_num; f0++)
	{
		if (size_max_array[f0] <= size_max) continue;
		Forest* F0 = forests[f0];
		Edge* e;
		Edge* e_next;
		Edge* edge0_first = NULL;

		int e_num = 0;
		for (e=F0->edge_first; e; e=e->list_next) e_num ++;
		Edge** e_list = new Edge*[e_num];
		e_num = 0;
		for (e=F0->edge_first; e; e=e->list_next) e_list[e_num ++] = e;
		RandomPermutation R(e_num);
		R.Start();

		for ( ; e_num > 0; e_num--)
		{
			e = e_list[R.GetNext()];
			int r = F0->GetRoot(e->i[0]);
			bool keep_edge = true;
			if (-F0->union_find_ptr[r] > size_max)
			{
				for (f=0; f<forest_num; f++)
				{
					if (f != f0 && forests[f]->AddEdge(e, size_max)) 
					{
						keep_edge = false;
						break;
					}
				}
			}
			if (keep_edge)
			{
				e->list_next = edge0_first;
				edge0_first = e;
			}
		}

		delete [] e_list;

		F0->Reset();
		for (e=edge0_first; e; e=e_next)
		{
			e_next = e->list_next;
			F0->AddEdge(e, node_num);
		}
	}

	delete [] size_max_array;
}

/*
void TestGraphDecomposition()
{
	int i, j, x, y, e;
	int X = 5, Y = 3;

	GraphDecomposition* G = new GraphDecomposition(X*Y, 4*X*Y);
	for (x=0; x<X; x++)
	for (y=0; y<Y; y++)
	{
		i = x + y*X;
		for (e=0; e<2; e++)
		{	
			if      (e == 0) { if (x < X-1) j = i + 1; else continue; }
			else if (e == 1) { if (y < Y-1) j = i + X; else continue; }
			else if (e == 2) { if (x < X-1 && y < Y-1) j = i + X + 1; else continue; }
			else if (e == 3) { if (x > 0   && y < Y-1) j = i + X - 1; else continue; }
			G->AddEdge(i, j);
		}
	}

	G->Decompose(GraphDecomposition::Method::GREEDY);

	G->PrintTrees();
	G->PrintStats();
	printf("*****************\n");
	G->BreakLargeTrees();
	G->PrintTrees();
	G->PrintStats();

	delete G;
}

int main()
{
	srand(10);
	TestGraphDecomposition();
	return 1;
}
*/
