#ifndef __GRAPH_DECOMPOSITION__
#define __GRAPH_DECOMPOSITION__

class GraphDecomposition
{
public:
	typedef int NodeId;
	typedef int EdgeId;

	GraphDecomposition(int node_num, int edge_num_max);
	~GraphDecomposition();

	EdgeId AddEdge(NodeId i, NodeId j);

	typedef enum
	{
		NONE, // NONE means first try GRID4, then GRID8, then OPT
		GRID4,
		GRID8,
		GREEDY,
		OPT // minimizes the number of forests. Can be slow on some graphs
	} Method;

	// NOTE: EACH NODE MUST HAVE AT LEAST ONE NEIGHBOR!!!

	void Decompose(Method method=NONE);
	void BreakLargeTrees();

	// after calling Decompose(), the trees can be obtained by calling three functions below.
	// Note, InitScan() is called automatically by Decompose(); if you want to repeat the scan,
	// call InitScan() again.
	void InitScan();
	NodeId GetNextRoot(int& tree_size); // returns -1 if no more roots
	EdgeId GetNextEdge(NodeId& parent, NodeId& child); // returns -1 if no more edges

	void PrintStats(bool print_tree_sizes=true);
	void PrintTrees();

private:
	struct Node;
	struct Forest;
	struct Edge
	{
		int i[2];
		Edge* next[2];
		Edge* list_next;
		Forest* F;
	};
	struct Node
	{
		Edge* first[2];
		int tmp;
	};

	struct Forest
	{
		Forest(int _node_num) : node_num(_node_num) { union_find_ptr = new int[node_num]; }
		~Forest() { delete [] union_find_ptr; }
		void Reset()
		{
			int i;
			for (i=0; i<node_num; i++) union_find_ptr[i] = -1;
			edge_first = NULL;
		}
		void CompressPath(int start, int end)
		{
			while (start != end) { int i = union_find_ptr[start]; union_find_ptr[start] = end; start = i; }
		}
		int GetRoot(int i0)
		{
			int i = i0;
			while (union_find_ptr[i] >= 0) i = union_find_ptr[i];
			CompressPath(i0, i);
			return i;
		}
		bool AddEdge(Edge* e, int size_max) // if merging components would create a tree of size larger than size_max, then don't merge
		{
			int i, j;
			i = e->i[0]; while (union_find_ptr[i] >= 0) i = union_find_ptr[i];
			j = e->i[1]; while (union_find_ptr[j] >= 0) j = union_find_ptr[j];
			if (i == j || union_find_ptr[i] + union_find_ptr[j] < -size_max)
			{
				CompressPath(e->i[0], i);
				CompressPath(e->i[1], j);
				return false;
			}
			union_find_ptr[i] += union_find_ptr[j];
			union_find_ptr[j] = i;
			CompressPath(e->i[0], i);
			CompressPath(e->i[1], i);
			e->list_next = edge_first;
			edge_first = e;
			e->F = this;
			return true;
		}

		int node_num;
		int* union_find_ptr;
		Edge* edge_first;
	};

	int node_num;
	int edge_num, edge_num_max;
	Node* nodes;
	Edge* edges;

	void DecomposeGreedy();
	void DecomposeOPT();

	////////////////////////
	Method method;

	//////////////////////////
	// data for grid graphs //
	//////////////////////////
	int sizeX, sizeY, x, y;
	EdgeId* grid_edges;
	int grid_next_chain, grid_nodes_left, grid_shift;
	int grid_p, grid_e;

	/////////////////////////////
	// data for general graphs //
	/////////////////////////////
	Forest** forests;
	int forest_num, forest_num_max;
	int f_next, root_current, current_list;
	Forest* F_current;

	Forest* AllocForest();

	NodeId GetNextRootGRID4(int& tree_size);
	NodeId GetNextRootGRID8(int& tree_size);
	EdgeId GetNextEdgeGRID(NodeId& parent, NodeId& child);

	NodeId GetNextRootGeneral(int& tree_size);
	EdgeId GetNextEdgeGeneral(NodeId& parent, NodeId& child);

	bool TryGrid();
};

#endif
