#ifndef __OPT_TRANSPORT__
#define __OPT_TRANSPORT__

class OptTransport
{
public:
	OptTransport() {}
	~OptTransport() {}

	void Solve(int A, int B, double* V, double* x, double* y, double* z);

	typedef int NodeId;
	typedef int EdgeId;

	void AddNodeExcess(NodeId i, double excess);

	// first call returns 0, second 1, and so on.
	// cap, rev_cap must be non-negative. 
	// cost can be negative.
	EdgeId AddEdge(NodeId i, NodeId j, double cap, double rev_cap, double cost);

	double SolveMinCost();

	///////////////////////////////////////////////////

	double GetRCap(EdgeId e);
	void SetRCap(EdgeId e, double new_rcap);
	double GetReverseRCap(EdgeId e);
	void SetReverseRCap(EdgeId e, double new_rcap);
	void PushFlow(EdgeId e, double delta);
	void UpdateCost(EdgeId e, double cap_orig, double delta);

	double GetDual(NodeId i) { return nodes[i].pi; }


private:

	struct Node;
	struct Arc;

	struct Node
	{
		Arc			*firstNonsaturated;
		Arc			*firstSaturated;

		Arc			*parent;
		Node		*next; // list of nodes with positive excesses

		double		excess;
		double		pi;
		int			flag;
		union
		{
			int		heap_ptr;
			Node*	next_permanent;
		};
	};

	struct Arc
	{
		Node		*head;
		Arc			*prev;
		Arc			*next;
		Arc			*sister;	// reverse arc

		double		r_cap;		// residual capacity
		double		cost;
		double		GetRCost() { return cost + head->pi - sister->head->pi; }
	};

	int		nodeNum, edgeNum, edgeNumMax;
	Node	*nodes;
	Arc		*arcs;
	Node*	firstActive;
	int		counter;
	double	cost;

	struct PriorityQueue
	{
		PriorityQueue();
		~PriorityQueue();
		void Reset();
		double GetKey(Node* i);
		void Add(Node* i, double key);
		void DecreaseKey(Node* i, double key);
		Node* RemoveMin(double& key);

	private:
		struct Item
		{
			Node*		i;
			double		key;
		}* array;
		int N, arraySize;
		void Swap(int k1, int k2);
	};

	PriorityQueue queue;

	void SetRCap(Arc* a, double new_rcap);
	void PushFlow(Arc* a, double delta);

	void Init();
	void DecreaseRCap(Arc* a, double delta);
	void IncreaseRCap(Arc* a, double delta);
	double Augment(Node* start, Node* end);
	void Dijkstra(Node* start);

	void TestOptimality();
};


inline OptTransport::PriorityQueue::PriorityQueue()
{
	N = 0;
	arraySize = 16;
	array = (Item*) malloc(arraySize*sizeof(Item));
}

inline OptTransport::PriorityQueue::~PriorityQueue()
{
	free(array);
}

inline void OptTransport::PriorityQueue::Reset()
{
	N = 0;
}

inline double OptTransport::PriorityQueue::GetKey(Node* i)
{
	return array[i->heap_ptr].key;
}

inline void OptTransport::PriorityQueue::Swap(int k1, int k2)
{
	Item* a = array+k1;
	Item* b = array+k2;
	a->i->heap_ptr = k2;
	b->i->heap_ptr = k1;
	Node* i = a->i;   a->i   = b->i;   b->i   = i;
	double key = a->key; a->key = b->key; b->key = key;
}

inline void OptTransport::PriorityQueue::Add(Node* i, double key)
{
	if (N == arraySize)
	{
		arraySize *= 2;
		array = (Item*) realloc(array, arraySize*sizeof(Item));
	}
	int k = i->heap_ptr = N ++;
	array[k].i = i;
	array[k].key = key;
	while (k > 0)
	{
		int k2 = (k-1)/2;
		if (array[k2].key <= array[k].key) break;
		Swap(k, k2);
		k = k2;
	}
}

inline void OptTransport::PriorityQueue::DecreaseKey(Node* i, double key)
{
	int k = i->heap_ptr;
	array[k].key = key;
	while (k > 0)
	{
		int k2 = (k-1)/2;
		if (array[k2].key <= array[k].key) break;
		Swap(k, k2);
		k = k2;
	}
}

inline typename OptTransport::Node* OptTransport::PriorityQueue::RemoveMin(double& key)
{
	if (N == 0) return NULL;

	Swap(0, N-1);
	N --;

	int k = 0;
	while ( 1 )
	{
		int k1 = 2*k + 1, k2 = k1 + 1;
		if (k1 >= N) break;
		int k_min = (k2 >= N || array[k1].key <= array[k2].key) ? k1 : k2;
		if (array[k].key <= array[k_min].key) break;
		Swap(k, k_min);
		k = k_min;
	}

	key = array[N].key;
	return array[N].i;
}


#endif
