#ifndef __PAIRWISE_COST__
#define __PAIRWISE_COST__

#include "../LRSolver/block.h"
#include "opt_transport.h"

// Implements pairwise term V[a,b] where a\in[0,A-1] and b\in[0,B-1]
// Sizes A, B don't need to be stored - they are passed in every call
struct PairwiseTerm
{
	double get(int a, int b, int A, int B);

	// must implement the following update for every 'b':
	// - find 'a' that minimizes Da[a]+V[a,b]
	// - update Db[b] += Da[a]+V[a,b], back_ptr_b[b] = a
	void pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B);

	// must implement the following
	// 1. for every 'b' set
	//   Db[b] = min_a { - minus_Da[a] + V[a,b] }
	// 2. compute delta = min_b Db
	//    Db[b] -= delta
	// 3. return delta
	double pass_message_srmp_fw(double* minus_Da, double* Db, int A, int B);

	// must implement the following update for every 'a':
	// - find 'b' that minimizes Db[b]+V[a,b]
	// - update Da[a] += Db[b]+V[a,b], back_ptr_a[a] = b
	void pass_message_bw(double* Da, double* Db, int* back_ptr_a, int A, int B);

	double pass_message_srmp_bw(double* Da, double* minus_Db, int A, int B);

	// this function must swap the meaning of 'a' and 'b'.
	// 'rbuf' is provided for convenience (doesn't have to be used).
	void swap(int A, int B, ReusableBuffer* rbuf);

	void print(int A, int B); // for debugging

	////////////////////////////////////////////////////////////////

	// The LRSolver code will contract pairwise terms, meaning that
	// it will work with a restriction of V[a,b] where a is restricted to 'AA' possible labels
	// and b will be restricted to 'BB' possible labels.

	// If this function returns true, then the code will take care of the contraction
	bool explicit_contraction(int A, int B, int AA, int BB); // 1<=AA<=A, 1<=BB<=B

	// Otherwise the user needs to implement the function below.
	// Labels for the first node: list_a[0]-a_shift, list_a[1]-a_shift, ..., list_a[AA]-a_shift
	// These labels form a strictly increasing sequence that belongs to 0,1,...,A-1
	// Similarly for the second node.
	// 'Da' is of size 'AA'
	// 'Db' and 'back_ptr_b' are of size 'BB'.
	//
	// must implement the following update for every 'b':
	// - find 'a' that minimizes Da[a]+V[ list_a[a]-a_shift , list_b[b]-b_shift ]
	// - update Db[b] += Da[a]+V[ list_a[a]-a_shift , list_b[b]-b_shift ], back_ptr_b[b] = a
	void pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B,
		int AA, int BB, int* list_a, int* list_b, int a_shift, int b_shift);


	double opt_transport(double* x, double* y, int A, int B);
	double opt_transport(double* x, double* y, int A, int B,
		int AA, int BB, int* list_a, int* list_b, int a_shift, int b_shift);

	////////////////////////////////////////////////////////////////

	typedef enum
	{
		GENERAL = 0,
		POTTS   = 1
	} Type;
	Type type;
};

/// Two specializations: 'PairwiseTermGeneral' and 'PairwiseTermPotts'

#include "pairwise_term_general.h"
#include "pairwise_term_potts.h"

inline PairwiseTerm* CheckAndAllocPairwiseTerm(int A, int B, double* V, Buffer* buf)
{
	PairwiseTerm* T = AllocPairwiseTermPotts(A, B, V, buf);
	if (T) return T;
	return AllocPairwiseTermGeneral(A, B, V, buf);
}


inline double PairwiseTerm::get(int a, int b, int A, int B)
{
	switch (type)
	{
		case GENERAL: return ((PairwiseTermGeneral*)this)->get(a, b, A, B);
		case POTTS:   return ((PairwiseTermPotts*  )this)->get(a, b, A, B);
		default: exit(1);
	}
}

inline void PairwiseTerm::pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B)
{
	switch (type)
	{
		case GENERAL: ((PairwiseTermGeneral*)this)->pass_message_fw(Da, Db, back_ptr_b, A, B); break;
		case POTTS:   ((PairwiseTermPotts*)this  )->pass_message(Da, Db, back_ptr_b, A); break;
		default: exit(1);
	}
}

inline double PairwiseTerm::pass_message_srmp_fw(double* minus_Da, double* Db, int A, int B)
{
	switch (type)
	{
		case GENERAL: return ((PairwiseTermGeneral*)this)->pass_message_srmp_fw(minus_Da, Db, A, B);
		case POTTS:   return ((PairwiseTermPotts*)this  )->pass_message_srmp(minus_Da, Db, A);
		default: exit(1);
	}
}

inline void PairwiseTerm::pass_message_bw(double* Da, double* Db, int* back_ptr_a, int A, int B)
{
	switch (type)
	{
		case GENERAL: ((PairwiseTermGeneral*)this)->pass_message_bw(Da, Db, back_ptr_a, A, B); break;
		case POTTS:   ((PairwiseTermPotts*)this  )->pass_message(Da, Db, back_ptr_a, A); break;
		default: exit(1);
	}
}

inline double PairwiseTerm::pass_message_srmp_bw(double* Da, double* minus_Db, int A, int B)
{
	switch (type)
	{
		case GENERAL: return ((PairwiseTermGeneral*)this)->pass_message_srmp_bw(Da, minus_Db, A, B);
		case POTTS:   return ((PairwiseTermPotts*)this  )->pass_message_srmp(minus_Db, Da, A);
		default: exit(1);
	}
}

inline void PairwiseTerm::swap(int A, int B, ReusableBuffer* rbuf)
{
	switch (type)
	{
		case GENERAL: ((PairwiseTermGeneral*)this)->swap(A, B, rbuf); break;
		case POTTS: break;
		default: exit(1);
	}
}

inline void PairwiseTerm::print(int A, int B)
{
	switch (type)
	{
		case GENERAL: ((PairwiseTermGeneral*)this)->print(A, B); break;
		case POTTS:   ((PairwiseTermPotts*)this  )->print(A); break;
		default: exit(1);
	}
}

inline bool PairwiseTerm::explicit_contraction(int A, int B, int AA, int BB)
{
	switch (type)
	{
		case GENERAL: return ((PairwiseTermGeneral*)this)->explicit_contraction(A, B, AA, BB);
		case POTTS:   return ((PairwiseTermPotts*  )this)->explicit_contraction(A, B, AA, BB);
		default: exit(1);
	}
}

inline void PairwiseTerm::pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B,
	int AA, int BB, int* list_a, int* list_b, int a_shift, int b_shift)
{
	switch (type)
	{
		case GENERAL: ((PairwiseTermGeneral*)this)->pass_message_fw(Da, Db, back_ptr_b, A, B, AA, BB, list_a, list_b, a_shift, b_shift); break;
		case POTTS:   ((PairwiseTermPotts*)this  )->pass_message_fw(Da, Db, back_ptr_b, A, B, AA, BB, list_a, list_b, a_shift, b_shift); break;
		default: exit(1);
	}
}

inline double PairwiseTerm::opt_transport(double* x, double* y, int A, int B)
{
	int a, b;
	int U = 0, V = 0;
	for (a=0; a<A; a++) { if (x[a] > 0) U ++; }
	for (b=0; b<B; b++) { if (y[b] > 0) V ++; }

	if (U == 0 || V == 0) return 0;

	char* buf = (char*) malloc((U + V + 2*U*V)*sizeof(double) + (U+V)*sizeof(int));
	char* ptr = buf;
	double* u = (double*) ptr; ptr += U*sizeof(double);
	double* v = (double*) ptr; ptr += V*sizeof(double);
	double* costs = (double*) ptr; ptr += U*V*sizeof(double);
	double* z = (double*) ptr; ptr += U*V*sizeof(double);
	int* map_x = (int*) ptr; ptr += U*sizeof(int);
	int* map_y = (int*) ptr; ptr += V*sizeof(int);

	U = V = 0;
	for (a=0; a<A; a++) { if (x[a] > 0) { u[U] = x[a]; map_x[U ++] = a; } }
	for (b=0; b<B; b++) { if (y[b] > 0) { v[V] = y[b]; map_y[V ++] = b; } }
	for (a=0; a<U; a++)
	for (b=0; b<V; b++)
	{
		costs[a + U*b] = get(map_x[a], map_y[b], A, B);
	}

	OptTransport OT;
	OT.Solve(U, V, costs, u, v, z);
	double xy_cost = 0;
	for (a=0; a<U*V; a++) xy_cost += costs[a]*z[a];

	free(buf);

	return xy_cost;
}

inline double PairwiseTerm::opt_transport(double* x, double* y, int A, int B,
	int AA, int BB, int* list_a, int* list_b, int a_shift, int b_shift)
{
	int a, b;
	int U = 0, V = 0;
	for (a=0; a<AA; a++) { if (x[a] > 0) U ++; }
	for (b=0; b<BB; b++) { if (y[b] > 0) V ++; }

	if (U == 0 || V == 0) return 0;

	char* buf = (char*) malloc((U + V + 2*U*V)*sizeof(double) + (U+V)*sizeof(int));
	char* ptr = buf;
	double* u = (double*) ptr; ptr += U*sizeof(double);
	double* v = (double*) ptr; ptr += V*sizeof(double);
	double* costs = (double*) ptr; ptr += U*V*sizeof(double);
	double* z = (double*) ptr; ptr += U*V*sizeof(double);
	int* map_x = (int*) ptr; ptr += U*sizeof(int);
	int* map_y = (int*) ptr; ptr += V*sizeof(int);

	U = V = 0;
	for (a=0; a<AA; a++) { if (x[a] > 0) { u[U] = x[a]; map_x[U ++] = a; } }
	for (b=0; b<BB; b++) { if (y[b] > 0) { v[V] = y[b]; map_y[V ++] = b; } }
	for (a=0; a<U; a++)
	for (b=0; b<V; b++)
	{
		costs[a + U*b] = get(list_a[map_x[a]]-a_shift, list_b[map_y[b]]-b_shift, A, B);
	}

	OptTransport OT;
	OT.Solve(U, V, costs, u, v, z);
	double xy_cost = 0;
	for (a=0; a<U*V; a++) xy_cost += costs[a]*z[a];

	delete [] buf;

	return xy_cost;
}

#endif
