#ifndef __PAIRWISE_TERM_GENERAL__
#define __PAIRWISE_TERM_GENERAL__

struct PairwiseTermGeneral : PairwiseTerm
{
	double get(int a, int b, int A, int B);
	void pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B);
	void pass_message_bw(double* Da, double* Db, int* back_ptr_a, int A, int B);
	double pass_message_srmp_fw(double* minus_Da, double* Db, int A, int B);
	double pass_message_srmp_bw(double* Da, double* minus_Db, int A, int B);
	void swap(int A, int B, ReusableBuffer* rbuf);
	void print(int A, int B);

	bool explicit_contraction(int A, int B, int AA, int BB) { return true; }
	void pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B,
		int AA, int BB, int* list_a, int* list_b, int a_shift, int b_shift); // doesn't need to be implemented, but let's implement anyway

	// used for contraction in TreeMRF.h
	void init(PairwiseTerm* T, int A, int B, int AA, int BB, int* list_a, int* list_b);

	double V[1];
};

inline int PairwiseTermGeneralSize(int A, int B)
{
	return sizeof(PairwiseTermGeneral) + (A*B-1)*sizeof(double);
}

inline PairwiseTermGeneral* AllocPairwiseTermGeneral(int A, int B, double* V, Buffer* buf)
{
	PairwiseTermGeneral* T = (PairwiseTermGeneral*) buf->Alloc(PairwiseTermGeneralSize(A, B));
	T->type = PairwiseTerm::Type::GENERAL;

	if (V)
	{
		memcpy(T->V, V, A*B*sizeof(double));
	}
	return T;
}

//////////////////////////

inline void PairwiseTermGeneral::swap(int A, int B, ReusableBuffer* rbuf)
{
	int a, b;
	double* tmp = (double*) rbuf->Alloc(A*B*sizeof(double));
	memcpy(tmp, V, A*B*sizeof(double));
	for (a=0; a<A; a++)
	for (b=0; b<B; b++)
	{
		V[b + B*a] = tmp[a + A*b];
	}
}

inline double PairwiseTermGeneral::get(int a, int b, int A, int B)
{
	return V[a+A*b];
}

inline void PairwiseTermGeneral::pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B)
{
	int a, b;

	double* ptr = V;
	for (b=0; b<B; b++)
	{
		int a_best = 0;
		double cost_best = Da[0] + (*ptr ++);
		for (a=1; a<A; a++)
		{
			double cost = Da[a] + (*ptr ++);
			if (cost_best > cost) { cost_best = cost; a_best = a; }
		}
		back_ptr_b[b] = a_best;
		Db[b] += cost_best;
	}
}

inline double PairwiseTermGeneral::pass_message_srmp_fw(double* minus_Da, double* Db, int A, int B)
{
	int a, b;

	double* ptr = V;
	for (b=0; b<B; b++)
	{
		double cost_best = -minus_Da[0] + (*ptr ++);
		for (a=1; a<A; a++)
		{
			double cost = -minus_Da[a] + (*ptr ++);
			if (cost_best > cost) { cost_best = cost; }
		}
		Db[b] = cost_best;
	}

	double delta = Db[0];
	for (b=1; b<B; b++)
	{
		if (delta > Db[b]) delta = Db[b];
	}
	for (b=0; b<B; b++) Db[b] -= delta;
	return delta;
}

inline void PairwiseTermGeneral::pass_message_bw(double* Da, double* Db, int* back_ptr_a, int A, int B)
{
	int a, b;

	for (a=0; a<A; a++)
	{
		int b_best = 0;
		double* ptr = V + a;
		double cost_best = Db[0] + (*ptr); ptr += A;
		for (b=1; b<B; b++)
		{
			double cost = Db[b] + (*ptr); ptr += A;
			if (cost_best > cost) { cost_best = cost; b_best = b; }
		}
		back_ptr_a[a] = b_best;
		Da[a] += cost_best;
	}
}

inline double PairwiseTermGeneral::pass_message_srmp_bw(double* Da, double* minus_Db, int A, int B)
{
	int a, b;

	for (a=0; a<A; a++)
	{
		double* ptr = V + a;
		double cost_best = -minus_Db[0] + (*ptr); ptr += A;
		for (b=1; b<B; b++)
		{
			double cost = -minus_Db[b] + (*ptr); ptr += A;
			if (cost_best > cost) { cost_best = cost; }
		}
		Da[a] = cost_best;
	}

	double delta = Da[0];
	for (a=1; a<A; a++)
	{
		if (delta > Da[a]) delta = Da[a];
	}
	for (a=0; a<A; a++) Da[a] -= delta;
	return delta;
}

inline void PairwiseTermGeneral::print(int A, int B)
{
	int a, b;
	double* ptr = V;
	for (a=0; a<A; a++)
	{
		printf("\t");
		for (b=0; b<B; b++) printf(" %f", *ptr ++);
		printf("\n");
	}
}

inline void PairwiseTermGeneral::pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B,
	int AA, int BB, int* list_a, int* list_b, int a_shift, int b_shift)
{
	int a, b;

	for (b=0; b<BB; b++)
	{
		int a_best = 0;
		double cost_best = Da[0] + V[(list_a[0]-a_shift) + A*(list_b[b]-b_shift)];
		for (a=1; a<AA; a++)
		{
			double cost = Da[a] + V[(list_a[a]-a_shift) + A*(list_b[b]-b_shift)];
			if (cost_best > cost) { cost_best = cost; a_best = a; }
		}
		back_ptr_b[b] = a_best;
		Db[b] += cost_best;
	}
}


inline void PairwiseTermGeneral::init(PairwiseTerm* T, int A, int B, int AA, int BB, int* list_a, int* list_b)
{
	int a, b;

	double* ptr = V;

	for (b=0; b<BB; b++)
	for (a=0; a<AA; a++)
	{
		*ptr ++ = T->get(list_a[a], list_b[b], A, B);
	}
}


#endif
