#ifndef __PRIORITY_QUEUE__
#define __PRIORITY_QUEUE__

struct PriorityQueue
{
	struct Node
	{
		int heap_ptr;
	};

	PriorityQueue();
	~PriorityQueue();
	void Reset();
	double GetKey(Node* i);
	void Add(Node* i, double key);
	void DecreaseKey(Node* i, double key);
	Node* RemoveMin(double& key);

//private:
	struct Item
	{
		Node*		i;
		double		key;
	}* items;
	int N, arraySize;
	void Swap(int k1, int k2);
};

inline PriorityQueue::PriorityQueue()
{
	N = 0;
	arraySize = 16;
	items = (Item*) malloc(arraySize*sizeof(Item));
}

inline PriorityQueue::~PriorityQueue()
{
	free(items);
}

inline void PriorityQueue::Reset()
{
	N = 0;
}

inline double PriorityQueue::GetKey(Node* i)
{
	return items[i->heap_ptr].key;
}

inline void PriorityQueue::Swap(int k1, int k2)
{
	Item* a = items+k1;
	Item* b = items+k2;
	a->i->heap_ptr = k2;
	b->i->heap_ptr = k1;
	Node* i = a->i;   a->i   = b->i;   b->i   = i;
	double key = a->key; a->key = b->key; b->key = key;
}

inline void PriorityQueue::Add(Node* i, double key)
{
	if (N == arraySize)
	{
		arraySize *= 2;
		items = (Item*) realloc(items, arraySize*sizeof(Item));
	}
	int k = i->heap_ptr = N ++;
	items[k].i = i;
	items[k].key = key;
	while (k > 0)
	{
		int k2 = (k-1)/2;
		if (items[k2].key <= items[k].key) break;
		Swap(k, k2);
		k = k2;
	}
}

inline void PriorityQueue::DecreaseKey(Node* i, double key)
{
	int k = i->heap_ptr;
	items[k].key = key;
	while (k > 0)
	{
		int k2 = (k-1)/2;
		if (items[k2].key <= items[k].key) break;
		Swap(k, k2);
		k = k2;
	}
}

inline PriorityQueue::Node* PriorityQueue::RemoveMin(double& key)
{
	if (N == 0) return NULL;

	Swap(0, N-1);
	N --;

	int k = 0;
	while ( 1 )
	{
		int k1 = 2*k + 1, k2 = k1 + 1;
		if (k1 >= N) break;
		int k_min = (k2 >= N || items[k1].key <= items[k2].key) ? k1 : k2;
		if (items[k].key <= items[k_min].key) break;
		Swap(k, k_min);
		k = k_min;
	}

	key = items[N].key;
	return items[N].i;
}



/////////////////////////////////////////////////////////////////////////////////


struct DynamicDijsktra // in undirected graph
{
	typedef int NodeId;
	typedef int EdgeId;

	DynamicDijsktra(int node_num, int edge_num_max);
	~DynamicDijsktra();

	EdgeId AddEdge(NodeId i, NodeId j, double w); // w >= 0
	void UpdateWeight(EdgeId e, double w) { edges[e].w = w; }

	// stops if either (i) the distance exceeds max_d, or (ii) sink becomes fixed.
	// In the latter case the distance from source to sink is returned.
	// Note, sink can be -1.
	// If source_distances != NULL then source_distances[i] is set to the distance from source ONLY FOR FIXED NODES.
	double Run(NodeId source, NodeId sink, double max_d, double* source_distances=NULL);

	// after Run(), the following functions can be called to get the result.
	bool isFixed(NodeId i) { return (nodes[i].heap_ptr <= -2); }
	EdgeId GetParentEdge(NodeId i) { return nodes[i].parent; } // can be called only for fixed nodes
	NodeId GetParentNode(NodeId i) { Edge* e = edges+nodes[i].parent; return (e->i[0] == i) ? e->i[1] : e->i[0]; } // can be called only for fixed nodes

	NodeId GetFirstFixed() { fixed_first_iterator = fixed_first; return GetNextFixed(); }   // get a list of fixed nodes
	NodeId GetNextFixed()
	{
			if (fixed_first_iterator >= node_num) return -1;
			int i = fixed_first_iterator;
			fixed_first_iterator = -2 - nodes[fixed_first_iterator].heap_ptr;
			return i;
	}   // (-1 indicates the end of the list)

//private:
	struct Edge;
	struct Node : PriorityQueue::Node
	{
		Edge* first[2];
		NodeId parent;
	};
	struct Edge
	{
		NodeId i[2];
		Edge* next[2];
		double w;
	};

	int node_num, edge_num_max, edge_num;
	int fixed_first, fixed_first_iterator;
	Node* nodes;
	Edge* edges;
	PriorityQueue queue;

	void ResetRun();
};

inline DynamicDijsktra::DynamicDijsktra(int _node_num, int _edge_num_max)
	: node_num(_node_num), edge_num_max(_edge_num_max), edge_num(0)
{
	int i;
	nodes = new Node[node_num];
	edges = new Edge[edge_num_max];
	for (i=0; i<node_num; i++)
	{
		nodes[i].first[0] = nodes[i].first[1] = NULL;
		nodes[i].parent = -1;
		nodes[i].heap_ptr = -1;
	}
	fixed_first = node_num;
}

inline DynamicDijsktra::~DynamicDijsktra()
{
	delete [] nodes;
	delete [] edges;
}

inline DynamicDijsktra::EdgeId DynamicDijsktra::AddEdge(NodeId i, NodeId j, double w)
{
	if (edge_num >= edge_num_max) { printf("too many edges\n"); exit(1); }
	Edge* e = &edges[edge_num];
	e->i[0] = i; e->i[1] = j;
	e->next[0] = nodes[i].first[0];
	e->next[1] = nodes[j].first[1];
	nodes[i].first[0] = nodes[j].first[1] = e;
	e->w = w;
	return edge_num ++;
}

inline double DynamicDijsktra::Run(NodeId source, NodeId sink, double max_d, double* source_distances)
{
	ResetRun();

	int dir;
	Node* i;
	Edge* e;

	queue.Add(&nodes[source], 0);
	while ( 1 )
	{
		double key;
		PriorityQueue::Node* pi = queue.RemoveMin(key);
		if (pi == NULL)
		{
			return 1e300;
		}
		i = ((Node*)pi);
		int _i = (int)(i-nodes);
		if (source_distances) source_distances[_i] = key;

		i->heap_ptr = -2-fixed_first;
		fixed_first = _i;

		if (_i == sink)
		{
			return key;
		}

		for (dir=0; dir<2; dir++)
		for (e=i->first[dir]; e; e=e->next[dir])
		{
			int j = e->i[1-dir];
			if (nodes[j].heap_ptr <= -2) continue;
			double j_key = key+e->w;
			if (j_key >= max_d) continue;
			if (nodes[j].heap_ptr == -1)
			{
				nodes[j].parent = (int)(e-edges);
				queue.Add(&nodes[j], j_key);
			}
			else
			{
				if (j_key < queue.GetKey(&nodes[j]))
				{
					nodes[j].parent = (int)(e-edges);
					queue.DecreaseKey(&nodes[j], j_key);
				}
			}
		}
	}
}

inline void DynamicDijsktra::ResetRun()
{
	if (fixed_first == node_num) return;

	PriorityQueue::Item* I;
	for (I=queue.items; I<queue.items+queue.N; I++)
	{
		Node* i = (Node*)(I->i);
		i->parent = -1;
		i->heap_ptr = -1;
	}
	while (fixed_first < node_num)
	{
		Node* i = nodes + fixed_first;
		fixed_first = -2-i->heap_ptr;
		i->parent = -1;
		i->heap_ptr = -1;
	}

	if ( 0 )
	{
		Node* i;
		for (i=nodes; i<nodes+node_num; i++)
		{
			if (i->parent != -1 || i->heap_ptr != -1)
			{
				printf("error in ResetRun()\n");
				exit(1);
			}
		}
	}
	fixed_first = node_num;
	queue.Reset();
}

#endif

